﻿#define _USE_MATH_DEFINES
#include "ClimbAnalyse.h"
		
ClimbAnalyse::ClimbAnalyse(OSM* O,short mintype, short maxtype, double minP,
								  double minS, double maxdwnHM, double minHM,
                                  HeightData* h){
    this->o = O;
    this->h = h;
	set_parameters(mintype, maxtype, minP, minS, maxdwnHM, minHM);
    f = new FibunacciHeap<CAfib>;
	cp = NULL;
	climbpool = NULL;
	climbcount = 0;
    minlat = -90;
    maxlat = 90;
    minlon = -180;
    maxlon = 180;
    analyse();
}
ClimbAnalyse::ClimbAnalyse(){
	f = NULL;
	cp = NULL;
	climbpool = NULL;
	climbcount = 0;
}

ClimbAnalyse::~ClimbAnalyse(){
    if(f != NULL) f->~FibunacciHeap();
	for(long i = 0; i < climbcount; i++){
		free_tpl(climbpool[i].c->t->prev,climbpool[i].c->prev->t);
	}
	for(long i = 0; i < climbcount; i++){
        if(climbpool[i].t != NULL)climbpool[i].t->~Track();
		if(climbpool[i].c->t != NULL) free(climbpool[i].c->t);
	}
	if(climbpool != NULL)
		free(climbpool);
}
void ClimbAnalyse::analyse(){
	reset_STRONG(o->get_first_KOO());
	init_fib(o->get_first_KOO());
    CAfib* c = f->extmin();
	while(c != NULL){
		if(all_active(c->c))
			expand(c->c);
		else{
			if(c->c->state != climbend)
				del_climb(c->c);
		}
		free(c);
		c = (CAfib*) f->extmin();
	}
	climbcount = 0;
    count_climbs(o->get_first_KOO());
    if(climbcount > 0){
        //QMessageBox(QMessageBox::NoIcon,"climbcount",QString("%1").arg(climbcount)).exec();
        climbpool = (climbpoolele*)malloc(climbcount*sizeof(climbpoolele));
        climbcount = 0;
        collect_climbs(o->get_first_KOO());
        if(climbcount > 0){
            sort_climbpool(0,climbcount-1);
            select_climbs();
            create_names();
        }
    }
    else
        climbpool = NULL;
}
float ClimbAnalyse::calc_P(trackpoint* t,int arraylength){
	float result = 0, hup, hdown;
	for(int i = 0; i < arraylength -1; i++){
		hup = 0;
		hdown = 0;
		for(int j = i+1; j < arraylength; j++){
			if(t[j].h > t[j-1].h)
				hup = hup + t[j].h - t[j-1].h;
			else
				hdown = hdown -t[j].h + t[j-1].h;
			if(		hup*maxdwnHM > hdown
				&&	hup / 1000. / (t[j].d - t[i].d) > minSlope
				&&	hup*hup / 1000. / (t[j].d - t[i].d) > result)
				result = float(hup*hup / 1000. / (t[j].d - t[i].d));
		}
	}
	return result;
}
void ClimbAnalyse::set_parameters(short mintype, short maxtype, double minP,
								  double minS, double maxdwnHM, double minHM){
	this->mintype = mintype;
	this->maxtype = maxtype;
	this->minP = minP;
	this->minSlope = minS;
	this->maxdwnHM = maxdwnHM;
	this->minHM = minHM;
    autonomP = 5;
}
void ClimbAnalyse::reset_STRONG(KOO* k){
	if(k!= NULL){
		if(k->cd != NULL){
			k->cd->s->from = NULL;
			k->cd->s->via = NULL;
		}
		reset_STRONG(k->r);
		reset_STRONG(k->l);
	}
}
void ClimbAnalyse::init_fib(KOO* k){
	if(k != NULL){
		if(k->cd != NULL){
			climb* c = (climb*)malloc(sizeof(climb));
			init_climb(c, k);
			expand(c);
		}
		init_fib(k->l);
		init_fib(k->r);
	}
}
void ClimbAnalyse::init_climb(climb* c, KOO* k){
	c->branches = 0;
	c->cst = NULL;
	c->d = 0;
	c->hend = 0;
	c->hmdown = 0;
	c->hmup = 0;
	c->home = k;
	c->hstart = 0;
	c->next = NULL;
	c->P = 0;
	c->Pmax = 0;
	c->Prec = 0;
	c->prev = NULL;
	c->root = c;
	c->state = climbstart;
	c->t = NULL;
	c->tend = NULL;
	c->tst = NULL;
	c->via = NULL;
}
void ClimbAnalyse::expand(climb* c){
    Neighbour* n = c->home->cd->neighbours;
	trackpointlist* t;
	
 	while(n != NULL){
		if(n->way->type <= maxtype && n->way->type >= mintype &&
            crossunused(c,n->nb) && !n->way->IsStair){
		t = add(c->t,n->tp,n->tpc,n->rd);
		if(!storeif_climb(n,t,c))
			free_tpl(t,c->t);
		else
			c->branches++;
		}
		n = n->next;
	}
}
void ClimbAnalyse::free_tpl(trackpointlist* tp, trackpointlist* tpstop){
	if(tp != NULL && tp != tpstop){
		free_tpl(tp->prev,tpstop);
		free(tp);
	}
	else{
		if(tp != NULL){
			tp->next = NULL;
		}
	}
}
trackpointlist* ClimbAnalyse::add(trackpointlist* tp, trackpoint* t, int tpc,
								  readdirection rd){
	trackpointlist *tpl;
	float doffset;
	if(tp != NULL){
        // append to existing list
		doffset = tp->t.d;
		tpl = tp;
	}
	else{
        // created new list
		doffset = 0;
		tpl = (trackpointlist*)malloc(sizeof(trackpointlist));
		tpl->prev = NULL;
		tpl->t.d = 0;
		if(rd == fw)
			tpl->t.h = t[0].h;
		else
			tpl->t.h = t[tpc-1].h;
	}
    // read and insert
	if(rd == fw){
		for(int i = 1; i < tpc; i++){
			tpl->next = (trackpointlist*)malloc(sizeof(trackpointlist));
			tpl->next->prev = tpl;
			tpl = tpl->next;
			tpl->t.h = t[i].h;
			tpl->t.d = t[i].d + doffset;
		}
	}
	else{
		float dmax = t[tpc-1].d;
		for(int i = tpc-2; i >= 0; i--){
			tpl->next = (trackpointlist*)malloc(sizeof(trackpointlist));
			tpl->next->prev = tpl;
			tpl = tpl->next;
			tpl->t.h = t[i].h;
			tpl->t.d = (dmax-t[i].d) + doffset;
		}
	}
	tpl->next = NULL;
	return tpl;
}
short ClimbAnalyse::is_climb(trackpointlist* tp){
	float hmdown= 0, hmup = 0;
	trackpointlist* t = tp;
	while(t->prev != NULL){
		if(t->prev->t.h < t->t.h)
			hmdown = hmdown - t->prev->t.h + t->t.h;
		else
			hmup = hmup + t->prev->t.h - t->t.h;
		if(  hmdown*hmdown/(tp->t.d-t->prev->t.d)/1000.> minP
			&&hmdown/(tp->t.d-t->prev->t.d)/1000. > minSlope
		   &&hmdown*maxdwnHM > hmup)
		   return 1;
		t = t->prev;
	}
	return 0;
}
short ClimbAnalyse::all_active(climb* c){
	climb* ct = c->prev;
	if(c->state == outofuse)
		return 0;
	while(ct != NULL){
		if(ct->state == outofuse || c->home == ct->home){
			return 0;
		}
		ct =ct->prev;
	}
	return 1;
}
inline short ClimbAnalyse::enough_active(climb* c){
	climb* ct = c, *cfirst = NULL ,*clast = NULL;
	float dautonom = 0;
	while(ct != c->cst){
		if(clast == NULL && ct->state == outofuse)
			clast = ct;
		if(ct->state == outofuse)
			cfirst = ct;
		else{
			if(ct == c)
				dautonom = dautonom + ct->tend->t.d;
			else
				dautonom = dautonom + ct->t->t.d;
			if(ct->prev != c->cst)
				dautonom = dautonom - ct->prev->t->t.d;
			else
				dautonom = dautonom - c->tst->t.d;
		}
		ct =ct->prev;
	}
	if(( cfirst == NULL) || 
	   ((c->Pmax - clast->P > autonomP || cfirst->prev->Pmax > autonomP) &&
		dautonom/(c->tend->t.d-c->tst->t.d) > 0.5))
		return 1;
	else
		return 0;
}
void ClimbAnalyse::calc_climb(climb* c){
	c->Prec = c->prev->Prec;
	c->cst = c->prev->cst;
	c->P = 0;
	c->Pmax = 0;
	c->d = 0;
	c->hend = 0;
	c->hstart = 0;
	c->hmup = 0;
    c->hmdown = 0;

    float hmdown= 0, hmup = 0;
    trackpointlist* t = c->t, *t2;
    climb* ccst = c;
    while(t->prev != NULL && t != c->prev->t){
        t2 = t;
        hmdown = 0;
		hmup = 0;
        ccst = c;
        while(t2->prev != NULL){
            // refresh height meters
            if(t2->prev->t.h < t2->t.h)
                hmdown	= hmdown - t2->prev->t.h + t2->t.h;
            else
                hmup	= hmup	 + t2->prev->t.h - t2->t.h;
            // posibly refhresh climb start
            if(ccst->t != NULL && t2->prev->t.d < ccst->t->t.d)
                ccst = ccst->prev;
            // calc climb
            if( (t->t.d-t2->prev->t.d) != 0 &&
                //(c->Prec < minP || hmdown > minHM) &&
				hmup < maxdwnHM*hmdown &&
                hmdown/(t->t.d-t2->prev->t.d)/1000.> minSlope &&
                hmdown/(t->t.d-t2->prev->t.d)/1000.< 0.35 &&
                hmdown*hmdown/(t->t.d-t2->prev->t.d)/1000.> c->Pmax){
                // climb definitions are matched
                c->Pmax = float(hmdown*hmdown/(t->t.d-t2->prev->t.d)/1000.);
                c->tst = t2->prev;
				c->tend = t;
                c->d = t->t.d-t2->prev->t.d;
				c->cst = ccst;
                c->hend = t->t.h;
                c->hstart = t2->prev->t.h;
				c->hmup = hmdown;
				c->hmdown = hmup;
                if(t == c->t){
                    // climb end exactly at cross
                    c->P = c->Pmax;
                }
                if(c->Pmax > c->Prec)
					c->Prec = c->Pmax;
			}
            t2 = t2->prev;
		}
        t = t->prev;
	}
}
short ClimbAnalyse::storeif_climb(Neighbour* n, trackpointlist* tp, climb* from){
	
	climb* c = (climb*)malloc(sizeof(climb));
	c->branches = 0;
	c->home = n->nb;
	c->prev = from;
	c->t = tp;
	c->via = n;
	c->state = failure;
    calc_climb(c);

	short keep, expand;
    // check whether climbs could exist on itseld
	if(c->P == 0){
        // climb cannot further expand as it breaks clibm definiton
		if(c->Pmax > minP){
            // but somewhere within the climb it reached climb state
			keep = 1;
			expand = 0;
			c->state = climbend;
		}
		else{
            // never reached climb state
			keep = 0;
			expand = 0;
		}
	}
	else{
        // climb matches climb definitions
		keep = 1;
		expand = 1;
		if(c->P == c->Prec)
			c->state = increasing;
		else
			c->state = decreasing;
	}

	if(expand){
        // climb is supposed to expand, but has to fight existing climbs
		climb* ct = (climb*)(n->nb->cd->s->from);
		while(ct != NULL){
            if(looses(ct,c)){
				ct = NULL;
				keep = expand = 0;
				c->state = outofuse;
			}
			else
				ct = ct->next;
		}
	}

	if(keep){
		c->next = (climb*) c->home->cd->s->from;
		c->home->cd->s->from = c;
		CAfib* cfib = (CAfib*) f->ins(c->t->t.h);
		cfib->c = c;
		climb* ct = c->next;
		while(ct != NULL){
            if(looses(c,ct))
				ct->state = outofuse;
			ct = ct->next;
		}
		return 1;
	}
	else{
		free(c);
		return 0;
	}
}
short ClimbAnalyse::crossunused(climb* c, KOO* k){
	while(c != NULL){
		if(c->home == k)
			return 0;
		c = c->prev;
	}
	return 1;
}
void  ClimbAnalyse::del_climb(climb* c){
    // delete only directly after extract_min() from FibunacciHeap, but bot due
    // to expanding climbs. Expanding climbs only set other climbs
    // c->state = outofuse
	if(c != NULL && c->branches == 0){
		climb dummy; dummy.next = (climb*)c->home->cd->s->from;
		climb* ct = &dummy;
		while(ct->next != NULL && ct->next != c){
			ct = ct->next;
		}
		if(ct->next != NULL){
			ct->next = ct->next->next;
			c->home->cd->s->from = dummy.next;
            if(c->prev != NULL){		//delete as far as possible
					free_tpl(c->t,c->prev->t);
					c->prev->branches--;
                    del_climb(c->prev); //automatically del if only 2 branches
			}
		}
		free(c);
	}
}
void ClimbAnalyse::count_climbs(KOO* k){
	if(k!= NULL){
		if(k->cd != NULL){
			climb* c = (climb*)(k->cd->s->from);
			while(c != NULL){
                if(c->Pmax == c->Prec){
					climbcount++;
                }
				c = c->next;
			}
		}
		count_climbs(k->l);
		count_climbs(k->r);
	}
}
void ClimbAnalyse::collect_climbs(KOO* k){
	if(k!= NULL){
		if(k->cd != NULL){
			climb* c = (climb*)(k->cd->s->from);
			while(c!= NULL){
                double a = 20;
                if(c->Pmax <9.7 && c->Pmax > 9.5 && c->hmup > 147)
                    a++;
                if(c->Pmax == c->Prec &&
				   c->hmup > minHM){
					climbpool[climbcount].c = c;
					climbpool[climbcount].P = c->Pmax;
					climbcount++;
				}
				c = c->next;
			}
		}
		collect_climbs(k->l);
		collect_climbs(k->r);
	}
}
void ClimbAnalyse::sort_climbpool(long s, long e){
	long l = s, r = e;
	climbpoolele c;
	float p = climbpool[(s+e)/2].P;
	do{
		while(climbpool[l].P < p) l++;
		while(climbpool[r].P > p) r--;
		if(l<=r){
			c = climbpool[l];
			climbpool[l] = climbpool[r];
			climbpool[r] = c;
			l++;
			r--;
		}
	}while(l <= r);
	if(s < r) sort_climbpool(s,r);
	if(l < e) sort_climbpool(l,e);
}
void ClimbAnalyse::select_climbs(){
	for(long i = climbcount-1; i >= 0; i--){
		if(!enough_active(climbpool[i].c)){
			climbpool[i].P = 0;
			climbpool[i].t = NULL;
		}
		else{
            create_climb_Track(&(climbpool[i]));
		}
        climbpool[i].name = NULL;
		climb* c = climbpool[i].c, *ct;
		while(c != climbpool[i].c->cst){
            // all climbs which belong to the pure climb
			ct = (climb*) c->home->cd->s->from;
			while(ct != NULL){
                // climbs which us this cross ...
				if(ct->via == c->via)
                    // ... and this Neighbour relation
					ct->state = outofuse;
				ct = ct->next;
			}
			c = c->prev;
		}
	}
}
short ClimbAnalyse::check(KOO* k){
	if(k != NULL){
		if(k->cd != NULL){
			if(k->cd->s->from != NULL){
				climb* c = (climb*)k->cd->s->from;
				while(c != NULL){
					c = c->next;
				}
			}
		}
		return check(k->l)*check(k->r);
	}
	else
		return 1;
}
void ClimbAnalyse::create_climb_Track(climbpoolele* c){
	climb* ct = c->c;
	int nc = 0;
	while(ct != c->c->cst){
		ct = ct->prev;
		nc++;
	}
    Neighbour** narray = (Neighbour**)malloc(nc*sizeof(Neighbour*));
	int i = nc;
	ct = c->c;
	while(ct != c->c->cst){
		i--;
		narray[i] = ct->via;
		ct = ct->prev;
	}

    c->t = new Track(narray,nc);
	float d0;
	if(c->c->cst->t == NULL)
		d0 = 0;
	else
		d0 = c->c->cst->t->t.d;
	c->t->trim_right(c->c->tend->t.d - d0);
	c->t->trim_left (c->c->tst->t.d - d0);
}
climblayer* ClimbAnalyse::get_ClimbLayer(){
	if(climbcount > 0 && climbpool != NULL){
		climblayer* result = (climblayer*)malloc(sizeof(climblayer));
        // count valid climbs
		result ->climbcount = 0;
		for(long i = 0; i < climbcount; i++)
			if(climbpool[i].P >= minP)
				result->climbcount++;
		result->climb = (climblayerelement*)malloc(result->climbcount*sizeof(climblayerelement));
		
		if(result->climbcount == 0){
			free(result);
			return NULL;
		}

        // extract valid climbs an note them
		long ci = 0;
        for(long i = 0; i < climbcount; i++){
			if(climbpool[i].P >= minP){
				result->climb[ci].P		= climbpool[i].P;
				result->climb[ci].hstart	= climbpool[i].c->hstart;
				result->climb[ci].hend	= climbpool[i].c->hend;
				result->climb[ci].hmdown	= climbpool[i].c->hmdown;
				result->climb[ci].hmup	= climbpool[i].c->hmup;
				result->climb[ci].d		= climbpool[i].c->d;
                result->climb[ci].dhcount  = climbpool[i].t->get_d_h_length();
                result->climb[ci].dharray = climbpool[i].t->get_d_h_array_copy();
                result->climb[ci].KOOcount = climbpool[i].t->get_KOOa_length();
                result->climb[ci].KOOarray = climbpool[i].t->get_KOOa_copy();
                result->climb[ci].slope	   = result->climb[ci].hmup/ result->climb[ci].d / 10.;
				result->climb[ci].bend = climbpool[i].t->get_bend()/climbpool[i].c->d;
				KOO k; climbpool[i].t->extract_KOO(&k,climbpool[i].c->d/2.);
				result->climb[ci].midlat = k.lat;
				result->climb[ci].midlon = k.lon;
                result->climb[ci].name = climbpool[i].name;
				ci++;
			}
		}
        // make local ranking of climbs
        for(long i = 0; i < result->climbcount; i++){
			float m = 0;
			for(long j = 0; j < result->climbcount; j++)
                m = qMax(m,
                        float(result->climb[j].P
						* exp(-distance(result->climb[i].midlat,
										result->climb[i].midlon,
										result->climb[j].midlat,
                                        result->climb[j].midlon)/30)));
			result->climb[i].LocRank = result->climb[i].P / m;
		}

		return result;
	}
	else
		return NULL;
}
climblayer* ClimbAnalyse::load_ClimbLayer(char* filename){
	FILE* f = fopen(filename,"rb");
	if(f != NULL){
		climblayer* c = (climblayer*)malloc(sizeof(climblayer));
		fread(&(c->climbcount),sizeof(c->climbcount),1,f);
		c->climb = (climblayerelement*)malloc(c->climbcount*sizeof(climblayerelement));
		for(long i = 0; i < c->climbcount; i++){
			fread(&(c->climb[i].d		 ),sizeof(c->climb[i].d		  ),1,f);
			fread(&(c->climb[i].dhcount	 ),sizeof(c->climb[i].dhcount ),1,f);
			fread(&(c->climb[i].hend	 ),sizeof(c->climb[i].hend	  ),1,f);
			fread(&(c->climb[i].hmdown	 ),sizeof(c->climb[i].hmdown  ),1,f);
			fread(&(c->climb[i].hmup	 ),sizeof(c->climb[i].hmup	  ),1,f);
			fread(&(c->climb[i].hstart	 ),sizeof(c->climb[i].hstart  ),1,f);
			fread(&(c->climb[i].KOOcount ),sizeof(c->climb[i].KOOcount),1,f);
			fread(&(c->climb[i].P		 ),sizeof(c->climb[i].P		  ),1,f);
			fread(&(c->climb[i].slope	 ),sizeof(c->climb[i].slope	  ),1,f);
			fread(&(c->climb[i].bend	 ),sizeof(c->climb[i].bend	  ),1,f);
			fread(&(c->climb[i].midlat	 ),sizeof(c->climb[i].midlat  ),1,f);
			fread(&(c->climb[i].midlon	 ),sizeof(c->climb[i].midlon  ),1,f);
			fread(&(c->climb[i].LocRank	 ),sizeof(c->climb[i].LocRank ),1,f);
            c->climb[i].dharray = new double*[2];
            c->climb[i].dharray[0] = new double[c->climb[i].dhcount];
            c->climb[i].dharray[1] = new double[c->climb[i].dhcount];
            fread(c->climb[i].dharray[0],sizeof(double),c->climb[i].dhcount,f);
            fread(c->climb[i].dharray[1],sizeof(double),c->climb[i].dhcount,f);
            c->climb[i].KOOarray = new double*[2];
            c->climb[i].KOOarray[0] = new double[c->climb[i].KOOcount];
            c->climb[i].KOOarray[1] = new double[c->climb[i].KOOcount];
            fread(c->climb[i].KOOarray[0],sizeof(double),c->climb[i].KOOcount,f);
            fread(c->climb[i].KOOarray[1],sizeof(double),c->climb[i].KOOcount,f);
            int length;
            fread(&length,sizeof(int),1,f);
            if(length > 0){
                c->climb[i].name = new char[length+1];
                fread(&(c->climb[i].name[0]),sizeof(char),length,f);
                c->climb[i].name[length] = '\0';
            }
            else
                c->climb[i].name = NULL;
		}
		return c;
	}
	else
		return NULL;
}
void ClimbAnalyse::save_ClimbLayer(char* projectname, char* folder,climblayer *c){
    if(c != NULL){
        QString filename = QString(folder).append(projectname).append(".bbl");
        FILE* f = fopen(filename.toLocal8Bit().data(),"wb");
        if(f != NULL){
            fwrite(&(c->climbcount),sizeof(c->climbcount),1,f);
            for(long i = 0; i < c->climbcount; i++){
                fwrite(&(c->climb[i].d		 ),sizeof(c->climb[i].d		  ),1,f);
                fwrite(&(c->climb[i].dhcount ),sizeof(c->climb[i].dhcount ),1,f);
                fwrite(&(c->climb[i].hend	 ),sizeof(c->climb[i].hend	  ),1,f);
                fwrite(&(c->climb[i].hmdown	 ),sizeof(c->climb[i].hmdown	  ),1,f);
                fwrite(&(c->climb[i].hmup	 ),sizeof(c->climb[i].hmup	  ),1,f);
                fwrite(&(c->climb[i].hstart	 ),sizeof(c->climb[i].hstart	  ),1,f);
                fwrite(&(c->climb[i].KOOcount),sizeof(c->climb[i].KOOcount),1,f);
                fwrite(&(c->climb[i].P		 ),sizeof(c->climb[i].P		  ),1,f);
                fwrite(&(c->climb[i].slope	 ),sizeof(c->climb[i].slope	  ),1,f);
                fwrite(&(c->climb[i].bend	 ),sizeof(c->climb[i].bend	  ),1,f);
                fwrite(&(c->climb[i].midlat	 ),sizeof(c->climb[i].midlat  ),1,f);
                fwrite(&(c->climb[i].midlon	 ),sizeof(c->climb[i].midlon  ),1,f);
                fwrite(&(c->climb[i].LocRank ),sizeof(c->climb[i].LocRank ),1,f);
                //if(c->climb[i].dhcount == 0 || i == 2390)
                //    QMessageBox().exec();
                fwrite(c->climb[i].dharray[0],sizeof(double),c->climb[i].dhcount,f);
                fwrite(c->climb[i].dharray[1],sizeof(double),c->climb[i].dhcount,f);
                fwrite(c->climb[i].KOOarray[0],sizeof(double),c->climb[i].KOOcount,f);
                fwrite(c->climb[i].KOOarray[1],sizeof(double),c->climb[i].KOOcount,f);

                int length = 0;
                if(c->climb[i].name != NULL)
                   length = strlen(c->climb[i].name);
                fwrite(&length,sizeof(int),1,f);
                if(length > 0)
                    fwrite(&(c->climb[i].name[0]),sizeof(char),length,f);
            }
            fclose(f);
        }
    }
}
void ClimbAnalyse::create_MAP(char* projectname, char* folder, HeightData* h,
                              double size){
	if(climbcount > 0){
		double minlat=90,maxlat=-90,minlon=180,maxlon=-180;
		for(long i = climbcount-1; i >= 0; i--){
			if(climbpool[i].P >= minP){
				climbpool[i].t->adapt_brect(&minlat,&maxlat,&minlon,&maxlon);
			}
		}
        MAP* mt = new MAP(h,o);
        QString filename = QString(folder).append(projectname).append("climblist.txt");
        mt->save_to_png(filename.toLocal8Bit().data());
		mt->~MAP();
	}
}					
void ClimbAnalyse::create_html(char* projectname, char* folder){
    if(climbcount > 0){
        //QMessageBox(QMessageBox::NoIcon,"",QString("%1").arg(folder)).exec();
        plot2D p;
        p.set_size(450,300);
        QString path0 = QString(folder).append("/").append(projectname).append("-%1.png");
        QString path1 = QString(folder).append("/").append(projectname).append("-BinaryList.src");
        QString path2 = QString(folder).append("/").append(projectname).append("-%1.trk");
        QString path4 = QString(folder).append("/").append(projectname).append("-names.txt");
        ushort usb;
		short sb;
		float fb;
		KOO kt;
        char newline = '\n';
		unsigned int realclimbs =0;
        for(int i = 0; i < climbcount; i++){
            if(climbpool[i].P >= minP){
                climbpool[i].t->extract_KOO(&kt,climbpool[i].c->d / 2.);
                if(o->isin(&kt,minlat,maxlat,minlon,maxlon))
                    realclimbs++;
            }
        }
        if(realclimbs>0){
            FILE* f;
            f = fopen(path1.toLocal8Bit().data(),"wb");
            FILE* fnames;
            fnames = fopen(path4.toLocal8Bit().data(),"w");
            fwrite(&realclimbs,4,1,f);
            int realindex = -1, length;
            for(int i = 0; i < climbcount; i++)
                if(climbpool[i].P >= minP){
                    climbpool[i].t->extract_KOO(&kt,climbpool[i].c->d / 2.);
                    if(o->isin(&kt,minlat,maxlat,minlon,maxlon)){
                        /*Binaere Daten schreiben*/
                        usb = ushort(climbpool[i].c->d*1000);
                        fwrite(&usb,2,1,f);
                        sb = short(climbpool[i].c->hstart*10);
                        fwrite(&sb,2,1,f);
                        sb = short(climbpool[i].c->hend*10);
                        fwrite(&sb,2,1,f);
                        sb = short(climbpool[i].c->hmdown*10);
                        fwrite(&sb,2,1,f);
                        sb = short(climbpool[i].t->get_bend()*100);
                        fwrite(&sb,2,1,f);
                        fb = float(kt.lon);
                        fwrite(&fb,4,1,f);
                        fb = float(kt.lat);
                        fwrite(&fb,4,1,f);
                        /*png-Datei erzeugen*/
                        realindex++;
                        p.set_data(climbpool[i].t->get_d_h_array(),
                                    climbpool[i].t->get_d_h_length());
                        QString filename = QString(path0).arg(realindex);
                        p.save_to_png(filename);
                        /*track erzeugen*/
                        filename = QString(path2).arg(realindex);
                        climbpool[i].t->to_binary_KOO(filename.toLocal8Bit()
                                                      .data());
                        /*write names file*/
                        length = strlen(climbpool[i].name);
                        if(length > 0)
                            fwrite(&(climbpool[i].name[0]),sizeof(char),
                                    length,fnames);
                        fwrite(&newline,sizeof(char),1,fnames);
                    }
                }
            fclose(f);
            fclose(fnames);
        }
	}
}
void ClimbAnalyse::create_html_singlefile(char* projectname, char* folder){
    if(climbcount > 0){
        //QMessageBox(QMessageBox::NoIcon,"",QString("%1").arg(folder)).exec();
        plot2D p;
        p.set_size(450,300);
        QString path0 = QString(folder).append("/").append(projectname).append("-KOO_Heights.dat");
        QString path1 = QString(folder).append("/").append(projectname).append("-BinaryList.src");
        QString path4 = QString(folder).append("/").append(projectname).append("-names.txt");
        ushort usb;
        short sb;
        float fb;
        KOO kt;
        char newline = '\n';
        unsigned int realclimbs =0;
        unsigned int *KOOcount = new unsigned int[climbcount+1];
        unsigned int *Heightscount = new unsigned int[climbcount+1];
        KOOcount[realclimbs]     = 0;
        Heightscount[realclimbs] = 0;
        // get data overview
        for(int i = 0; i < climbcount; i++){
            if(climbpool[i].P >= minP){
                climbpool[i].t->extract_KOO(&kt,climbpool[i].c->d / 2.);
                if(o->isin(&kt,minlat,maxlat,minlon,maxlon)){
                    realclimbs++;
                    KOOcount[realclimbs]     = climbpool[i].t->get_KOOa_length();
                    Heightscount[realclimbs] = climbpool[i].t->get_d_h_length();
                }
            }
        }
        if(realclimbs>0){
            FILE* f;
            f = fopen(path1.toLocal8Bit().data(),"wb");
            FILE* fnames;
            fnames = fopen(path4.toLocal8Bit().data(),"w");
            FILE* fdat;
            fdat = fopen(path0.toLocal8Bit().data(),"wb");
            // right binary list header
            fwrite(&realclimbs,4,1,f);
            // right data file header
            fwrite(&realclimbs,4,1,fdat);
            unsigned int ci = 4 + realclimbs*4*2;  // climbcount + register length
            for(int i = 0; i < realclimbs; i++){
               fwrite(&ci,4,1,fdat);
               ci += 4 + KOOcount[i+1]*2*4;
               fwrite(&ci,4,1,fdat);
               ci += 4 + Heightscount[i+1]*2*4;
            }
            //write data
            int realindex = 0, length;
            for(int i = 0; i < climbcount; i++)
                if(climbpool[i].P >= minP){
                    climbpool[i].t->extract_KOO(&kt,climbpool[i].c->d / 2.);
                    if(o->isin(&kt,minlat,maxlat,minlon,maxlon)){
                        //Binaere Daten schreiben
                        usb = ushort(climbpool[i].c->d*1000);
                        fwrite(&usb,2,1,f);
                        sb = ushort(climbpool[i].c->hstart*10);
                        fwrite(&sb,2,1,f);
                        sb = ushort(climbpool[i].c->hend*10);
                        fwrite(&sb,2,1,f);
                        sb = ushort(climbpool[i].c->hmdown*10);
                        fwrite(&sb,2,1,f);
                        sb = ushort(climbpool[i].t->get_bend()*100);
                        fwrite(&sb,2,1,f);
                        fb = float(kt.lon);
                        fwrite(&fb,4,1,f);
                        fb = float(kt.lat);
                        fwrite(&fb,4,1,f);
                        // generate track/height data
                        realindex++;
                        trackpoint* tpl;
                        KOOlist* kl;
                        climbpool[i].t->get_data_ref(&tpl, NULL, &kl, NULL);
                        float floatbuf;
                        fwrite(&(KOOcount[realindex]),4,1,fdat);
                        for(int j = 0; j < KOOcount[realindex]; j++){
                            floatbuf = float(kl[j].k->lat);
                            fwrite(&(floatbuf),sizeof(float),1,fdat);
                            floatbuf = float(kl[j].k->lon);
                            fwrite(&(floatbuf),sizeof(float),1,fdat);
                        }
                        fwrite(&(Heightscount[realindex]),4,1,fdat);
                        for(int j = 0; j < Heightscount[realindex]; j++){
                            floatbuf = float(tpl[j].d);
                            fwrite(&(floatbuf),sizeof(float),1,fdat);
                            floatbuf = float(tpl[j].h);
                            fwrite(&(floatbuf),sizeof(float),1,fdat);
                        }
                        //write names file
                        length = strlen(climbpool[i].name);
                        if(length > 0)
                            fwrite(&(climbpool[i].name[0]),sizeof(char),
                                    length,fnames);
                        fwrite(&newline,sizeof(char),1,fnames);
                    }
                }
            fclose(f);
            fclose(fnames);
            fclose(fdat);
        }
        delete KOOcount;
        delete Heightscount;
    }
}
void ClimbAnalyse::create_list(char* projectname, char* folder){
	if(climbcount > 0){
        QString filename = QString(folder).append("\\").append(projectname).append("climblist.txt");
        FILE* f = fopen(filename.toLocal8Bit().data(),"w");
        for(long i = climbcount; i >= 0; i--){
			if(climbpool[i].P > minP)
                fprintf(f,"%5.1f P\t %4.1f %% \t %.2f km\t %6.1f m\t %6.1f m\t %6.1f m\t %6.1f m\t %6.1f serpentine/km\t%s\t Start-id: %12lli\t End-id:%12lli \n",
						climbpool[i].P,
						climbpool[i].c->hmup/climbpool[i].c->d/10,
						climbpool[i].c->d,
						climbpool[i].c->hmup,
						climbpool[i].c->hmdown,
						climbpool[i].c->hstart,
						climbpool[i].c->hend,
						climbpool[i].t->get_bend()/climbpool[i].c->d,
                        climbpool[i].name,
						climbpool[i].c->cst->home->id,
						climbpool[i].c->home->id);
		}
		fclose(f);
	}
}
short ClimbAnalyse::looses(climb* resident, climb* aspirant){
    if(resident->prev->home == aspirant->prev->home){
        // climbs do not only meet each other, but already share a road
		if( (resident->cst == aspirant->cst && resident->t->t.d <= aspirant->t->t.d) ||
            (aspirant->d < resident->d &&
			 resident->P > aspirant->P)
			)
			return 1;
		else
			return 0;
	}
	else
		return 0;
}
nameaspirant::nameaspirant(char* name){
    this->name = QString().fromUtf8(name);
    d = e = s = 0;
}
nameaspirant::nameaspirant(char* name, float d){
    this->name = QString().fromUtf8(name);
    this->d = d;
    e = s = 0;
}
bool nameaspirant::operator ==(nameaspirant o){
    if(this->name == o.name)
        return true;
    else
        return false;
}
void ClimbAnalyse::create_names(){
    for(long i = climbcount-1; i >= 0; i--){
        if(climbpool[i].P >= minP){
            // determine street names
            QList<nameaspirant> wn;
            eval_Way_names(&wn,climbpool[i].c);
            // determine point names
            QList<nameaspirant> pn;
            eval_point_names(&pn,o->pt[2],1.0,climbpool[i].t,true);
            eval_point_names(&pn,o->pt[3],1.0,climbpool[i].t,true);
            eval_point_names(&pn,o->pt[4],1.0,climbpool[i].t,true);
            eval_point_names(&pn,o->pt[Coli],1.2,climbpool[i].t,true);
            eval_point_names(&pn,o->pt[Peaki],-1,climbpool[i].t,false);
            QString name;
            bool foundname = 0;
            // 1. check end point
            double max = 0;
            int maxi = -1;
            for(int i = 0; i < pn.length(); i++)
                if(max < (pn[i].e-pn[i].s)*pn[i].d > max){
                    max = (pn[i].e-pn[i].s)*pn[i].d;
                    maxi = i;
                }
            if(max>0.125 && !pn[maxi].name.isNull()){
                foundname = 1;
                name.append(pn[maxi].name);
                // 2. check start point
                max = 0;
                maxi = -1;
                for(int i = 0; i < pn.length(); i++)
                    if(max < (-pn[i].e+pn[i].s)*pn[i].d > max){
                        max = (-pn[i].e+pn[i].s)*pn[i].d;
                        maxi = i;
                    }
                if(max>0.125 && !pn[maxi].name.isNull())
                    name.append(" (").append(pn[maxi].name).append(")");
                }
            // 3. check way name
            QString wname;
            qSort(wn.begin(),wn.end());
            if(!wn[0].name.isEmpty()){
                if(wn[0].d > 0.5)
                    wname = wn[0].name;
                else
                    if(wn.length()>1 && !wn[1].name.isEmpty()
                       && wn[0].d + wn[1].d > 0.66)
                        wname.append(wn[0].name)
                             .append(", ")
                             .append(wn[1].name);
            }
            if(!wname.isEmpty()){
                if(foundname)
                    name.append(", ").append(wname);
                else
                    name = wname;
            }
            // 4. if way name at maximum, add closest point
            if(!foundname && pn.length() > 0){
                qSort(pn.begin(),pn.end());
                if(!pn[0].name.isEmpty() && pn[0].d > 0.5)
                    name.append(" [").append(pn[0].name).append("]");
            }
            // copy to data
            if(!name.isEmpty()){
                climbpool[i].name = new char[strlen(name.toUtf8().data())+1];
                strcpy(climbpool[i].name,name.toUtf8().data());
            }
            else{
                QString none = "none";
                climbpool[i].name = new char[strlen(name.toUtf8().data())+1];
                strcpy(climbpool[i].name,name.toUtf8().data());
            }
        }
        else
            climbpool[i].name = NULL;

    }
}
void ClimbAnalyse::eval_Way_names(QList<nameaspirant> *wn, climb* c0){
    climb* c = c0, *clast;
    while(c != c0->cst){
        int i = wn->indexOf(c->via->way->name);
        if( i > -1) (*wn)[i].d += c->via->d;
        else        wn->append(nameaspirant(c->via->way->name,
                                           c->via->d));
        clast = c;
        c = c->prev;
    }
    int is = wn->indexOf(c0->via->way->name);
    int ie = wn->indexOf(clast->via->way->name);
    (*wn)[is].d -= c0->t  ->t.d - c0->tend  ->t.d;
    if(c0->cst->t != NULL)
        (*wn)[ie].d -= c0->tst->t.d - c0->cst->t->t.d;
    else
        (*wn)[ie].d -= c0->tst->t.d;
    for(int i = 0; i < wn->length(); i++)
        (*wn)[i].d = (*wn)[i].d / c0->d;
}
void ClimbAnalyse::eval_point_names(QList<nameaspirant> *pn, points **p,
                                    double fac, Track* t, bool validstart){
    bool calcfromindex = fac < 0;
    points* pts = *p;
    double d,h0;
    KOO ks = t->get_firstKOO();
    KOO ke = t->get_lastKOO();
    KOO ktt;
    float hs = h->height(ks.lon,ks.lat,NULL,0);
    float he = h->height(ke.lon,ke.lat,NULL,0);
    float dlimit = distance(ks.lat,ks.lon,ke.lat,ke.lon);
    float dh = he-hs;
    double latmin,latmax,lonmin,lonmax;
    t->get_brect(&latmin,&latmax,&lonmin,&lonmax);
    double dd = sqrt((latmax-latmin)*(latmax-latmin)+
                     (lonmax-lonmin)*(lonmax-lonmin));
    latmin -= dd;
    latmax += dd;
    lonmin -= dd;
    lonmax += dd;
    while(pts != NULL){
        if(o->isin(pts->k,latmin,latmax,lonmin,lonmax)){
            if(calcfromindex) fac = 1.1*(1-exp(-sqrt(pts->index/800)));
            t->next_d(pts->k->lat,pts->k->lon,&d);
            t->extract_KOO(&ktt,d);
            d = distance(ktt.lat,ktt.lon,pts->k->lat,pts->k->lon);
            h0 = h->height(pts->k->lon,pts->k->lat,NULL,0);
            if(d<dlimit){
                pn->append(nameaspirant(pts->name,(1-d/dlimit)*fac));
                pn->last().e = qMax(0.,
                                    (1-distance(ke.lat,ke.lon,pts->k->lat,pts->k->lon)
                                    /dlimit)*(1-qAbs(h0-he)/dh));
                if(validstart)
                    pn->last().s = qMax(0.,
                                        (1-distance(ks.lat,ks.lon,pts->k->lat,pts->k->lon)
                                         /dlimit)*(1-qAbs(hs-h0)/dh));
                else
                    pn->last().s = 0;
            }
        }
        pts = pts->next;
    }
}
void ClimbAnalyse::set_brect(double minlat, double maxlat,
                             double minlon, double maxlon){
    this->minlat = minlat;
    this->maxlat = maxlat;
    this->minlon = minlon;
    this->maxlon = maxlon;
}
