#pragma once
#include "HeightData.h"
#include "MathExt.h"
#include <QFileInfo>
#include <QMessageBox>
#include <QTime>

typedef struct summitname{
	char* name;
	double lon,lat,distance;
	double height;
	summitname* next;
}summitname;

typedef struct summit{
    double height;
	double prominence;
    double lon, lat;
    double index,Indexprominence,OldIndex,AvgIndex;
    double AnglePmax[15];
	summit* parent;
	summitname* names;
}summit;
typedef struct summitlist{
	summit* data;
	summitlist* next;
}summitlist;

class SummitRanking:public QObject{
    Q_OBJECT
public:
    SummitRanking();
    SummitRanking(QString worldfile, QString destfolder);
    SummitRanking(int lat, int lon, QString srtm3folder,
                  QString destfolder);
    void CalcSummits(int lat, int lon, QString srtm3folder,
                     QString destfolder);
	SummitRanking(char* SRTM3path, char* summitlistfile,
				  double minlon, double maxlon, double minlat, double maxlat);
	~SummitRanking();
signals:
    void log(QString logfile,QString message);
private:
	void prepare(char* SRTM3path,
				 double minlon, double maxlon, double minlat, double maxlat);
    void unprepare();
	void calc();
	void load_names(char* file);
	void print();
	int trace_uphill(double* lon, double* lat);
	void shorten_list();
	void init();
	summit* create_summit();
	void sort_summit_list(long s, long e);
	double hgp(summit* child);			//return height of the greatest parent
	void inherit_parent(summit* taker, summit* giver, double topoprom);
	recti GridRect;
	int cx,cy;							//data grid count
	long cs;							//total summit count
	short* hs;							//height data sorted
    int *lons,*lats;					//reference to grid;
    int *gradx,*grady;                  //reference to height grid;
	long count;							//number of height data points
	summit*** s, **qsh;					//summit grid and collection array
	summitlist* ssorted;
	HeightData h;
    QMutex fmutex;

};
