﻿#ifndef BTP3DIALOGS_H
#define BTP3DIALOGS_H
#include "DataTyps.h"
#include <QtWidgets/QLabel>
#include <QtWidgets/QLineEdit>
#include <QtWidgets/QPushButton>
#include <QtWidgets/QCheckBox>
#include <QtWidgets/QGridLayout>
#include <QtWidgets/QMessageBox>
#include <QDir>
#include <QtWidgets/QFileDialog>
#include <QtWidgets/QColorDialog>
#include <QScrollArea>
#include <QPalette>
#include <QPainter>
#include "MathExt.h"

//! Button to display color and open color dialog
class ColorButton: public QPushButton{
    Q_OBJECT
public:
    ColorButton(QRgb* color);
    void paintEvent(QPaintEvent *);
private slots:
    void edit_color();
private:
    QRgb* color;
};
//! generic class to provide editing of specific types
class EditInputs: public QWidget{
    Q_OBJECT
public:
    EditInputs(QString name, double* v);
    EditInputs(QString name, float* v);
    EditInputs(QString name, int* v);
    EditInputs(QString name, char* v, bool folder);
    EditInputs(QString name, bool* v);
    EditInputs(QString name, short* v);
    EditInputs(QString name, QRgb* v);
    void refresh();
private slots:
    void checkedit(QString newtext);
    void checkbox();
    void select_folder();
private:
    QLabel label;
    QLineEdit edit;
    ColorButton *colorbutton;
    QCheckBox check;
    double* dv;
    float* fv;
    int* iv;
    char* cv;
    bool* bv;
    short* sv;
    QRgb* rgbv;
    char* folder;
};

//! provide all user settings
/*!
  all user setting is stored in a config struct, consisting of
    MapSettings
    STRONGsetting
    ClimbAnalyseSettings
    ProfilSettings
    PowerAnalysisSetting
  to edit these structs they are opened with the suitable constructor, an can
  be initialized with init() command
*/
class BTP3setups: public QDialog{
    Q_OBJECT
public:
    BTP3setups(STRONGsetting* ss, STRONGsetting* ms, STRONGsetting* cs);
    BTP3setups(ClimbAnalyseSettings* CA);
    BTP3setups(ExtMapSettings* ems);
    BTP3setups(ProfilSettings* PS);
    BTP3setups(MapSettings* MS);
    void init();
private:
    void set_NULL();
    void setLayouts();
    //! buffer struct to store user edits
    STRONGsetting *ss, *ms, *cs, ss0, ms0, cs0;
    //! buffer struct to store user edits
    ClimbAnalyseSettings* CA, CA0;
    //! buffer struct to store user edits
    ExtMapSettings* ems, ems0;
    //! buffer struct to store user edits
    ProfilSettings* PS, PS0;
    //! buffer struct to store user edits
    MapSettings* MS, MS0;
    //! buffer struct to store user edits
    QPushButton buttons[4];
    EditInputs* edit[86];
    QGridLayout* L;
public slots:
    void OK();                          /*!< copies buffer to real struct*/
    void init_buffer();                 /*!< sets buffer default */
    void init_dest();                   /*!< sets real struct default*/
    void cancel();                      /*!< close without changes*/
};

#endif // BTP3DIALOGS_H
