﻿#include "gpx.h"


gpx::gpx(char* filename){
	file = filename;
    char* extension = &(filename[strlen(filename)-4]);
	if(file != NULL){
        if(!strcmp(extension,".gpx")) from_gpx();
        if(!strcmp(extension,".fit")) from_fit();
		calc_distance();
        remove_stops_teleportations();
        relax_heights(4);
        correct_bpm();
        determin_typ();
	}
}
gpx::~gpx(){
}
void gpx::add_point(double lon, double lat, double h, long int time){
    add_point(lon,lat,h,time,0,0.);
}
void gpx::add_point(double lon, double lat, double h, long int time,
                    double bpm, double W){
    if(lon < -180 || lon > 180 || lat < -90 || lat > 90 ||
       h < -500 || h > 10000)
        return;                         // invalid data
    l.append(gpxlist(0,h,lat,lon,time,bpm,W));
}
void gpx::from_fit(){
     FILE *f;
     FIT_UINT8 buf[8];
     FIT_CONVERT_RETURN convert_return = FIT_CONVERT_CONTINUE;
     FIT_UINT32 buf_size;
     FIT_UINT32 mesg_index = 0;
     #if defined(FIT_CONVERT_MULTI_THREAD)
        FIT_CONVERT_STATE state;
     #endif
     #if defined(FIT_CONVERT_MULTI_THREAD)
        FitConvert_Init(&state, FIT_TRUE);
     #else
        FitConvert_Init(FIT_TRUE);
     #endif

     if((f = fopen(file, "rb")) == NULL)
     {
        return;
     }

     while(!feof(f) && (convert_return == FIT_CONVERT_CONTINUE)){
        for(buf_size=0;(buf_size < sizeof(buf)) && !feof(f); buf_size++)
           buf[buf_size] = (FIT_UINT8)getc(f);

        do{
           #if defined(FIT_CONVERT_MULTI_THREAD)
              convert_return = FitConvert_Read(&state, buf, buf_size);
           #else
              convert_return = FitConvert_Read(buf, buf_size);
           #endif

           switch (convert_return){
              case FIT_CONVERT_MESSAGE_AVAILABLE:{
                 #if defined(FIT_CONVERT_MULTI_THREAD)
                    const FIT_UINT8 *mesg = FitConvert_GetMessageData(&state);
                    FIT_UINT16 mesg_num = FitConvert_GetMessageNumber(&state);
                 #else
                    const FIT_UINT8 *mesg = FitConvert_GetMessageData();
                    FIT_UINT16 mesg_num = FitConvert_GetMessageNumber();
                 #endif


                if(mesg_num == FIT_MESG_NUM_RECORD){
                    const FIT_RECORD_MESG *record = (FIT_RECORD_MESG *) mesg;
                    add_point(
                        double(record->position_long)*8.38190317153931E-008,
                        double(record->position_lat)*8.38190317153931E-008,
                        double(record->altitude)/5-500,
                        long(record->timestamp)+599616000,
                        double(record->heart_rate),
                        double(record->power));
                }
             }
             default:
                break;
          }
       } while (convert_return == FIT_CONVERT_MESSAGE_AVAILABLE);
    }

     fclose(f);
}
void gpx::from_gpx(){
	FILE* f = fopen(file,"r");
	if(f != NULL){
		/*Datei existiert*/
		char line[10000],*readstart,*newstart;
        double lon = 0, lat = 0, h = 0, bpm = 0, W = 0;
		long int time;
		int read_out;
		while(!feof(f)){
			/*Zeilenweise einlesen*/
			fgets(line,10000,f);
			replace_quotation(line);
			read_out = 0;
			readstart = line;
			while(!read_out){
				newstart = strstr(readstart,"</trkpt>");
                if(newstart == NULL)
                    newstart = strstr(readstart,"</rtept>");
				if (newstart != NULL){
					newstart[0] = '\0';
					newstart = &newstart[1];
				}
				else{
					read_out = 1;
				}
				extract_attribute("lat",readstart,&lat);
				extract_attribute("lon",readstart,&lon);	
				extract_element("ele",readstart,&h);
                extract_element("power",readstart,&W);
                extract_element("gpxtpx:hr",readstart,&bpm);
				extract_time(readstart,&time);

				if(!read_out){
                    add_point(lon,lat,h,time,bpm,W);
					readstart = newstart;
				}
			}
		}
	}
}
void gpx::replace_quotation(char* a){
	int i = 0;
	while(a[i] != '\0'){
		if(a[i] == '\"')
			a[i] = '\'';
		i++;
	}
}
void gpx::extract_attribute(const char* attr, char* src, double* result){
	char *x1;
	char *x2;
	x1 = strstr(src, attr);
	if(x1 != NULL){
		x1 = &x1[strlen(attr)+2*sizeof(char)];
		x2 = strchr(x1,'\'');
        if(x2 != NULL){
            x2[0] = '\0';
            *result = atof(x1);
            x2[0] = '\'';
        }
	}
}
void gpx::extract_element(const char* elem, char* src, double* result){
	char *x1;
	char *x2;
	char pattern[100];
    strcpy(&pattern[2],elem);
    pattern[1]='<';
    pattern[2+strlen(elem)] = '>';
    pattern[3+strlen(elem)] = '\0';
    x1 = strstr(src,&pattern[1]);
	if(x1 != NULL){
        x1 = &x1[strlen(&(pattern[1]))];
		pattern[1] = '/';
		pattern[0] = '<';
		x2 = strstr(src,pattern);
		x2[0] = '\0';
		*result = atof(x1);
		x2[0] = '<';
	}
}
void gpx::extract_time(char* src, long int* unix_time){
	char *x1,*x2;
	char pattern[] = "</time>";
	pattern[1]='<';
	x1 = strstr(src,&pattern[1]);
	if(x1 != NULL){
		x1 = &x1[strlen(pattern)-1];
		pattern[1] = '/';
		x2 = strstr(src,pattern);
		x2[0] = '\0';
		int totaldays,year,month,day,hour,minute,second;
		char *x3 = strstr(x1,"T");
		if(x3 != NULL){
			x1 = x1;		x1[4]='\0';		year = atoi(x1);	x1[4]='-';
			x1 = &x1[5];	x1[2]='\0';		month = atoi(x1);	x1[2]='-';
			x1 = &x1[3];	x1[2]='\0';		day = atoi(x1);		x1[2]='T';
			x1 = &x3[1];	x1[2]='\0';		hour = atoi(x1);	x1[2]=':';
			x1 = &x1[3];	x1[2]='\0';		minute = atoi(x1);	x1[2]=':';
			x1 = &x1[3];	x1[2]='\0';		second = atoi(x1);	x1[2]='Z';
			
			int is_leap_year = 0;
			if( year % 4 == 0 
				&& (((year-1901) % 100 != 0) || ((year-1901) % 400 == 0))) 
				is_leap_year = 1;
			
			/*Tage seit 1970 inkl. Schalt- und Saekularisierungsjahr*/
			totaldays = (year-1970)*365;				//normale Tagesanzahl
			totaldays = totaldays + (year-1969) / 4;	//Schaltjahr normal
			totaldays = totaldays - (year-1901) / 100;	//Seakularisierung 1
			totaldays = totaldays + (year-1601) / 400;	//Seakularisierung 2

			/*Aufschlüsselung der Monatstage*/
			switch(month){
				case 12: totaldays = totaldays + 30;
				case 11: totaldays = totaldays + 31;
				case 10: totaldays = totaldays + 30;
				case  9: totaldays = totaldays + 31;
				case  8: totaldays = totaldays + 31;
				case  7: totaldays = totaldays + 30;
				case  6: totaldays = totaldays + 31;
				case  5: totaldays = totaldays + 30;
				case  4: totaldays = totaldays + 31;
				case  3: totaldays = totaldays + 28 + is_leap_year;
				case  2: totaldays = totaldays + 31;
			}

			*unix_time =  24*60*60 * (totaldays + (day -1))
						+    60*60 * hour
						+       60 * minute 
						+            second;
		}
		x2[0] = '<';
	}
}
//void* gpx::get_complete_List(){
//	return l;
//}
void gpx::calc_distance(){
    for(int i = 1; i < l.length(); i++)
        l[i].add_distance_to(l[i-1]);
}
gpxlist::gpxlist(double d, double h, double lat, double lon, double time){
    this->d = d;
    this->h = h;
    this->lat = lat;
    this->lon = lon;
    this->time = time;
    this->bpm = 0;
    this->W = 0;
}
gpxlist::gpxlist(double d, double h, double lat, double lon, double time,
                 double bpm, double Ws){
    this->d = d;
    this->h = h;
    this->lat = lat;
    this->lon = lon;
    this->time = time;
    this->bpm = bpm;
    if(Ws < 3000)
        this->W = Ws;
    else
        this->W = 0;
}
void gpxlist::add_distance_to(gpxlist nb){
    double lat1 = this->lat;
    double lon1 = this->lon;
    double lat2 = nb.lat;
    double lon2 = nb.lon;
    double check = cos(lat2/180*M_PI)*cos(lat1/180*M_PI)*
            (cos(lon2/180*M_PI-(lon1)/180*M_PI))+sin(lat2/180*M_PI)
            *sin((lat1)/180*M_PI);
    this->d = nb.d + acos(qMin(1.0,check))*R0;
}
double** gpx::get_complete(long int* n){
    *n = long int(l.length());
	double** result;
    if(l.length()>1){
        result = new double*[7];
        for(int i = 0; i < 7; i++)
            result[i] = new double[l.length()];
        for(int i = 0; i < l.length(); i++){
            result[0][i] = l[i].d;
            result[1][i] = l[i].h;
            result[2][i] = l[i].lat;
            result[3][i] = l[i].lon;
            result[4][i] = l[i].time;
            result[5][i] = l[i].bpm;
            result[6][i] = l[i].W;
		}
	}
	else
		result = NULL;

	return result;
}
void gpx::remove_stops_teleportations(){
    // calculate avg speed > 1 m/s && < 25 m/s without position jumps
    double dr = 0;
    long int tr = 0;
    double dd = 0;
    double dt = 0;
    double dh = 0;
    double vavg;
    for(int i = 1; i < l.length(); i++)
        if(v(i) > 1 && v(i) < 20 && l[i].d - l[i-1].d < 0.1){
            dr += l[i].d - l[i-1].d;
            tr += l[i].time - l[i-1].time;
        }
    if(tr > 0)  vavg = dr/tr;
    else        return;
    // remove stops
    for(int i = 1; i < l.length(); i++){
        l[i].h -= dh;
        l[i].d -= dd;                   // include previous knowledge
        l[i].time -= dt;
        if(v(i) < 1){
            // position jump or stop found
            dt += l[i].time-l[i-1].time;
            dd += l[i].d-l[i-1].d;
            dh += l[i].h-l[i-1].h;
            l.removeAt(i);
            i--;
        }
    }
    // remove teleportations
    dd = 0;
    dt = 0;
    dh = 0;
    for(int i = 1; i < l.length(); i++){
        l[i].d -= dd;                   // include previous knowledge
        l[i].time -= dt;
        l[i].h -= dh;
        if(l[i].d-l[i-1].d > 0.1){
            // position jump or stop found
            dt += l[i].time-l[i-1].time -1;
            dd += l[i].d-l[i-1].d       -1.*vavg;
            dh += l[i].h-l[i-1].h;
            l[i].time = l[i-1].time + 1;
            l[i].d    = l[i-1].d    + 1.*vavg;
            l[i].h    = l[i-1].h;
        }
    }
}
void gpx::relax_heights(int repetitions){
    for(int k = 0; k < repetitions; k++){
        for(int i = 1; i < l.length()-1; i++){
            double dh1 = (l[i].h - l[i-1].h)/3.;
            double dh2 = (l[i-1].h - l[i].h)/3.;
            l[i-1].h    += dh1;
            l[i].h      -= dh1;
            l[i].h      += dh2;
            l[i+1].h    -= dh2;
        }
    }
}
double gpx::v(int i){
    if(i < l.length()){
        double dt = l[i].time-l[i-1].time;
        if(dt > 0)
            return (l[i].d-l[i-1].d)*1000. / dt;
        else
            return 0;
    }
    else
        return 0;
}
void gpx::determin_typ(){
    double t12 = 0.001, t20 = 0.001, t25 = 0.001, txx = 0.001;
    bool Power_available = false;
    for(int i = 1+10; i < l.length()-10; i++){
        if(l[i].W > 0)  Power_available = true;
        double v = (l[i+10].d - l[i-10].d)*1000
                 / qMax(long(1),(l[i+10].time - l[i-10].time))
                 * 3.6;
        if(v < 12)
            t12++;
        else
            if(v < 20)
                t20++;
            else
                if(v < 25)
                        t25++;
                else
                    if(v > 25)
                        txx++;
    }
    if(Power_available == true || (txx+t25)/t20 > 0.5)
        typ = 1;
    else{
        if((t25+txx)/t12 > 0.03)
            typ = 2;
        else
            typ = 3;
    }
}
void gpx::correct_bpm(){
    for(int i = 0; i < l.length(); i++){
        if(l[i].bpm > 195)l[i].bpm = 195;
        if(l[i].bpm < 75)l[i].bpm = 75;
    }
}
int gpx::get_typ(){
    return typ;
}
