﻿/*! \mainpage Manual and Source Code Documentation
 *
 * \section Introduction
 *
 * This is BergTourenPlaner 3, an offline map resource, routing and
 * analysing system written for passionated climbers on bike. The Program is
 * based on OpenStreetMap (%OSM) data and the Digital Elevation Model (DEM)
 * %SRTM3 of the NASA.
 *
 * Beside the ordenary shortes path Dijkstra routing algorithm  BTP3 provides
 * routing routines to route paths with least altitude gain (i.g. least vertical
 * or height meters (Hm)) as well as paths with maximized Hm, in the following
 * called %STRONG algorithm, with the associated Track called %STRONG.
 * Second important thing is Power analysis to calculate average power output
 * at user defined rider profil and external conditions. This also works for
 * gpx tracks with existing time stemps.
 *
 * Last important thing is climb analyis, investigating the whole street
 * network for user defined climbs. Consequently maps with the calculated
 * clibms can be calculated.
 *
 * This is open source project, feel free to try out and enrich the code,
 * written for Qt 5.1 and tested so far at XP, Win7 and Win8 machines. For
 * beginning consider the following important classes:
 *
 * MainWindow, MAP, Routing, STRONGcalc, PowerAnalysis, Track and WayHeight
 *
 * Please report bugs and wishes to degga87[ at ]web.de.
 *
 * \section sec1 Getting Started
 *
 * \subsection Installation
 *
 * You just have to copy the complete folder anywhere on your computer, no
 * additional action is necessary.
 *
 * \subsection secsrtm3 SRTM3 DEM
 *
 * %SRTM3 data is necessary to render terrain and calculate street data from %OSM
 * data. %SRTM3 consists of hgt files representing the while terrain within
 * 1°x1°-area. Best source to get data is http://www.viewfinderpanoramas.org/Coverage%20map%20viewfinderpanoramas_org3.htm.
 * Download desired area files into a folder and reference it with
 * settings->%SRTM3 data folder. The map window will be instantly rerendered.
 * Use DataProcessing->correct hgt-files to get rid of voids.
 *
 * \subsection secosmdata OSM map data
 *
 * BTP3 works with btp files which are binary files generated from %OSM data.
 * I recommend downloading the desired data area from
 * http://download.geofabrik.de/ , cutting the area with osmconvert to an
 * *.osm file, which can be read with BTP3 via data->read osm. %SRTM3 data
 * should be present at the process, the process may take up to a hour. Maximum
 * *.osm file size is 2 GB for 32 bit Windows, otherwise unlimited. Save the
 * generate data via file->save btp. Please note that only the data viewed in
 * the map window will be save. You might use this to cut your btp file into
 * smaller pieces. IF no cutting is undesired you might zoom out sufficiently.
 *
 * You can instruct BTP3 to load a btp file at program start with
 * settings->btp startup file, otherwise data->load btp to load data.
 *
 * \subsubsection Editing
 * Editing btp data is possible by pressing e. Select a way and assign new type
 * with number keys od delete it.
 *
 * \section secrotuing Routing
 *
 * The following algorithms are availavle:\n
 * - shortes path (press k)\n
 * - least height meters path (press f)\n
 * - maximum Hm/km (press m)\n
 * - STRONG (press s)\n
 * - CLIMB (as STRONG, but muximizing total climb index (press c)\n
 * By clicking a cross at the track you can pin it to fix sections, drag and
 * drop to modify track is possible in routing modes k and f.\n
 * There is Undo and Redo function called by z and y key which loads the last
 * released track.\n
 * Every routing mode can be done without end point pressing ENTER.\n
 * Restrict the road size by specifying upper and lower allowed road type,
 * pressing a number between 0 and 5 with our withour ctrl.
 *
 * \section secstrong STRONG
 * The investigation of the shortes track with 10000 Hm was main purpose of
 * BTP. This can be done by routing mode s. Setup such a calculation with
 * Settings->%STRONG settings. Be aware that depending on your choice of
 * parameters the calculation is memory intense (you can easily exceed your RAM
 * capacity) and may take several hours depending on extent of calculation and
 * CPU power.\n
 * Play arround easpecially with the step parameter and der Hm parameter and
 * check its influence observing system monitor.\n
 * It is not necessary to state a start or endpoint of the track allowing to
 * find the %STRONG-track in a global view not starting from your home, but
 * restricted to your homeland.\n
 * You can stop such a calculation every time you want by ESC. Choose the
 * desired Hm number by PAGE UP/DOWN und accept by enter. As one calculation is
 * time consuming an therefore valuable, you can save and load it with
 * SAVE/LOAD %STRONG calculation.\n
 * If you are interested in a track ending in a specific point, press n after
 * calculation is ready.
 *\subsection secstronglayer STRONG layer
 * With a finished calculation it is possible to create an overlay indicating
 * the most convenient roads to plan a Hm maximized route. The roads are
 * broadend by its usage frequency when backtracing the %STRONG routes from a
 * collection of arbitary but homogeniously distributed end points.
 *
 * \section secpower Power calculation
 * A assumption of power output of the rider is calculated for every created
 * released track. It can be done for existing gpx files as well.\n
 * There are the following contributions:
 * - rolling resistence (depending on speed, mass, rolling coefficient)
 * - aerodynamic drag (speed (wind + movement), cwA, air density (sea level
 * density, height, sea level temperature))
 * - accelaration (after breaking with breaking acceleration and maximum
 * lateral acceleration)\n
 * Major assumption of the whole calculation is a speed dependend power output
 * effecting in both enhanced power for climbing an accelaration.
 *
 * \section secprocess Data Processing
 * Aside the STRONGlayer control, void correction and STRONG save and load support
 * it is possible to perform a CimbAnalysis on the loaded dataset. It will
 * extract all sufficiently independent climbs of the data area, whereat the
 * definition of climb is done by the user.\n
 * Further it is possible to influence the sectional Hm between cross making
 * ascents to descents and vis versa. This is of particular interest when
 * creating meaningful STRONG layers. Normalizing this kind of data will avoid
 * STRONG calculations ending at most highest mountains.
 *
 * \section secbug Bugs and Contribution
 * Please feel free to send me Bugs you notice when working with BTP3. I aim
 * for a stable, user friendly BTP under Win7 and Win8. With the source code
 * it might be possible for you the build a Linux executable or even other
 * operating system supported by Qt.\n
 * If you like to support the project or realize your own ideas within the
 * framwork of BTP3, so do it: Get http://qt-project.org/wiki/Qt_5.0 and begin
 * coding, I hope this
 * documentation helps. For me it would be greates satisfaction to get BTP
 * growing by the help of other enthusiastic programmers. Please refer to my
 * Email if you like to collaborate.\n\n
 * Thanks so far to the OpenStreetMap community, Qt und NASA to enrich mankind
 * by its free data and programms.\n
 * Remember BTP being only as good as its underlying data set from
 * OpenStreetMap, you personally can influence the quality of the data set by
 * editing it at http://www.openstreetmap.org.
 */



#ifndef MAINWINDOW_H
#define MAINWINDOW_H
#pragma once
#include <QtWidgets/QMainWindow>
#include <QPixmap>
#include <QResizeEvent>
#include <stdio.h>
#include <QtWidgets/QFileDialog>
#include <QTime>
#include <QtWidgets/QMessageBox>
#include <QtConcurrent/QtConcurrent>
#include "DataTyps.h"
#include "HeightData.h"
#include "MAP.h"
#include "Routing.h"
#include "STRONGcalc.h"
#include "UndoRedo.h"
#include "WayHeight.h"
#include "btp3setups.h"
#include "ClimbAnalyse.h"
#include "poweranalysis.h"
#include "gpx.h"
#include "webdata.h"
#include "SummitRanking.h"
#include "terrain.h"
#include "FIT_converter/fit_convert.h"


//! main class
/*!
  handles user Inputs, draws maps, routes, calculates power
*/
namespace Ui {
class MainWindow;
}

class MainWindow : public QMainWindow
{
    Q_OBJECT
    
public:
    explicit MainWindow(QWidget *parent = 0);
    ~MainWindow();
    void closeEvent(QCloseEvent *);
    void load_btp(QString filename);
    void save_btp(QString filename, double minlat, double maxlat,
                  double minlon, double maxlon);
    void read_osm(QString filename);
    void analyse_gpx(QString filename);
    HeightData* hd;
    OSM* osm;
    config c;
    int autosave;
private slots:
    void update_map_track();            /*!< draws markers/tracks on raw map*/
    void got_to_latlon();
    void set_SRTM3_folder();
    void set_SRTM1_folder();
    void set_SRTM066_folder();
    void set_SRTM018_folder();
    void DGM2_to_SRTM018();
    void DGM20_to_SRTM066();
    void set_start_btp();
    void load_btp();
    void save_btp();
    void read_osm();
    void analyse_gpx();                 //!< gpx is loaded disregarding data
    void import_gpx();                  //!< gpx is to be mapped onto data
    void training_diary();              //!< diary entries from fit files

    //! conroles communication from profil to mainwindow
    void toggle_poweranalysis();
    void profilinfo(double x);
    void profil_settings();
    void map_settings();
    void STRONG_preview(int i);         /*!< refreshes STRONG calc*/
    void STRONG_settings();             /*!< settings user interface*/

    void route_STRONG_finalize(int succes);
                                        /*!< finallizes succesful STRONG calc*/
    void save_profile();
    void save_gpx();
    void save_segment_gpx();
    void save_track_binary();
    void save_track_binaryPlus();
    void convert_binary_to_binaryPlus();
    void save_btr();
    void load_btr();
    void create_STRONGlayer();
    void save_STRONGlayer();
    void load_STRONGlayer();
    void toggle_STRONGlayer();
    void load_STRONGcalc();
    void save_STRONGcalc();
    void AnalyseClimbs();
    void load_Climblayer();
    void toggle_Climblayer();
    void toggle_HikeBike();
    void coordinate_to_clipbard();
    void network_length_to_clipboard();
    void network_length_to_clipboard(KOO* k, double* d, double* elev, int *C, int *N);
    void generate_Map();                /*!< high end map after user dialog*/
    /*! for routing algorithms: total height meters from to cross to cross
        are reduced by the height difference of the related two cross
    */
    void normalize_Hm();
    /*! for routing algorithms: fwd- and bwd-height meters are swapped,
        interesting for STRONG algorithm forced to finish at top of a mountain
        when the mountain is chosen as start point
     */
    void invert_Hm();
    void swap_Hm_d();
    void update_profil();               /*!< draws ct/pt/st to profil*/
    /*! connected to plot2d::clicked signal, centers map around trackpoint */
    void center_track(double d);
    //! repaints profil with current dataset, but refreshed titel
    void zoom_profil(double dmin, double dmax);
    void toggle_showpower();            /*!< speed is/not shown in plot*/
    void toggle_showspeed();            /*!< power is/not shown in plot*/
    void toggle_bpm();            /*!< bpm is/not shown in plot*/
    //! creates tiles for map server
    /*! loads precalculated btp data and generates tiles from zoomlevel 5 to
     * 14, output are png files, used by openlayers or google map API
     */
    void generate_tiles();
    //! generates 3D view
    void show3D();
    //! splits europe.pbf $osm data
    /*! takes europe data set of $OSM data in pbf file format and splits into
     * either files representing zoomlevel 9 or higher depending on the maximum
     * output file size of 400 MB
     */
    void split_europe_pbf();
    //! slot to organize multithreating task
    /*! receives finished signals from finished tasks restarts new threats if
     * necesary. These leads to a maximum count of threats being active, i.g.
     * optimized for example for eight active threads
     */
    void next_command();
    //! slot to organize multithreating task
    /*! receives signals from working threads to expand tast list Cmdl
     */
    void add_Cmd(webdatacmd* w);
    //! slot to organize multithreating task
    /*! check all SRTM3 hgt-files of a folder for voids and interpolate lineary
     */
    void correct_hgt();
    //! slot to organize multithreating task
    /*! reads $*.osm files and saves btp
     */
    void read_europe();
    //! slot to organize multithreating task
    /*! loads btp and executes ClimbAnalysem, than saves html data for website
     */
    void CA_europe();
    //! slot to organize multithreating task
    /*! loads btp, calcs and saves STRONG and creates transparent tiles
     */
    void STRONG_europe();
    //! copy high zoom level tiles for desired areas in europe
    void zoom14_collection();
    //! split world summit data file into SRTM tiles
    void summitranking_split();
    //! calc world summit lists in SRTM3 tiles size
    void summitranking_calc();
    //! collect moutains with >500 points from the lst files
    void summitranking_collect();
    //! logging for multithreaded classes
    void log_calc(QString logfilename,QString message);
    //! copies list of sites along track to clipboard;
    void list_sites_to_clipboard();
    //! analysis climbs und currently selected track segment
    void climb_analysis_current_segment();
signals:
    void stop_STRONG();
private:
    //! loads settings, initialized variables
    void init();
    void init_config();
    void autoload_setting();
    void autosave_setting();
    //! change of main window size
    /*!
      generates new map and plots it in the size changed window
    */
    void resizeEvent ( QResizeEvent * event );
    //! handles keyboard input
    void keyPressEvent(QKeyEvent *event);
    //! handles all mouse inputs
    void mousePressEvent(QMouseEvent *event);
    void mouseReleaseEvent(QMouseEvent *event);
    void mouseMoveEvent(QMouseEvent *event);
    void mouseDoubleClickEvent(QMouseEvent *event);
    void wheelEvent(QWheelEvent * event );
    //! handles all mouse inputs
    void MouseToTrack();
    //! handles all map modification (mooving, zooming)
    void zoom(double zoomfactor,double dlonfactor, double dlatfactor);
    void update_statusbar();
    //! sets routing mode and shows hints in statusbar
    void set_routingmode(inputstate ip);
    void set_mintype(char mt);          /*!< sets minimum road category*/
    void set_maxtype(char mt);          /*!< sets maximum road category*/
    void delete_to_cross(int cc);       /*!< deletes last cc crosses rom ct */
    void undo();                        /*!< loads news valid backup*/
    void redo();                        /*!< loads newest backup before undo*/
    void enter();                       /*!< handles all return events*/
    void update_map_raw();              /*!< creates map fitting main window*/
    void update_map_MovObj(Way* w);     /*!< draws short time objects*/
    void drawRoute(QPainter *p);        /*!< update_map: draws current tracks*/
    void drawCross(QPainter *p);        /*!< update_map: draws selected cross*/
    void route_airline();               /*!< straight way without osm data*/
    void route_short();                 /*!< shortes path from kold to knew*/
    void route_flat();                  /*!< least Hm from kold to knew*/
    void route_salesman();              /*!< solve traveling sales man problem*/
    void route_time();                  /*!< least time from kold to knew*/
    void route_STRONG();                /*!< most Hm from kold to knew*/
    void acceptSTRONG();                /*!< accepts STRONG as new segment*/
    void selectSTRONG(readdirection rd, short steps);
                                        /*!< show the result of STRONG calc*/
    void delete_STRONG_tracks();
    void toggle_preview();              /*!< choose any endpoint at STRONG?*/
    int  route_preview(int x, int y);   /*!< preview of open end routing*/
    void accept_ct();                   /*!< releases ct and sets lrt*/
    void reset_track();                 /*!< resets track*/

    Ui::MainWindow *ui;
    MAP* map;
    Routing* r;
    PowerAnalysis* PA;
    QPixmap* cmap;                      /*!< last generated map*/
    int x,y;                            /*!< current mouse postion*/
    int xold,yold;                      /*!< old/new mouse postion*/
    KOO *kold,*knew,*kdot;              /*!< old/new click cross postion*/
    KOO* anchorfrom, *anchorto;			/*!< drag&drop between these anchors*/
    int kl_i_from, kl_i_to;             /*!< drag&drop between these anchors*/
    Track* ct;							/*!< current Track*/
    Track* lrt;
    Track** STRONGtracks;				/*!< STRONG-tracks*/
    int sN;                             /*!< STRONG-tracks count*/
    int STRONGindex;                    /*!< STRONG-tracks: current index*/
    UndoRedo* UR;                       /*!< Undo Redo list*/
    dotstate dstate;                    /*!< state of the dot to be drawn*/
    int ydot,xdot;                      /*!< buffer for dots to be drawn*/
    inputstate oldroutemode;
    Way* selected_way;					/*!< way to be editted*/
    QString status1, status2;           /*!< status bar contents*/
    short readytotrace;                 /*!< routing calculation is finished,
                                             any point can be backtraced*/
    int STRONGstate;                    /*!< 0: two ele, 1: threat running,
                                             2: multiple ele*/
    short traceback;                    /*!< try to trace back selected point*/
    bool btp_modified;                  /*!< to remind user to save edits*/
    bool use_marked_ways;
    bool apply_traffic_restrictions;
    int trace_back(KOO* k);             /*!< if ready to trace tries to trace
                                             back beginning with k*/
    void recalc_power();                /*!< triggers PowerAnalyse to calc*/
    //! STRONGlayer
    QColor colorgradient(float x);
    long int SLnbc;
    int SLN, SLc;
    STRONGlayer* SL;
    //! Climblayer
    climblayer* CL;
    QLabel * statusLeft;                /*!< status bar splitting*/
    QLabel * statusMiddle;              /*!< status bar splitting*/
    //! command list to calculate web content multithreated
    QList<webdatacmd> Cmdl;
    int threadscount;                   /*!< number of active threads*/
};

#endif // MAINWINDOW_H
