﻿
#include "HeightData.h"
#define PI 3.14159265
#define RO 6371000

HeightData::HeightData(){
	nReserve = 2;
	sx = sy = -1;
	grid = NULL;
}
void HeightData::set_rect(double minlat,double maxlat,double minlon,double maxlon){
    if(minlat > maxlat) maxlat = minlat = 0;
    if(minlon > maxlon) maxlon = minlon = 0;
    DEMRect.minlat = minlat;
	DEMRect.maxlat = maxlat;
	DEMRect.minlon = minlon;
	DEMRect.maxlon = maxlon;
}
HeightData::~HeightData(void){
	free_grid();
}
void HeightData::create_grid(){
	free_grid();
	int xminlon = floor(DEMRect.minlon*1200);
	int xmaxlon = floor(DEMRect.maxlon*1200);
	int yminlat = floor(DEMRect.minlat*1200);
	int ymaxlat = floor(DEMRect.maxlat*1200);
	GridRect.minlon = xminlon-nReserve;
	GridRect.minlat = yminlat-nReserve;
	GridRect.maxlon = xmaxlon+nReserve+1;
	GridRect.maxlat = ymaxlat+nReserve+1;
	sx = GridRect.maxlon-GridRect.minlon+1;
	sy = GridRect.maxlat-GridRect.minlat+1;

    grid = new short*[sy];
	for(int i = 0; i < sy; i++){
        grid[i] = new short[sx];
		for( int j = 0; j < sx; j++)
			grid[i][j] = -10000;
	}
	
	/*Gitter auslesen*/
	for( int lon = floor(GridRect.minlon/1200.); lon <= floor(GridRect.maxlon/1200.);lon++)
		for( int lat = floor(GridRect.minlat/1200.); lat <= floor(GridRect.maxlat/1200.);lat++){
			char filename[12];
			char path[1000];
			if(lat < 0 && lon < 0)   sprintf(&(filename[0]),"S%02iW%03i.hgt",-lat,-lon);
			if(lat < 0 && lon >= 0)  sprintf(&(filename[0]),"S%02iE%03i.hgt",-lat, lon);
			if(lat >= 0 && lon < 0)  sprintf(&(filename[0]),"N%02iW%03i.hgt", lat,-lon);
			if(lat >= 0 && lon >= 0) sprintf(&(filename[0]),"N%02iE%03i.hgt", lat, lon);

			strcpy(path,SRTM3folder);
			strcat(path,"\\");
			strcat(path,filename);

			int ystart = (lat + 1) * 1200 - max(GridRect.minlat, lat    * 1200);
			int yend   = (lat + 1) * 1200 - min(GridRect.maxlat,(lat+1) * 1200);
			int yos    = (lat + 1) * 1200 - GridRect.minlat;
			int xstart = max(lon     * 1200,GridRect.minlon) - lon * 1200;
			int xend   = min((lon+1) * 1200,GridRect.maxlon) - lon * 1200;
			int xread  = xend - xstart + 1;
			int xos    = max(lon     * 1200,GridRect.minlon) - GridRect.minlon;

			FILE* f = fopen(path,"rb");
			if(f != NULL){
				/*Datei existiert*/
				for(int y = ystart; y >= yend; y--){
					fseek(f,2*(1201*y+xstart),SEEK_SET);
					fread(&(grid[yos-y][xos]),2,xread,f);
				}
				fclose(f);
			}
			else
				/*falls Datei nicht existiert alle Null setzen*/
				for(int y = ystart; y >= yend; y--)
					for(int x = xos; x < xos + xread; x++)
						grid[yos-y][x]=0;
		}

	/*Bytereihenfolge umkehren (Little-Big-Endian)*/
	for(int y = 0; y < sy; y++)
		for( int x = 0; x < sx; x++)
			grid[y][x] = (((grid[y][x])>>8)&0xff)+(((grid[y][x]) << 8)&0xff00);
}
void HeightData::free_grid(){
	if(grid != NULL){
		for(int i = 0; i < sy; i++)
            delete grid[i];
		if(sy>0)
            delete grid;
		grid = NULL;
	}
}
int HeightData::lty(double lat){
	return floor(1200*lat) - GridRect.minlat;
}
int HeightData::ltx(double lon){
	return floor(1200*lon) - GridRect.minlon;
}
double HeightData::ltdy(double lat){
	return (1200*lat - floor(1200*lat));
}
double HeightData::ltdx(double lon){
	return (1200*lon - floor(1200*lon));
}
void HeightData::interpolate_bicubic_conv3(int width,int height,double alpha,
										   double** h, double** dhdx, double** dhdy){
	nReserve = 3;
	create_grid();

	/*Rechenhilfen vorberechnen*/
	double dlon = DEMRect.maxlon - DEMRect.minlon;
	double dlat = DEMRect.maxlat - DEMRect.minlat;
    CubicKernel3* kx = new CubicKernel3[width];
    CubicKernel3* kxx = new CubicKernel3[width];
	for(int x = 0; x < width; x++){
		kxx[x].index[2] = kx[x].index[2] = ltx(1.*x/(width-1)*dlon + DEMRect.minlon);
		kxx[x].index[0] = kx[x].index[0] = kx[x].index[2] - 2;
		kxx[x].index[1] = kx[x].index[1] = kx[x].index[2] - 1;
		kxx[x].index[3] = kx[x].index[3] = kx[x].index[2] + 1;
		kxx[x].index[4] = kx[x].index[4] = kx[x].index[2] + 2;
		kxx[x].index[5] = kx[x].index[5] = kx[x].index[2] + 3;
		double dx = ltdx(1.*x/(width-1)*dlon + DEMRect.minlon);
		kx[x].weights[0] = Cubic_conv_kernel3(dx+2);
		kx[x].weights[1] = Cubic_conv_kernel3(dx+1);
		kx[x].weights[2] = Cubic_conv_kernel3(dx+0);
		kx[x].weights[3] = Cubic_conv_kernel3(dx-1);
		kx[x].weights[4] = Cubic_conv_kernel3(dx-2);
		kx[x].weights[5] = Cubic_conv_kernel3(dx-3);
		kxx[x].weights[0] = Cubic_conv_kernel3_d(dx+2);
		kxx[x].weights[1] = Cubic_conv_kernel3_d(dx+1);
		kxx[x].weights[2] = Cubic_conv_kernel3_d(dx+0);
		kxx[x].weights[3] = Cubic_conv_kernel3_d(dx-1);
		kxx[x].weights[4] = Cubic_conv_kernel3_d(dx-2);
		kxx[x].weights[5] = Cubic_conv_kernel3_d(dx-3);
	}

    CubicKernel3* ky = new CubicKernel3[height];
    CubicKernel3* kyy = new CubicKernel3[height];
	for(int y = 0; y < height; y++){
		kyy[y].index[2] = ky[y].index[2] = lty(1.*y/(height-1)*dlat + DEMRect.minlat);
		kyy[y].index[0] = ky[y].index[0] = ky[y].index[2] - 2;
		kyy[y].index[1] = ky[y].index[1] = ky[y].index[2] - 1;
		kyy[y].index[3] = ky[y].index[3] = ky[y].index[2] + 1;
		kyy[y].index[4] = ky[y].index[4] = ky[y].index[2] + 2;
		kyy[y].index[5] = ky[y].index[5] = ky[y].index[2] + 3;
		double dy = ltdy(1.*y/(height-1)*dlat + DEMRect.minlat);
		ky[y].weights[0] = Cubic_conv_kernel3(dy+2);
		ky[y].weights[1] = Cubic_conv_kernel3(dy+1);
		ky[y].weights[2] = Cubic_conv_kernel3(dy+0);
		ky[y].weights[3] = Cubic_conv_kernel3(dy-1);
		ky[y].weights[4] = Cubic_conv_kernel3(dy-2);
		ky[y].weights[5] = Cubic_conv_kernel3(dy-3);
		kyy[y].weights[0] = Cubic_conv_kernel3_d(dy+2);
		kyy[y].weights[1] = Cubic_conv_kernel3_d(dy+1);
		kyy[y].weights[2] = Cubic_conv_kernel3_d(dy+0);
		kyy[y].weights[3] = Cubic_conv_kernel3_d(dy-1);
		kyy[y].weights[4] = Cubic_conv_kernel3_d(dy-2);
		kyy[y].weights[5] = Cubic_conv_kernel3_d(dy-3);
	}
	if(h != NULL){
		for(int x = 0; x < width; x++)
			for(int y = 0; y < height; y++){
				h   [x][height-1-y] = Bicubic_conv3(kx [x],ky [y]);
				/*fuer genaue aber halb so schnelle Berechnung folgende 2 Zeilen
				aktivieren*/
				//dhdx[x][height-1-y] = Bicubic_conv3(kxx[x],ky [y]);
				//dhdy[x][height-1-y] = Bicubic_conv3(kx [x],kyy[y]);
			}
	}
	free_grid();
    delete(kx);
    delete(kxx);
    delete(ky);
    delete(kyy);
	/*hier die schnell-Berechnung des Gradienten*/
	if(dhdx != NULL){
		double pxperx = width/((DEMRect.maxlon-DEMRect.minlon)*1200.);
		for(int x = 1; x < width-1; x++)
			for(int y = 1; y < height-1; y++)
				dhdx[x][y] = (h[x+1][y]-h[x-1][y])/2.*pxperx;
		for(int i = 0 ; i < width; i++)
			dhdx[i][0] = dhdx[i][height-1] = 0;
	}
	if(dhdy != NULL){
		double pxpery = height/((DEMRect.maxlat-DEMRect.minlat)*1200.);
		for(int x = 1; x < width-1; x++)
			for(int y = 1; y < height-1; y++)
				dhdy[x][y] = (h[x][y+1]-h[x][y-1])/2.*pxpery;
		for(int i = 0 ; i < height; i++)
			dhdy[0][i] = dhdy[width-1][i] = 0;
	}
}
void HeightData::interpolate_bicubic_conv3_mercator(int width,int height,
                                                    double alpha,double** h,
                                                    double** dhdx, double** dhdy){
    nReserve = 3;
    create_grid();

    /*Rechenhilfen vorberechnen*/
    double dlon = DEMRect.maxlon - DEMRect.minlon;
    CubicKernel3* kx = new CubicKernel3[width];
    CubicKernel3* kxx = new CubicKernel3[width];
    for(int x = 0; x < width; x++){
        kxx[x].index[2] = kx[x].index[2] = ltx((0.5+x)/width*dlon + DEMRect.minlon);
        kxx[x].index[0] = kx[x].index[0] = kx[x].index[2] - 2;
        kxx[x].index[1] = kx[x].index[1] = kx[x].index[2] - 1;
        kxx[x].index[3] = kx[x].index[3] = kx[x].index[2] + 1;
        kxx[x].index[4] = kx[x].index[4] = kx[x].index[2] + 2;
        kxx[x].index[5] = kx[x].index[5] = kx[x].index[2] + 3;
        double dx = ltdx((0.5+x)/width * dlon + DEMRect.minlon);
        kx[x].weights[0] = Cubic_conv_kernel3(dx+2);
        kx[x].weights[1] = Cubic_conv_kernel3(dx+1);
        kx[x].weights[2] = Cubic_conv_kernel3(dx+0);
        kx[x].weights[3] = Cubic_conv_kernel3(dx-1);
        kx[x].weights[4] = Cubic_conv_kernel3(dx-2);
        kx[x].weights[5] = Cubic_conv_kernel3(dx-3);
        kxx[x].weights[0] = Cubic_conv_kernel3_d(dx+2);
        kxx[x].weights[1] = Cubic_conv_kernel3_d(dx+1);
        kxx[x].weights[2] = Cubic_conv_kernel3_d(dx+0);
        kxx[x].weights[3] = Cubic_conv_kernel3_d(dx-1);
        kxx[x].weights[4] = Cubic_conv_kernel3_d(dx-2);
        kxx[x].weights[5] = Cubic_conv_kernel3_d(dx-3);
    }

    double* ylat = new double[height];
    CubicKernel3* ky = new CubicKernel3[height];
    CubicKernel3* kyy = new CubicKernel3[height];
    double mmax = lattomercator(DEMRect.maxlat);
    double mmin = lattomercator(DEMRect.minlat);
    double dlat = mmax - mmin;
    for(int y = 0; y < height; y++){
        ylat[y] = mercatortolat((0.5+y)/height*dlat + mmin);
        kyy[y].index[2] = ky[y].index[2] = lty(ylat[y]);
        kyy[y].index[0] = ky[y].index[0] = ky[y].index[2] - 2;
        kyy[y].index[1] = ky[y].index[1] = ky[y].index[2] - 1;
        kyy[y].index[3] = ky[y].index[3] = ky[y].index[2] + 1;
        kyy[y].index[4] = ky[y].index[4] = ky[y].index[2] + 2;
        kyy[y].index[5] = ky[y].index[5] = ky[y].index[2] + 3;
        double dy = ltdy(ylat[y]);
        ky[y].weights[0] = Cubic_conv_kernel3(dy+2);
        ky[y].weights[1] = Cubic_conv_kernel3(dy+1);
        ky[y].weights[2] = Cubic_conv_kernel3(dy+0);
        ky[y].weights[3] = Cubic_conv_kernel3(dy-1);
        ky[y].weights[4] = Cubic_conv_kernel3(dy-2);
        ky[y].weights[5] = Cubic_conv_kernel3(dy-3);
        kyy[y].weights[0] = Cubic_conv_kernel3_d(dy+2);
        kyy[y].weights[1] = Cubic_conv_kernel3_d(dy+1);
        kyy[y].weights[2] = Cubic_conv_kernel3_d(dy+0);
        kyy[y].weights[3] = Cubic_conv_kernel3_d(dy-1);
        kyy[y].weights[4] = Cubic_conv_kernel3_d(dy-2);
        kyy[y].weights[5] = Cubic_conv_kernel3_d(dy-3);
    }
    if(h != NULL){
        for(int x = 0; x < width; x++)
            for(int y = 0; y < height; y++){
                h   [x][height-1-y] = Bicubic_conv3(kx [x],ky [y]);
                /*fuer genaue aber halb so schnelle Berechnung folgende 2 Zeilen
                aktivieren*/
                //dhdx[x][height-1-y] = Bicubic_conv3(kxx[x],ky [y]);
                //dhdy[x][height-1-y] = Bicubic_conv3(kx [x],kyy[y]);
            }
    }
    free_grid();
    delete(kx);
    delete(kxx);
    delete(ky);
    delete(kyy);
    /*hier die schnell-Berechnung des Gradienten*/
    if(dhdx != NULL){
        double pxperx = width/((DEMRect.maxlon-DEMRect.minlon)*1200.);
        for(int x = 1; x < width-1; x++)
            for(int y = 1; y < height-1; y++)
                dhdx[x][y] = (h[x+1][y]-h[x-1][y])/2.*pxperx;
        for(int i = 0 ; i < width; i++)
            dhdx[i][0] = dhdx[i][height-1] = 0;
    }
    if(dhdy != NULL){
        double pxpery;// = height/((DEMRect.maxlat-DEMRect.minlat)*1200.);

        for(int y = 1; y < height-1; y++){
            pxpery = 2./((ylat[y+1]-ylat[y-1])*1200.);
            for(int x = 1; x < width-1; x++)
                dhdy[x][y] = (h[x][y+1]-h[x][y-1])/2.*pxpery;
        }
        for(int i = 0 ; i < height; i++)
            dhdy[0][i] = dhdy[width-1][i] = 0;
    }
    delete ylat;
}
inline double HeightData::Cubic_conv_kernel3(double x){
    x = qAbs(x);
	if(x >= 3) return 0;
	if(x > 2) return x*x*x/12.-2.*x*x/3.+21.*x/12.-3./2.;
	if(x > 1) return -7.*x*x*x/12.+3*x*x-59.*x/12+15./6.;
	if(x >=0) return 4.*x*x*x/3.-7.*x*x/3.+1.;
    else return 0;
}
inline double HeightData::Cubic_conv_kernel3_d(double x){
	int IsGreaterZero = 1;
	if(x < 0){
		IsGreaterZero = 0;
		x = -x;
	}
	double result = 0;
	if(x > 3) result = 0;
	if(x < 3 && x >= 2) result = x*x/4.-4.*x/3.+7./4.;
	if(x < 2 && x >= 1) result = -7.*x*x/4.+6*x-59./12.;
	if(x < 1 && x >= 0) result = 4.*x*x-14.*x/3.;

	if(IsGreaterZero)
		return result;
	else
		return -result;
}
inline double HeightData::Bicubic_conv3(CubicKernel3 kx, CubicKernel3 ky){
	double result = 0;
	for( int x = 0; x < 6; x++)
		for(int y = 0; y < 6; y++)
			result = result + grid[ky.index[y]][kx.index[x]] 
							  * kx.weights[x] * ky.weights[y];
	return result;
}
void HeightData::set_SRTM3_folder(char* folder){
	strcpy(SRTM3folder,folder);
}
void HeightData::reset_t(double minlat,double maxlat,double minlon,double maxlon){
	nReserve = 3;
	set_rect(minlat,maxlat,minlon,maxlon);
	create_grid();
}
int HeightData::max_t(){
	int result = -10000;
	if(grid != NULL){
		result = int(grid[0][0]);
		for(int y = 0; y < sy; y++)
			for( int x = 0; x < sx; x++)
				if(grid[y][x] > result)
					result = grid[y][x];
	}
	return result;
}
int HeightData::max_dt(){
    int min,max;
    if(grid != NULL){
        min = max = int(grid[0][0]);
        for(int y = 0; y < sy; y++)
            for( int x = 0; x < sx; x++){
                if(grid[y][x] > max)
                    max = grid[y][x];
                if(grid[y][x] < min)
                    min = grid[y][x];
            }
        return max-min;
    }
    else
        return 0;
}
double HeightData::height_t_save(double lon,double lat,void* hc,short mode){
    if(lon > DEMRect.minlon &&
       lon < DEMRect.maxlon &&
       lat > DEMRect.minlat &&
       lat < DEMRect.maxlat)
        return height_t(lon,lat,hc,mode);
    else return 0;
}

double HeightData::height_t(double lon,double lat,void* hc,short mode){
	CubicKernel3 kx;
	kx.index[2] = ltx(lon);
	kx.index[0] = kx.index[2] - 2;
	kx.index[1] = kx.index[2] - 1;
	kx.index[3] = kx.index[2] + 1;
	kx.index[4] = kx.index[2] + 2;
	kx.index[5] = kx.index[2] + 3;
	CubicKernel3 ky;
	ky.index[2] = lty(lat);
	ky.index[0] = ky.index[2] - 2;
	ky.index[1] = ky.index[2] - 1;
	ky.index[3] = ky.index[2] + 1;
	ky.index[4] = ky.index[2] + 2;
	ky.index[5] = ky.index[2] + 3;

	double dx = ltdx(lon);
	double dy = ltdy(lat);

	/*kontrolliere, dass indeces im grid enthalten sind*/
	if(ky.index[0] < 0 || ky.index[5] >= sy || 
	   kx.index[0] < 0 || kx.index[5] >= sx)
	   return 0;

    double h = 0,dhdx=0,dhdy=0;
	if(mode == 0 || hc != NULL){
		/*einfache Hoehe berechnen*/
		for(int i = 0; i < 6; i++){
			kx.weights[i] = Cubic_conv_kernel3(dx+2-i);
			ky.weights[i] = Cubic_conv_kernel3(dy+2-i);
		}
        h = Bicubic_conv3(kx,ky);
	}
	if(mode == 1 || mode == 3 || mode == 4){
		/*Ableitung in x berechnen*/
		for(int i = 0; i < 6; i++){
			kx.weights[i] = Cubic_conv_kernel3_d(dx+2-i);
			ky.weights[i] = Cubic_conv_kernel3(dy+2-i);
		}
        dhdx = Bicubic_conv3(kx,ky);
	}
	if(mode == 2 || mode == 3 || mode == 4){
		/*Ableitung in y ausgeben*/
		for(int i = 0; i < 6; i++){
			kx.weights[i] = Cubic_conv_kernel3(dx+2-i);
			ky.weights[i] = Cubic_conv_kernel3_d(dy+2-i);
		}
		dhdy = Bicubic_conv3(kx,ky);
	}

	if(hc != NULL)						//Hoehe mitgeben
		*((float*)hc) = h;
	if(mode == 0)						//Hoehe ausgeben
		return h;
	if(mode == 1)						//x-Ableitung ausgeben
		return dhdx;
	if(mode == 2)						//y-Ableitung ausgeben
		return dhdy;
	if(mode == 3)						//Gradientenrichtung ausgeben
		return atan2(double(dhdx),double(dhdy*cos(lat/180*PI)));
	if(mode == 4){						//Vertrauensindex mit- u. Hoehe ausgeben
		//((double*)hc)[0] = abs(((double*)hc)[2]*dhdy-((double*)hc)[1]*dhdx/cos(lat/180*PI));
		double hn[4];
		double dlat = 0.00007, dlon = dlat/cos(lat/180.*PI);
        hn[0] = height_t(lon-2*dlon*((double*)hc)[1],lat+2*dlat*((double*)hc)[2],NULL,0);
		hn[1] = height_t(lon-1*dlon*((double*)hc)[1],lat+1*dlat*((double*)hc)[2],NULL,0);
		hn[2] = height_t(lon+1*dlon*((double*)hc)[1],lat-1*dlat*((double*)hc)[2],NULL,0);
		hn[3] = height_t(lon+2*dlon*((double*)hc)[1],lat-2*dlat*((double*)hc)[2],NULL,0);
		((double*)hc)[0]= (2*hn[0]-hn[1]-2*h-hn[2]+2*hn[3])/14.*
                          qAbs(((double*)hc)[2]*dhdy-((double*)hc)[1]*dhdx/cos(lat/180*PI));
		return h;
	}
    else
        return 0;
}
double HeightData::height(double lon,double lat,void* hc,short mode){
	set_rect(lat,lat,lon,lon);
	nReserve = 3;
	create_grid();
	double result = height_t(lon,lat,hc,mode);
	free_grid();
	return result;
}
void HeightData::paek_pro(double lon, double lat, qint64** c,float* Ahigher,
						  int cc, int ac,float d,float mperlon, float mperlat){
	int j,i,x,y,ymin,ymax,xmin,xmax, Ac = 0, AcF = 0;
	float x0 = lon*1200 - GridRect.minlon;	//Gipfellage
	float y0 = lat*1200 - GridRect.minlat;
	float sx = mperlon/1000;			//Skalierung km je x
	float sy = mperlat/1000;			//Saklierung km je y
	float h0 = height_t(lon,lat,NULL,0);

	/*Punktzaehler anlegen*/
    int **wc = new int*[ac];
 	for(i = 0; i < ac; i++)
        wc[i] = new int[cc];

	/*Höhenzaehler und Punktzaehler initialisieren*/
	for(i = 0; i < ac; i++)
		for(j = 0; j < cc; j++){
			c[i][j] = 0;					//Zählt alle Höhen
			wc[i][j] = 0;					//Anzahl enthaltener Punkte
		}

    xmin = qMax(0         ,int(floor(x0-(d*cc)/sx)));
    xmax = qMin(this->sx-1,int (ceil(x0+(d*cc)/sx)));
	for(x = xmin; x <= xmax; x++){
        ymin = qMax(0    ,int(floor(y0-sqrt((d*cc*d*cc)-((x-x0)*sx*(x-x0)*sx)+1)/sx)));
        ymax = qMin(this->sy-1,int(ceil (y0+sqrt((d*cc*d*cc)-((x-x0)*sx*(x-x0)*sx)+1)/sx)));
		for(y = ymin; y <= ymax;  y++){
			i = int(floor((atan2(y-y0,x-x0)/PI+1)/2*ac));
			j = int(floor(sqrt(((y-y0)*sy*(y-y0)*sy)+((x-x0)*sx*(x-x0)*sx))/d));
            if(i < cc && grid[y][x] < 9000
			   && i < ac && j < cc){
				c[i][j] = c[i][j] + long(grid[y][x]);
				wc[i][j]++;
				if(grid[y][x] > h0) Ac++;
				AcF++;
			}
		}
	}
	/*Durchschnitt bilden*/
	for(i = 0; i < ac; i++)
		for(j = 0; j < cc; j++)
			if(wc[i][j] > 0)
				c[i][j] = c[i][j] / wc[i][j];
			else
				c[i][j] = 12000;
	
	/*Punktzaehler löschen*/
	for(i = 0; i < ac; i++)
        delete(wc[i]);
    delete(wc);

	*Ahigher = float(1.*Ac/AcF);

}
void HeightData::interpolate_linear(int width,int height,double alpha,
							double** h, double** dhdx, double** dhdy){
	nReserve = 3;
	create_grid();
	
	int ix,iy;
	for(int x = 0; x < width; x++)
		for(int y = 0; y < height; y++){
			ix = ltx(DEMRect.minlon + (DEMRect.maxlon-DEMRect.minlon)*(1.*x/(width -1)));
			iy = lty(DEMRect.minlat + (DEMRect.maxlat-DEMRect.minlat)*(1.*y/(height-1)));
			h[x][height-1-y]	   = 1.*( + grid[iy  ][ix] + grid[iy  ][ix+1]
							  + grid[iy+1][ix] + grid[iy+1][ix+1])/4.;
			dhdx[x][height-1-y] = 1.*( - grid[iy  ][ix] + grid[iy  ][ix+1]
						      - grid[iy+1][ix] + grid[iy+1][ix+1])/2.;
			dhdy[x][height-1-y] =-1.*( - grid[iy  ][ix] - grid[iy  ][ix+1]
						      + grid[iy+1][ix] + grid[iy+1][ix+1])/2.;
		}
}
void HeightData::interpolate_linear_mercator(int width,int height,double alpha,
                            double** h, double** dhdx, double** dhdy){
    nReserve = 3;
    create_grid();
    double mmax = lattomercator(DEMRect.maxlat);
    double mmin = lattomercator(DEMRect.minlat);
    int ix,iy;
    for(int y = 0; y < height; y++){
        iy = lty(mercatortolat(mmin + (mmax-mmin)*((0.5+y)/height)));
        for(int x = 0; x < width; x++){
            ix = ltx(DEMRect.minlon + (DEMRect.maxlon-DEMRect.minlon)*((0.5+x)/width));
            h[x][height-1-y]	   = 1.*( + grid[iy  ][ix] + grid[iy  ][ix+1]
                              + grid[iy+1][ix] + grid[iy+1][ix+1])/4.;
            dhdx[x][height-1-y] = 1.*( - grid[iy  ][ix] + grid[iy  ][ix+1]
                              - grid[iy+1][ix] + grid[iy+1][ix+1])/2.;
            dhdy[x][height-1-y] =-1.*( - grid[iy  ][ix] - grid[iy  ][ix+1]
                              + grid[iy+1][ix] + grid[iy+1][ix+1])/2.;
        }
    }
}

void HeightData::get_sorted_grid_data(short** heights, int** lons, int** lats,
                                      int** gradx, int** grady, long* count){
	/*extracts all grid values with KOO into three list, sorted by height
	beginning with the highest*/
	*count = sy*sx;
	qsh = *heights = new short[*count];
    qslon = *lons = new int[*count];
    qslat= *lats = new int[*count];
    qsx= *gradx = new int[*count];
    qsy= *grady = new int[*count];
    int x,y,deadlevel;
    double h0,dh,dht,Acorr,gcorr,dtt;
	/*Extracting*/
    for(int ii = 0; ii < sy; ii++){
        (*lats)[ii*sx] = 1.*(GridRect.minlat+ii)/1200.;
        Acorr = cos((*lats)[ii*sx]*0.0174532922);
        gcorr = 1./sqrt(Acorr*Acorr+1);
        Acorr = 1./Acorr;
        for( int jj = 0; jj < sx; jj++){
			(*heights)[ii*sx+jj] = grid[ii][jj];
            (*lons)[ii*sx+jj] = jj;
            (*lats)[ii*sx+jj] = ii;
            //now trace uphill
            deadlevel = 1;
            h0 = grid[ii][jj];
            x = jj;
            y = ii;
            if(y > 0 && y < sy-1 && sx > 0 && x < sx-1){
                dh = 0;
                deadlevel = 1;
                int gt[8][2] = {{-1,-1},{-1,+1},{+1,-1},{+1,+1},
                                {0,-1},{0,+1},{+1,0},{-1,0}};
                double gf[8];
                gf[0] = gcorr;
                gf[1] = gcorr;
                gf[2] = gcorr;
                gf[3] = gcorr;
                gf[4] = Acorr;
                gf[5] = Acorr;
                gf[6] = 1;
                gf[7] = 1;

                dh = 0;

                for(int j = 0; j < 8; j++){
                    dtt = (grid[ii+gt[j][0]][jj+gt[j][1]]-h0)*gf[j];
                    if(dtt > dh){
                        dh = dtt;
                        x = jj+gt[j][1];
                        y = ii+gt[j][0];
                    }
                    if(dtt != dh)
                        deadlevel = 0;
                }
            }
            if(!deadlevel){
                (*gradx)[ii*sx+jj] = x;
                (*grady)[ii*sx+jj] = y;
            }
            else{
                (*gradx)[ii*sx+jj] = -1;
                (*grady)[ii*sx+jj] = -1;
            }
		}
    }
	/*QuickSort*/
	sort_GridList(0,*count-1);
}
void HeightData::sort_GridList(long s, long e){
	short sb;
    int ib;
	long l = s, r = e;
	double p = qsh[(s+e)/2], db;
	do{
		while(qsh[l] < p) l++;
		while(qsh[r] > p) r--;
		if(l<=r){
			sb = qsh[l]; qsh[l] = qsh[r]; qsh[r] = sb;
			db = qslon[l]; qslon[l] = qslon[r]; qslon[r] = db;
			db = qslat[l]; qslat[l] = qslat[r]; qslat[r] = db;
            ib = qsx[l]; qsx[l] = qsx[r]; qsx[r] = ib;
            ib = qsy[l]; qsy[l] = qsy[r]; qsy[r] = ib;
			l++;
			r--;
		}
	}while(l <= r);
	if(s < r) sort_GridList(s,r);
	if(l < e) sort_GridList(l,e);
}
recti HeightData::get_GridRect(){
	return GridRect;
}
int HeightData::correct_hgt(char* filename){
    char fn[100];
    short badheight = -1000;
    int lnl = strlen(filename);
    strcpy(&(fn[0]),&(filename[strlen(filename)-11]));
    int lon = -190,lat = -190;
    fn[7]='\0';
    if(fn[3]=='E') lon =   atoi(&(fn[4]));
    if(fn[3]=='W') lon = - atoi(&(fn[4]));
    fn[3]='\0';
    if(fn[0]=='N') lat =   atoi(&(fn[1]));
    if(fn[0]=='S') lat = - atoi(&(fn[1]));
    bool corrected = false;
    if(lon >= -180 && lon < 180 && lat >= -90 && lat < 90){
        set_rect(lat,lat+1,lon,lon+1);
        nReserve = 100;
        create_grid();
        for(int x = 99; x < 1300; x++)
            for(int y = 99; y < 1300; y++)
                if(grid[y][x] < badheight){
                    corrected = true;
                    int ym = y, yp = y, xm = x, xp = x;
                    double hy = badheight-100., hx = badheight-100.;
                    while(grid[ym][x] < badheight && ym > 0)    ym--;
                    while(grid[yp][x] < badheight && yp < sy-1) yp++;
                    while(grid[y][xm] < badheight && xm > 0)    xm--;
                    while(grid[y][xp] < badheight && xp < sx-1) xp++;
                    short y0 = grid[y][x];
                    short y1 = grid[yp][x];
                    short y2 = grid[ym][x];
                    short y3 = grid[y][xp];
                    short y4 = grid[y][xm];
                    if(grid[ym][x] > badheight && grid[yp][x] > badheight)
                        hy = ((yp-y)*grid[ym][x]+(y-ym)*grid[yp][x])/(yp-ym);
                    if(grid[ym][x] > badheight && grid[yp][x] > badheight)
                        hx = ((xp-x)*grid[y][xm]+(x-xm)*grid[y][xp])/(xp-xm);
                    if(hy > badheight && hx > badheight)
                        grid[y][x] = (hx+hy)/2;
                    else{
                        if(hy > badheight)
                            grid[y][x] = hy;
                        else
                            grid[y][x] = hx;
                    }
                }
        if(corrected){
            /*Datei existiert*/
            FILE* f = fopen(filename,"wb");
            for(int y = 1300; y >= 99; y--){
                for(int x = 99; x < 1300; x++)
                    grid[y][x] = (((grid[y][x])>>8)&0xff)+(((grid[y][x]) << 8)&0xff00);
                fwrite(&(grid[y][100]),2,1201,f);
            }
            fclose(f);
        }
        return corrected;
    }
    else
        return 0;
}
