﻿#include "osm.h"

OSM::OSM(string fileName){
    setfilename(fileName);
    initiallists();
}
OSM::OSM(){
    initiallists();
}
OSM::~OSM(){
    delete_AVL(*(D.nodes));
    delete_AVL(*(D.waysavl));
    for(int i = 0; i <= MaxNature; i++)
        delete_lines(*(ln[i]));
    for(int i = 0; i <= MaxPoints; i++)
        delete_points(*(pt[i]));
    delete_relation(*woodr);
    delete_relation(*waterr);
    delete_relation(*router);
}
qint64 OSM::extract_valuei(char* s, const char* a){
    char *x1;
    char *x2;
    x1 = strstr(s,a)+strlen(a)+2*sizeof(char);
    if(x1 == NULL)
        QMessageBox(QMessageBox::NoIcon,"Error",QString("Could not extract integer value from '%1'").arg(s)).exec();
    x2 = strchr(x1,'\'');
    *x2 = '\0';
    qint64 result = QString(x1).toLongLong();
    *x2 = '\'';
    return result;
}
double OSM::extract_valuef(char* s, const char* a){
    char *x1;
    char *x2;
    x1 = strstr(s,a)+strlen(a)+2*sizeof(char);
    if(x1 == NULL)
        QMessageBox(QMessageBox::NoIcon,"Error",QString("Could not extract double value from '%1'").arg(s)).exec();

    x2 = strchr(x1,'\'');
    *x2 = '\0';
    double result = atof(x1);
    *x2 = '\'';
    return result;
}
void OSM::read_OSM(){
    if(status >= 1){
        FILE *f;
        f = fopen(filename.c_str(),"r");
        if (f == NULL){
            printf("Fehler beim öffnen der Datei ");
            printf(filename.c_str());
        }
        else{
            char line[10000],lineattr[10000],linename[10000];

            qint64 currentseek;
            dminlat = 0;
            dmaxlat = 0;
            dminlon = 0;
            dmaxlon = 0;

            fseek(f,0,SEEK_SET);
            while(!feof(f)){
                currentseek = _ftelli64(f);
                fgets(line,999,f);
                quotationreplace(line);
                if(strstr(line,"bbox=") != NULL
                    || strstr(line,"<bound box=") != NULL)
                    extract_bbox(line);
                if(strstr(line,"<bounds ") != NULL){
                    dminlat = extract_valuef(line," minlat");
                    dmaxlat = extract_valuef(line," maxlat");
                    dminlon = extract_valuef(line," minlon");
                    dmaxlon = extract_valuef(line," maxlon");
                }
                if (strstr(line,"<node ") != NULL)
                    nodehandling(f,line,lineattr,linename);
                if (strstr(line,"<way ") != NULL)
                    Wayhandling(f,line,lineattr,linename,&currentseek);
                if (strstr(line,"<relation ") != NULL)
                    relationhandling(f,line,lineattr,linename,&currentseek);
            }
            fclose(f);
            status = 2;
            //invert_points(pt[7]);
        }
    }
    else{
        printf("Initialisieren korrekten Dateinamen mit OSM(fileName)!\n");
    }
}
KOO* OSM::search(qint64 id, KOO* s){
    while (s != NULL && s->id != id){
        if(s->id > id) s = s->l;
        else s = s->r;
    }
    return s;
}
KOO* OSM::insert(qint64 id, KOO* s){
    KOO* p = s;
    if (s != NULL){
        /*Falls der Baum schon angelegt ist*/
        while (s != NULL && s->id != id){
            /*suche das zugehörige Blatt*/
            if(s->id > id){
                p = s;
                s = s->l;
            }
            else
                if(s->id < id){
                    p = s;
                    s = s->r;
                }
        }
        if(s != NULL)
            /*das Blatt ist schon vorhanden*/
            return s;
        else{
            /*Blatt muss angelegt werden*/
            if(p->id > id){
                /*Blatt links*/
                p->l = new KOO(id);
                s = p->l;
                s->prev = p;
                return s;
            }
            else{
                /*Blatt rechts*/
                p->r = new KOO(id);
                s = p->r;
                s->prev = p;
                return s;
            }
        }
    }
    else{
        /*Baum anlegen, weil noch nicht geschehen*/
        s = new KOO(id);
        return s;
    }
}
KOO* OSM::rotate(KOO** s, branch b){
    KOO* k;
    switch(b){
        case l:	switch((*s)->balance){
                case +1:switch((*s)->r->balance){
                        case -1:(*s)->balance = 0;
                                (*s)->r->balance = -2;
                                break;
                        case 0: (*s)->balance = 0;
                                (*s)->r->balance = -1;
                                break;
                        case 1: (*s)->balance = -1;
                                (*s)->r->balance = -1;
                                break;
                        }
                        break;
                case +2:switch((*s)->r->balance){
                        case -1:(*s)->balance = 1;
                                (*s)->r->balance = -2;
                                break;
                        case 0: (*s)->balance = 1;
                                (*s)->r->balance = -1;
                                break;
                        case 1: (*s)->balance = 0;
                                (*s)->r->balance = 0;
                                break;
                        case 2: (*s)->balance = -1;
                                (*s)->r->balance = 0;
                                break;
                        }
                        break;
                }
                break;
        case r:	switch((*s)->balance){
                case -1:switch((*s)->l->balance){
                        case -1:(*s)->balance = 1;
                                (*s)->l->balance = 1;
                                break;
                        case 0: (*s)->balance = 0;
                                (*s)->l->balance = 1;
                                break;
                        case 1: (*s)->balance = 0;
                                (*s)->l->balance = 2;
                                break;
                        }
                        break;
                case -2:switch((*s)->l->balance){
                        case -2:(*s)->balance = 1;
                                (*s)->l->balance = 0;
                                break;
                        case -1:(*s)->balance = 0;
                                (*s)->l->balance = 0;
                                break;
                        case 0: (*s)->balance = -1;
                                (*s)->l->balance = 1;
                                break;
                        case 1: (*s)->balance = -1;
                                (*s)->l->balance = 2;
                                break;
                        }
                        break;
                }
                break;
    default: break;
    }
    if (b == l){
        k = (*s)->r->l;
        (*s)->r->l = (*s);
        (*s) = (*s)->r;
        (*s)->l->r = k;
        (*s)->prev = (*s)->l->prev;
        (*s)->l->prev = (*s);
        if (k != NULL) k->prev = (*s)->l;
        return (*s);
    }
    else{
        k = (*s)->l->r;
        (*s)->l->r = (*s);
        (*s) = (*s)->l;
        (*s)->r->l = k;
        (*s)->prev = (*s)->r->prev;
        (*s)->r->prev = (*s);
        if (k != NULL) k->prev = (*s)->r;
        return (*s);
    }
}
KOO* OSM::insert_AVL(qint64 id, KOO** s){
    if(*s != NULL){
        int fertig = 0;
        KOO* i = insert(id,*s);
        KOO* result = i;
        if(i->status > -1)				//Falls die KOO schon vorhanden war
            return i;					//ist die gesammte Routine zu beenden

        i->status = 0;
        while(!fertig){
            i = i->prev;
            if (id < i->id){
                i->balance--;
                switch(i->balance){
                    case-2: switch(i->l->balance){
                                case -1: i = rotate(&i,r);
                                         break;
                                case  1: rotate(&(i->l),l);
                                         i = rotate(&i,r);
                                         break;
                            }
                            fertig = 1;
                            if(i->prev != NULL){
                                if(i->prev->id > i->id)
                                    i->prev->l = i;
                                else
                                    i->prev->r = i;
                            }
                            break;
                    case-1: if(i->prev == NULL) fertig = 1;
                            break;
                    case 0: fertig = 1;
                            break;
                }
            }
            else{
                i->balance++;
                switch(i->balance){
                    case 2: switch(i->r->balance){
                                case  1: i = rotate(&i,l);
                                         break;
                                case -1: rotate(&(i->r),r);
                                         i = rotate(&i,l);
                                         break;
                            }
                            fertig = 1;
                            if(i->prev != NULL){
                                if(i->prev->id > i->id)
                                    i->prev->l = i;
                                else
                                    i->prev->r = i;
                            }
                            break;
                    case 1: if(i->prev == NULL) fertig = 1;
                            break;
                    case 0: fertig = 1;
                            break;
                }
            }
        }
        if (i->prev == NULL){
            *s = i;
            return result;
        }
        else
            return result;
    }
    else{
        KOO* result = new KOO(id);
    *s = result;
    return result;
    }
}
Way* OSM::search(qint64 id, Way* s){
    while (s != NULL && s->id != id){
        if(s->id > id) s = s->l;
        else s = s->r;
    }
    return s;
}
Way* OSM::insert(qint64 id, Way* s){
    Way* p = s;
    if (s != NULL){
        /*Falls der Baum schon angelegt ist*/
        while (s != NULL && s->id != id){
            /*suche das zugehörige Blatt*/
            if(s->id > id){
                p = s;
                s = s->l;
            }
            else
                if(s->id < id){
                    p = s;
                    s = s->r;
                }
        }
        if(s != NULL)
            /*das Blatt ist schon vorhanden*/
            return s;
        else{
            /*Blatt muss angelegt werden*/
            if(p->id > id){
                /*Blatt links*/
                p->l = new Way(id);
                s = p->l;
                s->prev = p;
                return s;
            }
            else{
                /*Blatt rechts*/
                p->r = new Way(id);
                s = p->r;
                s->prev = p;
                return s;
            }
        }
    }
    else{
        /*Baum anlegen, weil noch nicht geschehen*/
        s = new Way(id);
        return s;
    }
}
Way* OSM::rotate(Way** s, branch b){
    Way* k;
    switch(b){
        case l:	switch((*s)->balance){
                case +1:switch((*s)->r->balance){
                        case -1:(*s)->balance = 0;
                                (*s)->r->balance = -2;
                                break;
                        case 0: (*s)->balance = 0;
                                (*s)->r->balance = -1;
                                break;
                        case 1: (*s)->balance = -1;
                                (*s)->r->balance = -1;
                                break;
                        }
                        break;
                case +2:switch((*s)->r->balance){
                        case -1:(*s)->balance = 1;
                                (*s)->r->balance = -2;
                                break;
                        case 0: (*s)->balance = 1;
                                (*s)->r->balance = -1;
                                break;
                        case 1: (*s)->balance = 0;
                                (*s)->r->balance = 0;
                                break;
                        case 2: (*s)->balance = -1;
                                (*s)->r->balance = 0;
                                break;
                        }
                        break;
                }
                break;
        case r:	switch((*s)->balance){
                case -1:switch((*s)->l->balance){
                        case -1:(*s)->balance = 1;
                                (*s)->l->balance = 1;
                                break;
                        case 0: (*s)->balance = 0;
                                (*s)->l->balance = 1;
                                break;
                        case 1: (*s)->balance = 0;
                                (*s)->l->balance = 2;
                                break;
                        }
                        break;
                case -2:switch((*s)->l->balance){
                        case -2:(*s)->balance = 1;
                                (*s)->l->balance = 0;
                                break;
                        case -1:(*s)->balance = 0;
                                (*s)->l->balance = 0;
                                break;
                        case 0: (*s)->balance = -1;
                                (*s)->l->balance = 1;
                                break;
                        case 1: (*s)->balance = -1;
                                (*s)->l->balance = 2;
                                break;
                        }
                        break;
                }
                break;
    default: break;
    }
    if (b == l){
        k = (*s)->r->l;
        (*s)->r->l = (*s);
        (*s) = (*s)->r;
        (*s)->l->r = k;
        (*s)->prev = (*s)->l->prev;
        (*s)->l->prev = (*s);
        if (k != NULL) k->prev = (*s)->l;
        return (*s);
    }
    else{
        k = (*s)->l->r;
        (*s)->l->r = (*s);
        (*s) = (*s)->l;
        (*s)->r->l = k;
        (*s)->prev = (*s)->r->prev;
        (*s)->r->prev = (*s);
        if (k != NULL) k->prev = (*s)->r;
        return (*s);
    }
}
Way* OSM::insert_AVL(qint64 id, Way** s){
    if(*s != NULL){
        int fertig = 0;
        Way* i = insert(id,*s);
        if(i->type > -1)				//der Weg war schon vorhanden und die
            fertig = 1;					//Routine kann beendet werden
        Way* result = i;
        while(!fertig){
            i = i->prev;
            if (id < i->id){
                i->balance--;
                switch(i->balance){
                    case-2: switch(i->l->balance){
                                case -1: i = rotate(&i,r);
                                         break;
                                case  1: rotate(&(i->l),l);
                                         i = rotate(&i,r);
                                         break;
                            }
                            fertig = 1;
                            if(i->prev != NULL){
                                if(i->prev->id > i->id)
                                    i->prev->l = i;
                                else
                                    i->prev->r = i;
                            }
                            break;
                    case-1: if(i->prev == NULL) fertig = 1;
                            break;
                    case 0: fertig = 1;
                            break;
                }
            }
            else{
                i->balance++;
                switch(i->balance){
                    case 2: switch(i->r->balance){
                                case  1: i = rotate(&i,l);
                                         break;
                                case -1: rotate(&(i->r),r);
                                         i = rotate(&i,l);
                                         break;
                            }
                            fertig = 1;
                            if(i->prev != NULL){
                                if(i->prev->id > i->id)
                                    i->prev->l = i;
                                else
                                    i->prev->r = i;
                            }
                            break;
                    case 1: if(i->prev == NULL) fertig = 1;
                            break;
                    case 0: fertig = 1;
                            break;
                }
            }
        }
        if (i->prev == NULL){
            *s = i;
            return result;
        }
        else
            return result;
    }
    else{
    Way* result = new Way;
    result->id = id;
    *s = result;
    return result;
    }
}
void OSM::get_KOO_Waysavl(){
    get_KOO_Waysavl(*(D.waysavl),1);
    // remove elements
    for (int i = 0; i <= MaxNature; i++){
        lines* first = NULL;
        lines lbuf; lbuf.next = *(ln[i]);
        Way wbuf; wbuf.type = i; lbuf.way = &wbuf;
        lines* l = &lbuf;
        while(l->next != NULL){
            if(l->next->way->type > MaxNature)
                l->next = l->next->next;
            else{
                if(first == NULL)
                    first = l->next;
                l = l->next;
            }
        }
        *(ln[i]) = first;
    }
}
void OSM::get_KOO_Waysavl(Way* l,short next){
    if(status >=2){
        status = 3;
        qint64 id;
        if(l != NULL){
            l->maxlat = -90;
            l->maxlon = -180;
            l->minlat =  90;
            l->minlon =  180;

            for(int j = 0; j < l->nodec; j++){
                id = l->node[j]->id;
                delete(l->node[j]);
                l->node[j] = search(id, *D.nodes);
                if(l->node[j] == NULL){
                    l->type = 99;
                    break;
                }

                /*Bestimme das umrandende Rechteck*/
                if(l->node[j]->lat > l->maxlat )
                    l->maxlat = float(l->node[j]->lat);
                if(l->node[j]->lat < l->minlat )
                    l->minlat = float(l->node[j]->lat);
                if(l->node[j]->lon > l->maxlon )
                    l->maxlon = float(l->node[j]->lon);
                if(l->node[j]->lon < l->minlon )
                    l->minlon = float(l->node[j]->lon);
            }
            if(next){
                get_KOO_Waysavl(l->l,1);	//rekursiver Aufruf des linken Kindes
                get_KOO_Waysavl(l->r,1);	//rekursiver Aufruf des rechten Kindes
            }
        }
    }
    else{
        printf("Zuerst Datei einlesen mittels read_OSM_all()\n");
    }
}
char* OSM::extract_valuec(char* s, const char* a){
    char *x1;
    char *x2;
    x1 = strstr(s,a)+strlen(a)+2*sizeof(char);
    if(x1 == NULL)
        QMessageBox(QMessageBox::NoIcon,"Error",QString("Could not char value from '%1'").arg(s)).exec();
    x2 = strchr(x1,'\'');
    *x2 = '\0';
    return x1;
}
void OSM::extract_bbox(char* s){
    if(strstr(s,"bbox=") != NULL){
        char* t = strstr(s,"bbox=");
        t = &(t[5]);
        char* t2 = strstr(t,",");
        t2[0]='\0';
        dminlon = atof(t);
        t = &(t2[1]);
        t2 = strstr(t,",");
        t2[0]='\0';
        dminlat = atof(t);
        t = &(t2[1]);
        t2 = strstr(t,",");
        t2[0]='\0';
        dmaxlon = atof(t);
        t = &(t2[1]);
        t2 = strstr(t,"'");
        t2[0]='\0';
        dmaxlat = atof(t);
    }
    else{
        char* t = strstr(s,"box=");
        t = &(t[5]);
        char* t2 = strstr(t,",");
        t2[0]='\0';
        dminlat = atof(t);
        t = &(t2[1]);
        t2 = strstr(t,",");
        t2[0]='\0';
        dminlon = atof(t);
        t = &(t2[1]);
        t2 = strstr(t,",");
        t2[0]='\0';
        dmaxlat = atof(t);
        t = &(t2[1]);
        t2 = strstr(t,"'");
        t2[0]='\0';
        dmaxlon = atof(t);
    }
}
short OSM::isin(lines* l,double minlat, double maxlat,
                         double minlon, double maxlon){
    if	  ( ( (l->way->minlat >= minlat && l->way->minlat <= maxlat)
            ||(l->way->maxlat <= maxlat && l->way->maxlat >= minlat)
            ||(l->way->minlat <= minlat && l->way->maxlat >= maxlat)
            )
        &&
            ( (l->way->minlon >= minlon && l->way->minlon <= maxlon)
            ||(l->way->maxlon <= maxlon && l->way->maxlon >= minlon)
            ||(l->way->minlon <= minlon && l->way->maxlon >= maxlon)
            ) )
        return 1;
    else
        return 0;
}
short OSM::isin(points* l,double minlat, double maxlat,
                          double minlon, double maxlon){
    return isin(l->k,minlat,maxlat,minlon,maxlon);
}
short OSM::isin(KOO* k,double minlat, double maxlat,
                       double minlon, double maxlon){
    if(    k->lat < maxlat
        && k->lat > minlat
        && k->lon > minlon
        && k->lon < maxlon)
        return 1;
    else
        return 0;
}
short OSM::isin(Way* w,double minlat, double maxlat,
                       double minlon, double maxlon){
    if	  ( ( (w->minlat >= minlat && w->minlat <= maxlat)
            ||(w->maxlat <= maxlat && w->maxlat >= minlat)
            ||(w->minlat <= minlat && w->maxlat >= maxlat)
            )
        &&
            ( (w->minlon >= minlon && w->minlon <= maxlon)
            ||(w->maxlon <= maxlon && w->maxlon >= minlon)
            ||(w->minlon <= minlon && w->maxlon >= maxlon)
            ) )
        return 1;
    else
        return 0;
}
short OSM::iscompletein(Way* w,double minlat, double maxlat,
                       double minlon, double maxlon){
    if(	   w->minlat > minlat
        && w->maxlat < maxlat
        && w->minlon > minlon
        && w->maxlon < maxlon)
        return 1;
    else
        return 0;
}
void OSM::setfilename(string fileName){
    if(fileName.compare(fileName.size()-4,4,".osm") == 0){
        filename = fileName;
        status = 1;
    }
    else{
        printf("%s ist keien .osm-Datei!\n",fileName.c_str());
        status = 0;
    }
}
void OSM::initiallists(){
    avail_ID_max = 0;

    maxlat = -90;
    minlat = 90;
    maxlon = -180;
    minlon = 180;
    dmaxlat = -90;
    dminlat = 90;
    dmaxlon = -180;
    dminlon = 180;

    D.nodes = new KOO*;
    D.waysavl = new Way*;
    *(D.nodes) = NULL;
    *(D.waysavl) = NULL;

    for(int i = 0; i <= MaxNature; i++){
        ln[i] = new lines*;
        *ln[i] = NULL;
    }
    for(int i = 0; i <= MaxPoints; i++){
        pt[i] = new points*;
        *pt[i] = NULL;
    }

    woodr = new relation*;
    waterr = new relation*;
    router = new relation*;
    *woodr = *waterr = *router = NULL;
}
int  OSM::get_status(){
    return status;
}
int OSM::count_cross(KOO *k){
    if(k != NULL){
        if(k->cd != NULL)
            return 1+count_cross(k->l)+count_cross(k->r);
        else
            return count_cross(k->l)+count_cross(k->r);
    }
    else
        return 0;
}
int OSM::count_STRONG(KOO *k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->s != NULL)
            return 1+count_STRONG(k->l)+count_STRONG(k->r);
        else
            return count_STRONG(k->l)+count_STRONG(k->r);
    }
    else
        return 0;
}

void OSM::nodehandling(FILE *f,char* line,char* lineattr,char* linename){
    char* attr = &line[0];
    char* name = NULL; char namebuf = '\0';name = &namebuf;
    short foundplace = 0,foundpeak = 0, foundpopulation = 0, foundname = 0, foundcol = 0;
    float hiketype = 0;
    int population = 0;
    int normalname = 0;


    KOO* k;
    k = insert_AVL(extract_valuei(line," id"),D.nodes);
    k->lat = extract_valuef(line," lat");
    k->lon = extract_valuef(line," lon");
    k->cd = NULL;

    while(strstr(line,"</node>") == NULL &&
          !(strstr(line,"<node ")!=NULL && strstr(line,"/>") != NULL)){
                fgets(line,999,f);
                quotationreplace(line);
                if(strstr(line,"<tag k='place' v='") != NULL){
                    strcpy(lineattr,line);
                    attr = extract_valuec(lineattr,"v");
                    foundplace = 1;
                }
                if(strstr(line,"<tag k='tourism' v='alpine_hut'/>") != NULL ||
                   strstr(line,"<tag k='tourism' v='attraction'/>") != NULL){
                    attr = "alpine_hut";
                    foundplace = 1;
                }
                if(strstr(line,"<tag k='natural' v='peak'/>") != NULL){
                    foundpeak = 1;
                }
                if(strstr(line,"<tag k='natural' v='saddle'/>") != NULL ||
                   strstr(line,"<tag k='mountain_pass' v='yes'/>") != NULL){
                    foundcol = 1;
                }
                if(strstr(line,"<tag k='name' v='") != NULL){
                    // primary name source
                    strcpy(linename,line);
                    name = extract_valuec(linename,"v");
                    normalname = 2;
                    foundname = 1;
                }
                if(normalname < 2 &&
                   strstr(line,"<tag k='openGeoDB:name' v='")!=NULL){
                    // secondary name source
                    strcpy(linename,line);
                    name = extract_valuec(linename,"v");
                    normalname = 1;
                    foundname = 1;
                }
                if(normalname < 1 &&
                    strstr(line,"<tag k='name:")!=NULL){
                    // tertiary name source
                    strcpy(linename,line);
                    name = extract_valuec(linename,"v");
                    foundname = 1;
                }
                if(strstr(line,"tag k='population' v='") != NULL){
                    population = extract_valuei(line,"v");
                    foundpopulation = 1;
                }
                check_hike_type(line, &hiketype);

    }
    //falls kein Name gefunden sollte Punkt nicht verzeichnet werden
    if(name[0] == '\0')
        foundplace = foundpeak = 0;
    else{
        htmlreplace(name);
    }

    if(foundpopulation && foundplace && !foundcol)
        //d.h. ein Ort mit Einwohnerzahlangabe ist ausgelesen wurden
        if(population > 0){
            //Klassifizierung nach Einwohnerzahl ist sicherer
            if(population>100000)
                add_points(pt[0],new points(k,name,population));
            else{
                if(population>10000)
                    add_points(pt[1],new points(k,name,population));
                else{
                    if(population>3000 || !strcmp(attr,"town"))
                        add_points(pt[2],new points(k,name,population));
                    else
                        if(population>50)
                            add_points(pt[3],new points(k,name,population));
                        else
                            add_points(pt[4],new points(k,name,population));

                }
            }
        }
    if(foundplace && !foundpopulation  && !foundcol){
            //...als Klassifizierung nach OSM-tags, da keine Einwohnerzahl
            if(!strcmp(attr,"city"))
                add_points(pt[0],new points(k,name,500000));
            else{
                if(!strcmp(attr,"town"))
                    add_points(pt[2],new points(k,name,5000));
                else{
                    if(   !strcmp(attr,"suburb")
                        ||!strcmp(attr,"village"))
                        add_points(pt[3],new points(k,name,500));
                    else
                        if(  !strcmp(attr,"hamlet")
                           ||!strcmp(attr,"locality")
                           ||!strcmp(attr,"alpine_hut")
                           ||!strcmp(attr,"isolated_dwelling"))
                            add_points(pt[4],new points(k,name,20));
                }
            }
    }
    if(foundcol && foundname){
        add_points(pt[5],new points(k,name,20));
    }
    if(foundpeak && foundname && !foundcol)
        add_points(pt[6],new points(k,name,0));
    if(hiketype > 0 && hiketype < 98)
        add_points(pt[7],new points(k,name,hiketype));
}
void OSM::check_hike_type(char* line, float *hiketype){
    if(strstr(line,"<tag k='highway' v='give_way'/>") != NULL)
        *hiketype = 10;
    if(strstr(line,"<tag k='highway' v='stop'/>") != NULL)
        *hiketype = 11;
    if(strstr(line,"<tag k='highway' v='traffic_signals'/>") != NULL)
        *hiketype = 12;
    if(strstr(line,"<tag k='amenity' v='restaurant'/>") != NULL)
        *hiketype = 50;
    if(strstr(line,"<tag k='amenity' v='cafe'/>") != NULL)
        *hiketype = 42;
    if(strstr(line,"<tag k='amenity' v='pub'/>") != NULL  ||
       strstr(line,"<tag k='amenity' v='bar'/>") != NULL  ||
       strstr(line,"<tag k='amenity' v='biergarten'/>") != NULL)
        *hiketype = 43;
    if(strstr(line,"<tag k='amenity' v='food_court'/>") != NULL  ||
       strstr(line,"<tag k='amenity' v='fast_food'/>") != NULL  ||
       strstr(line,"<tag k='amenity' v='ice_cream'/>") != NULL)
        *hiketype = 51;
    if(strstr(line,"<tag k='shop' v='supermarket'/>") != NULL  ||
       strstr(line,"<tag k='shop' v='beverages'/>") != NULL  ||
        strstr(line,"<tag k='shop' v='convenience'/>") != NULL)
        *hiketype = 31;
    if(strstr(line,"<tag k='shop' v='bakery'/>") != NULL  ||
       strstr(line,"<tag k='shop' v='general'/>") != NULL  ||
       strstr(line,"<tag k='shop' v='kiosk'/>") != NULL  ||
       strstr(line,"<tag k='shop' v='pastry'/>") != NULL)
        *hiketype = 32;
    if(strstr(line,"<tag k='amenity' v='fuel'/>") != NULL)
        *hiketype = 30;
    if(strstr(line,"<tag k='information' v='map'/>") != NULL)
        *hiketype = 40;
    if(strstr(line,"<tag k='information' v='board'/>") != NULL)
        *hiketype = 39;
    if(strstr(line,"<tag k='natural' v='cave_entrance'/>") != NULL)
        *hiketype = 61;
    if(strstr(line,"<tag k='tourism' v='viewpoint'/>") != NULL)
        *hiketype = 60;
    if(strstr(line,"<tag k='natural' v='tree'/>") != NULL)
        *hiketype = 35;
    if(strstr(line,"<tag k='waterway' v='waterfall'/>") != NULL)
        *hiketype = 71;
    if(strstr(line,"<tag k='natural' v='spring'/>") != NULL)
        *hiketype = 70;
    if(strstr(line,"tag k='building' v='yes'/>") != NULL||
       strstr(line,"tag k='landuse'") != NULL)
        if(*hiketype < 20)
            *hiketype = 37;
    if(strstr(line,"<tag k='historic' v='memorial'/>") != NULL)
        *hiketype = qMax(*hiketype,float(44.));
    if(strstr(line,"<tag k='generator:source' v='wind'/>") != NULL)
        *hiketype = qMax(*hiketype,float(57.));
    if(strstr(line,"<tag k='historic' v='castle'/>") != NULL)
        *hiketype = qMax(*hiketype,float(45.));
    if(strstr(line,"<tag k='historic' v='ruins'/>") != NULL)
        *hiketype = qMax(*hiketype,float(53.));
    if(strstr(line,"<tag k='museum'") != NULL)
        *hiketype = qMax(*hiketype,float(44.));
    if(strstr(line,"<tag k='amenity' v='place_of_worship'/>") != NULL ||
       strstr(line,"<tag k='building' v='church'/>") != NULL)
        *hiketype = qMax(*hiketype,float(52.));
    if(strstr(line,"<tag k='man_made' v='tower'/>") != NULL)
        *hiketype = qMax(*hiketype,float(47.));
    if(strstr(line,"<tag k='amenity' v='parking'/>") != NULL)
        *hiketype = 21;
    if(strstr(line,"<tag k='highway' v='bus_stop'/>") != NULL ||
       strstr(line,"<tag k='railway' v='tram_stop'/>") != NULL)
        *hiketype = 22;
    if(strstr(line,"<tag k='amenity' v='shelter'/>") != NULL)
        *hiketype = 38;
    if(strstr(line,"<tag k='tourism' v='picnic_site'/>") != NULL ||
       strstr(line,"<tag k='amenity' v='bench'/>") != NULL)
        if(*hiketype != 38)
            *hiketype = 36;
    if(strstr(line,"<tag k='information' v='guidepost'/>") != NULL)
        *hiketype = 64;
    if(strstr(line,"<tag k='barrier' v='gate'/>") != NULL ||
       strstr(line,"<tag k='barrier' v='swing_gate'/>") != NULL ||
       strstr(line,"<tag k='barrier' v='lift_gate'/>") != NULL)
        *hiketype = 62;
    if(strstr(line,"<tag k='power' v='substation'/>") != NULL)
        *hiketype = 98;
}
void OSM::relationhandling(FILE *f,char* line,char* lineattr,char* linename,
                           qint64* currentseek){

    int relationlength = 0;
    int iswood = 0, iswater = 0, isroute = 0;
    qint64 tID;
    Way* tw;
    while(strstr(line,"</relation>") == NULL){
        if(strstr(line,"<member type='way'") != NULL){
            relationlength++;
            tID = extract_valuei(line,"ref");
            tw = search(tID,*(D.waysavl));
            //if(tw != NULL && tw->type == 6)
            //	iswood = 1;
        }
        if(strstr(line,"<tag k='landuse' v='forest'/>") != NULL
        || strstr(line,"<tag k='natural' v='wood'/>") != NULL)
            iswood = 1;
        if(strstr(line,"<tag k='waterWay' v='riverbank'/>") != NULL
        || strstr(line,"<tag k='landuse' v='reservoir'/>") != NULL
        || strstr(line,"<tag k='natural' v='water'/>") != NULL)
            iswater = 1;
        if(strstr(line,"<tag k='route' v='hiking'/>") != NULL
        || strstr(line,"<tag k='route' v='bicycle'/>") != NULL
        //|| strstr(line,"<tag k='type' v='route'/>") != NULL
        || strstr(line,"<tag k='route' v='foot'/>") != NULL)
            isroute = 1;

        fgets(line,999,f);
        quotationreplace(line);
    }

    if ((iswood || iswater || isroute) && relationlength > 0){
        relation* r =new relation;
        r->memberscount = relationlength;
        r->members = new Way*[relationlength];
        _fseeki64(f,*currentseek,SEEK_SET);
        fgets(line,999,f);
        quotationreplace(line);
        char linename[10000];

        relationlength = -1;
        while(strstr(line,"</relation>") == NULL){
            fgets(line,999,f);
            quotationreplace(line);
            if(strstr(line,"<member type='way'") != NULL){
                relationlength++;
                r->members[relationlength] = new Way;
                r->members[relationlength]->id = extract_valuei(line,"ref");
            }
            if(r->name == NULL && strstr(line,"<tag k='name' v='") != NULL){
                // take name only if no ref tag
                strcpy(linename,line);
                char* namet = extract_valuec(linename,"v");
                htmlreplace(namet);
                r->name = new char[strlen(namet)+1];
                strcpy(r->name,namet);
            }
            if(strstr(line,"<tag k='osmc:symbol' v='") != NULL){
                strcpy(linename,line);
                char* namet = extract_valuec(linename,"v");
                namet[3] = '\0';
                if(strstr(namet,"blu")) r->color = blue;
                if(strstr(namet,"yel")) r->color = yellow;
                if(strstr(namet,"gre")) r->color = green;
                if(strstr(namet,"red")) r->color = red;
                if(strstr(namet,"bla")) r->color = black;
            }
            if(strstr(line,"<tag k='route' v='bicycle'/>") != NULL){
                r->color = pink;
            }
        }
        if(iswood){
            add_relation(woodr,r);
            r->type = Woodi;
        }
        if(iswater){
            add_relation(waterr,r);
            r->type = Wateri;
        }
        if(isroute){
            if(r->name != NULL && strstr(r->name,"Dresdner Heide, ") == r->name){
                if(r->color == red) r->color = none;
                r->name = &(r->name[16]);
            }
            add_relation(router,r);
            r->type = 9;
        }
    }
}
void OSM::Wayhandling(FILE *f,char* line,char* lineattr,char* linename,
                      qint64* currentseek){
    char* attr = &line[0];
    char* name = NULL;
    float hiketyp;
    int Waylength = 0;
    while(strstr(line,"</way>") == NULL){
        if(strstr(line,"<nd ref=") != NULL) Waylength++;
        fgets(line,999,f);
        quotationreplace(line);
        if(name == NULL && strstr(line,"<tag k='name' v='") != NULL){
            // take name only if no ref tag
            strcpy(linename,line);
            name = extract_valuec(linename,"v");
            htmlreplace(name);
        }
        if(strstr(line,"<tag k='ref' v='") != NULL){
            // take re
            strcpy(linename,line);
            name = extract_valuec(linename,"v");
            htmlreplace(name);
        }
    }

    if (Waylength > 0){
        _fseeki64(f,*currentseek,SEEK_SET);	//fur id neu lesen
        fgets(line,999,f);
        quotationreplace(line);
        Way* k = insert_AVL(extract_valuei(line,"id"),D.waysavl);
        k->nodec = Waylength;
        k->node = new KOO*[Waylength];
        k->type = 98;					//falls sich erweisen sollte, dass der
                                        //Way ungenutzt bleibt, muss er beim
                                        //speichern der fertig berechneten
                                        //Daten verworfen werden
        Waylength = -1;
        int smoothness = 0;
        float width = -1;
        while(strstr(line,"</way>") == NULL){
            fgets(line,999,f);
            quotationreplace(line);
            if(strstr(line,"<nd ref=") != NULL){
                Waylength++;
                k->node[Waylength] = new KOO;
                k->node[Waylength]->id = extract_valuei(line,"ref");
            }
            if(strstr(line,"<tag k='highway' v='") != NULL ||
               strstr(line,"<tag k='construction' v='") != NULL){
                k->type = qMin(k->type,(char)9);
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");
                if(!strcmp(attr,"motorway")||
                   !strcmp(attr,"motorway_link")||
                   !strcmp(attr,"trunk")||
                   !strcmp(attr,"trunk_link"))
                    k->type = qMin(k->type,(char)0);
                if(!strcmp(attr,"primary")||
                   !strcmp(attr,"primary_link"))
                    k->type = qMin(k->type,(char)1);
                if((!strcmp(attr,"secondary")
                    ||!strcmp(attr,"secondary_link")))
                    k->type = qMin(k->type,(char)2);
                if(!strcmp(attr,"tertiary")
                    ||!strcmp(attr,"tertiary_link"))
                    k->type = qMin(k->type,(char)3);
                if(!strcmp(attr,"residential")
                    ||!strcmp(attr,"unclassified")
                    ||!strcmp(attr,"road")
                    ||!strcmp(attr,"construction")
                    ||!strcmp(attr,"service"))
                    k->type = qMin(k->type,(char)4);
                if( !strcmp(attr,"cycleway"))
                    k->type = qMin(k->type,(char)5);
                if( !strcmp(attr,"steps"))
                    k->IsStair = true;
                if(!strcmp(attr,"path")||
                   !strcmp(attr,"footway"))
                    k->type = qMin(k->type,(char)9);
            }
            if(strstr(line,"<tag k='surface' v='") != NULL){
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");
                if(!strcmp(attr,"asphalt")||
                   !strcmp(attr,"concrete")||
                   !strcmp(attr,"concrete:lanes")||
                   !strcmp(attr,"concrete:plates")||
                   !strcmp(attr,"paving_stones ")||
                   !strcmp(attr,"paved")||
                   !strcmp(attr,"sett"))
                    k->type = qMin(k->type,(char)4);
                if(!strcmp(attr,"compacted")||
                   !strcmp(attr,"fine_gravel"))
                    k->type = (char)5;
                if(!strcmp(attr,"cobblestone:flattened"))
                    k->type = qMin(k->type,(char)4);
                if(!strcmp(attr,"cobblestone"))
                    k->type = qMin(k->type,(char)4);
                if(!strcmp(attr,"earth") ||
                   !strcmp(attr,"ground")||
                   !strcmp(attr,"dirt"))
                    smoothness = 1;
                if(!strcmp(attr,"sand") ||
                   !strcmp(attr,"gravel") ||
                   !strcmp(attr,"mud"))
                    smoothness = -1;
            }
            if(strstr(line,"<tag k='smoothness' v='") != NULL){
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");
                if(!strcmp(attr,"excellent")||
                   !strcmp(attr,"good"))
                   smoothness = 1;
                if(!strcmp(attr,"intermediate"))
                   smoothness = 0;
                if(!strcmp(attr,"bad")||
                    !strcmp(attr,"very_bad")||
                    !strcmp(attr,"horrible")||
                    !strcmp(attr,"very_horrible")||
                    !strcmp(attr,"impassable"))
                   smoothness = -1;
            }
            if(strstr(line,"<tag k='mtb:scale' v='") != NULL){
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");
                if(!strcmp(attr,"0"))
                    smoothness = qMax(smoothness,0);
                if(!strcmp(attr,"1") ||
                   !strcmp(attr,"2") ||
                   !strcmp(attr,"3") ||
                   !strcmp(attr,"4") ||
                   !strcmp(attr,"0") ||
                   !strcmp(attr,"0"))
                    smoothness = qMin(smoothness,-1);
            }
            if(strstr(line,"<tag k='tracktype' v='") != NULL){
                strcpy(lineattr,line);
                attr = extract_valuec(lineattr,"v");
                if(!strcmp(attr,"grade1"))
                    k->type = qMin(k->type,(char)4);
                if(!strcmp(attr,"grade2"))
                    k->type = qMin(k->type,(char)5);
                if(!strcmp(attr,"grade3"))
                    k->type = qMin(k->type,(char)6);
                if(!strcmp(attr,"grade4"))
                    k->type = qMin(k->type,(char)7);
                if(!strcmp(attr,"grade5"))
                    k->type = qMin(k->type,(char)7);
            }
            if(strstr(line,"<tag k='width' v='") != NULL){
                strcpy(lineattr,line);
                width = extract_valuef(lineattr," v");
            }
            if(strstr(line,"<tag k='tunnel' v='yes'/>") != NULL)
                k->IsTunnel = true;
            if(strstr(line,"<tag k='bridge' v='") != NULL)
                k->IsBridge = true;
            if(strstr(line,"<tag k='priority_road' v='designated'")!= NULL
             ||strstr(line,"<tag k='priority_road' v='yes") != NULL)
                k->IsPriorityRoad = true;
            if((strstr(line,"<tag k='landuse' v='forest'/>") != NULL
             || strstr(line,"<tag k='natural' v='wood'/>") != NULL))
                k->type = qMin(k->type,(char)Woodi);
            if((strstr(line,"<tag k='waterway' v='riverbank'/>") != NULL
            ||  strstr(line,"<tag k='landuse' v='reservoir'/>") != NULL
            ||  strstr(line,"<tag k='natural' v='water'/>") != NULL))
                k->type = qMin(k->type,(char)Wateri);
            if(strstr(line,"<tag k='waterway' v='river'/>") != NULL)
                k->type = qMin(k->type,(char)Riveri);
            if(strstr(line,"<tag k='waterway' v='stream'/>") != NULL)
                k->type = qMin(k->type,(char)Creeki);
            if(strstr(line,"<tag k='waterway' v='stream'/>") != NULL)
                k->type = qMin(k->type,(char)Creeki);
            if((strstr(line,"<tag k='railway' v='rail'/>") != NULL
            ||  strstr(line,"<tag k='railway' v='preserved'/>") != NULL
            ||  strstr(line,"<tag k='railway' v='narrow_gauge'/>") != NULL
            ||  strstr(line,"<tag k='railway' v='disused'/>") != NULL
            ||  strstr(line,"<tag k='railway' v='monorail'/>") != NULL
            ||  strstr(line,"<tag k='railway' v='construction'/>") != NULL))
                k->type = qMin(k->type,(char)Raili);
            check_hike_type(line,&hiketyp);
            if(strstr(line,"<tag k='path' v='climbing_access'/>") != NULL){
                k->add_color(black);
            }
        }
        // fine tuning with help of smoothness parameter
        if(width > 2.5 && k->type == Creeki)
            k->type = Riveri;
        switch(k->type){
            case 5:
            case 6:
            case 7:
                k->type = qMin(7,qMax(5,k->type-smoothness));
                break;
            case 8:
            case 9:
            case 10:
                k->type = qMin(10,qMax(8,k->type-smoothness));
                break;
        }
        // add to line array
        if(k->type <= MaxNature)
            add_line(ln[k->type],k,name);
        else
            if(hiketyp >= 20 && hiketyp < 90){
                k->type = char(hiketyp);
                k->set_name(name);
            }
    }
}
void OSM::add_line(lines** l,Way* Way,char* name){
    lines* a = new lines;
    a->way = Way;
    if(*l == NULL) a->next = NULL;
    else a->next = *l;
    *l = a;
    if(name != NULL){
        Way->name = a->name = new char[strlen(name)+1];
        strcpy(a->name,name);
    }
    else
        a->name = Way->name = NULL;
}
points* OSM::add_points(points** dest, points *src){
    points* p = *dest;
    src->k->status = 3;
    if(p == NULL)
        *dest = src;
    else
        *dest = p->insert_order_increasing(src);
    return src;
}
void OSM::add_relation(relation** l, relation* r){
    r->next = *l;
    *l = r;
}
void OSM::invert_points(points** l){
    points* b = *l;
    points* p = *l;
    while(b != NULL){
        p = b;
        b = p->next;
        p->next = p->prev;
        p->prev = b;
    }
    *l = p;
}
void OSM::RelationstoWay(relation** l){
    relation* r = *l;
    Way* w;
    char buf = '\0';
    while(r != NULL){
        for(int i = 0; i < r->memberscount; i++){
            w = search(r->members[i]->id,*(D.waysavl));
            if(w != NULL && w->type != 99){
                w->add_color(r->color);
                if(r->name != NULL){
                    if(w->name == NULL) w->name = &buf;
                    // append path name
                    char* newname = new char[strlen(r->name)+strlen(w->name)+1+2];
                    strcpy(newname,w->name);
                    newname[strlen(w->name)] = ';';
                    newname[strlen(w->name)+1] = ' ';
                    strcpy(&(newname[strlen(w->name)+2]),r->name);
                    w->name = newname;
                 }
            }
        }
        r = r->next;
    }
}

void OSM::RelationstoFace(relation** l,lines** f){
    relation* r = *l;
    short *used, exists,foundneighbourmember;
    used = new short;
    lines* cf;
    int i,j,k,cm,cn;
    Way** Ways = new Way*;
    Way* grandWay;
    Way* tempWay;
    int nodecount;
    while(r != NULL){
        delete used;
        KOO *e,*s;
        used = new short[r->memberscount];
        for(i = 0; i < r->memberscount; i++){
            used[i] = 0;
            tempWay = search(r->members[i]->id,*(D.waysavl));
            if(tempWay != NULL && tempWay->type != 99){
                /*der Weg existiert und kennt alle seine nodes*/
                r->members[i] = tempWay;
                used[i] = 0;
            }
            else
                used[i] = 1;
        }
        for(i = 0; i < r->memberscount; i++)
            if(!used[i]){
                if(r->members[i]->node[0					  ]->id ==
                   r->members[i]->node[r->members[i]->nodec-1]->id){
                    /*falls der Member geschlossen ist, dann pruefen, ob er in
                    f schon existiert, falls nicht und der Weg noch keinen Typ
                    hat soll er in f eingefuegt werden*/
                    exists = 0;
                    cf = *f;
                    while(cf != NULL){
                        if(cf->way->id == r->members[i]->id){
                            exists = 1;
                            cf = NULL;
                        }
                        else
                            cf = cf->next;
                    }
                    if(!exists && r->members[i]->type > MaxNature){
                        add_line(f,r->members[i],NULL);
                        r->members[i]->type = r->type;
                        if(r->name == NULL || r->name[0] == '\0')
                            r->name = r->name;
                    }
                    used[i] = 1;
                }
                else{
                    /*falls der Member nicht geschlossen ist muss versucht
                    werden, mit Hilfe der anderen Member eine geschlossene
                    kurve zu konzentrieren. Members, die dabei helfen, liegen
                    im Array Ways*/
                    delete(Ways);
                    Ways = new Way*[r->memberscount];
                    cm = 0;
                    e = r->members[i]->node[0];
                    s = r->members[i]->node[r->members[i]->nodec-1];
                    Ways[cm]=r->members[i];
                    used[i] = 1;
                    do{
                        foundneighbourmember = 0;
                        for(j = 0; j < r->memberscount; j++){
                            if(!used[j]
                            &&(  s->id == r->members[j]->node[0]->id
                               ||s->id == r->members[j]->node[r->members[j]->nodec-1]->id)
                               ){
                                /*falls der Weg j an den cm-Weg anknuepft*/
                                cm++;
                                Ways[cm] = r->members[j];
                                if(s->id == r->members[j]->node[0]->id)
                                    s = r->members[j]->node[r->members[j]->nodec-1];
                                else
                                    s = r->members[j]->node[0];
                                used[j] = 1;
                                j = r->memberscount;
                                foundneighbourmember = 1;
                            }
                        }
                    }while(    foundneighbourmember
                            && (e->id != Ways[cm]->node[0]				  ->id)
                            && (e->id != Ways[cm]->node[Ways[cm]->nodec-1]->id));
                    /*falls es gelungen ist, Members zu einer geschlossenen
                    Kurve zu gruppieren, soll nun diese Kurve erstellt werden
                    und in f abgelegt werden*/
                    if(foundneighbourmember){
                        /*suche nach Weg unter Ways, der exklusiv von Relation
                        genutzt wird*/
                        int FoundExclusive = 1;
                        int CheckMax = Wateri; if(l == woodr) CheckMax = Woodi;
                        for(j = 0; j <= cm; j++){
                            FoundExclusive = 1;
                            for(int k = 0; k < CheckMax; k++){
                                cf = *(ln[k]);
                                while(cf != NULL){
                                    if(Ways[j]->id == cf->way->id){
                                        FoundExclusive = -1;
                                        cf = NULL;
                                        k = CheckMax;
                                    }
                                    else
                                        cf = cf->next;
                                }
                            }
                            if(FoundExclusive == 1){
                                FoundExclusive = j;
                                j = cm + 1;
                            }
                        }
                        if(FoundExclusive > -1){
                            /*es gibt einen exklusiven Weg der zum verknuepfen
                            genutzt werden kann*/
                            nodecount = 0;
                            for(j = 0; j <= cm; j++)
                                nodecount = nodecount + Ways[j]->nodec;
                            nodecount = nodecount - cm;
                            grandWay = new Way;

                            /*grandWay intialisieren*/
                            grandWay->balance = r->members[i]->balance;
                            grandWay->id = Ways[FoundExclusive]->id;
                            grandWay->nodec = nodecount;
                            grandWay->node =new KOO*[nodecount];
                            grandWay->type = r->type;

                            cn = 0;
                            s = r->members[i]->node[0];

                            grandWay->maxlat = -90;
                            grandWay->minlat = 90;
                            grandWay->maxlon = -180;
                            grandWay->minlon = 180;
                            for(j = 0; j <= cm; j++){
                                if(Ways[j]->node[0]->id == s->id){
                                    /*vorwaerts fuellen*/
                                    for(k = 0; k < Ways[j]->nodec-1; k++){
                                        grandWay->node[cn] = Ways[j]->node[k];
                                        cn++;
                                    }
                                    s = Ways[j]->node[Ways[j]->nodec-1];
                                }
                                else{
                                    /*rueckwaerts fuellen*/
                                    for(k = Ways[j]->nodec-1; k > 0; k--){
                                        grandWay->node[cn] = Ways[j]->node[k];
                                        cn++;
                                    }
                                    s = Ways[j]->node[0];
                                }
                                /*umrandendes Rechteck bestimmen*/
                                if(Ways[j]->maxlat > grandWay->maxlat )
                                    grandWay->maxlat = Ways[j]->maxlat;
                                if(Ways[j]->minlat < grandWay->minlat )
                                    grandWay->minlat = Ways[j]->minlat;
                                if(Ways[j]->maxlon > grandWay->maxlon )
                                    grandWay->maxlon = Ways[j]->maxlon;
                                if(Ways[j]->minlon < grandWay->minlon )
                                    grandWay->minlon = Ways[j]->minlon;
                            }
                            grandWay->node[grandWay->nodec-1]=grandWay->node[0];
                            add_line(f,grandWay,NULL);
                        }
                    }
                }
            }
        r = r->next;
    }
    delete(used);
}
void OSM::reset_KOO_status(KOO* k,char i){
    if(k != NULL){
        k->status = i;
        reset_KOO_status(k->l,i);
        reset_KOO_status(k->r,i);
    }
}
void OSM::calc_brect(){
    maxlat = -90;
    minlat = 90;
    maxlon = -180;
    minlon = 180;
    for(int i = 0; i <= MaxNature; i++)
        calc_brect(ln[i]);
    for(int i = 0; i <= MaxPoints; i++)
        calc_brect(pt[i]);
}
void OSM::calc_brect(points** l){
    points* p = *l;
    while(p != NULL){
        if(p->k->lat > maxlat) maxlat = p->k->lat;
        if(p->k->lat < minlat) minlat = p->k->lat;
        if(p->k->lon > maxlon) maxlon = p->k->lon;
        if(p->k->lon < minlon) minlon = p->k->lon;
        p = p->next;
    }
}
void OSM::calc_brect(lines** l){
    lines* p = *l;
    while(p != NULL){
        for(int i = 0; i < p->way->nodec; i++){
            if(p->way->node[i]->lat > maxlat) maxlat = p->way->node[i]->lat;
            if(p->way->node[i]->lat < minlat) minlat = p->way->node[i]->lat;
            if(p->way->node[i]->lon > maxlon) maxlon = p->way->node[i]->lon;
            if(p->way->node[i]->lon < minlon) minlon = p->way->node[i]->lon;
        }
        p = p->next;
    }
}
Way* OSM::get_first_Way(){
    return *(D.waysavl);
}
KOO* OSM::get_first_KOO(){
    return *(D.nodes);
}
Neighbour* OSM::add_Neighbour(KOO* k, KOO** node, KOO* nb, Way* w, float d,
                        int KOOc, readdirection rd){
    Neighbour *n;
    if(k->cd == NULL){ alloc_crossdata(k); k->status = 2;}
    if(nb->cd == NULL){ alloc_crossdata(nb); nb->status = 2;}
    if( k->cd->neighbours == NULL){
        k->cd->neighbours =new Neighbour;
        n = k->cd->neighbours;
        n->next = NULL;
    }
    else{
        n = k->cd->neighbours;
        while(n->next != NULL && (n->nb != nb || n->way != w))
            /*Eine Nachbarbeziehung ist dann identisch, wenn die 2 Kreuzungen
            und der benutzte Weg übereinstimmen*/
            n = n->next;

        if(n->nb != nb || n->way != w){
            /*Die Nachbarbeziehung war ueberhaupt noch nicht vorhanden*/
            n->next =new Neighbour;
            n = n->next;
            n->next = NULL;
        }
        else{
            if(d < n->d){
                /*Die Nachbarbeziehung war mit einer laengeren Distanz
                vorhanden und sollte ueberschrieben werden*/
                if(n->tp != NULL){
                    delete(n->tp);
                    n->tp = NULL;
                    n->brthr->tp = NULL;
                }

            }
            else
                /*die bestehende Nachbarbeziehung hat einen kuerzeren Weg,
                einfuegen abbrechen*/
                return NULL;
        }
    }
    n->d = d;
    n->hm = 0;
    n->t = d;
    n->nb = nb;
    n->brthr = NULL;
    n->way = w;
    n->nodec = KOOc;
    n->node = node;
    n->rd = rd;
    n->tpc = 0;
    n->tp = NULL;
    n->tr = 0;
    k->status = 2;
    return n;
}
double OSM::getminlat(){
    return minlat;
}
double OSM::getminlon(){
    return minlon;
}
double OSM::getmaxlat(){
    return maxlat;
}
double OSM::getmaxlon(){
    return maxlon;
}

void OSM::calc_peak_index_pro(HeightData* h){
    points* p = *(pt[6]);
    int wc = 50, j, i, ac = 10;
    float h0;
    float d = float(0.2);

    float* index = new float[ac];
    float* index2 = new float[ac];
    float is,is2;
    float Ahigher;
    qint64**c = new qint64*[ac];
    for(i = 0; i < ac; i++)
        c[i] = new qint64[wc];
    float mperlon, mperlat;
    while(p != NULL){
        mperlon = 2*M_PI*6371/360*cos(p->k->lat/180.*M_PI) * 1000/1200;
        mperlat = 2*M_PI*6371/360 * 1000/1200;
        h0 = h->height_t(p->k->lon,p->k->lat,NULL,0);
        h->paek_pro(p->k->lon,p->k->lat,c,&Ahigher,wc,ac,d,mperlon,mperlat);
        for(i = 0; i < ac; i++){
            index[i] = 0;
            index2[i] = 0;
            /*Berechnet Volumen*/
            if( h0 > c[i][0])
                for(j = 1; j < wc; j++)
                    if(c[i][j]<c[i][j-1])
                        index[i] = index[i] + (h0 - c[i][j]);
                    else
                        j = wc;
            /*Berechnet Bergformel*/
            for(j = 0; j < wc; j++)
                if( h0 > c[i][j]
                    &&(h0-c[i][j])*(h0-c[i][j])/(i*d*1000) > index2[i])
                    index2[i] = float((h0-c[i][j])*(h0-c[i][j])/((0.5+j)*d*1000));
        }
        is = 0;
        is2 = 0;
        p->index = 0;
        for(j = 0; j < ac; j++){
            is  = is  + sqrt(index[j]);
            is2 = is2 + index2[j];
        }
        p->index = sqrt(is*is*is2/ac);
        p->index = float(p->index * max(1.,1+(0.002-Ahigher)*500));

        p = p->next;
    }

    for(i = 0; i < ac; i++)
        delete(c[i]);
    delete(c);
    delete(index);
    delete(index2);
}
void OSM::sort_peaks(){
    short change = 1;
    points p,*l = *(pt[6]);
    if(l != NULL){
        p.next = l;
        l->prev = &p;
        FILE* f;
        f = fopen("pointssortiert.txt","w");
        while(change){
            change = 0;
            l = p.next;
            while(l->next != NULL){
                /*a-b-c-d ==> a-c-b-d*/
                if(l->index < l->next->index){
                    l->prev->next = l->next;//a->c
                    l->next->prev = l->prev;//a<-c
                    l->prev	= l->next;		//c<-b
                    if(l->next->next != NULL){
                        l->next->next->prev = l;//b<-d
                        l->next = l->next->next;//b->d;
                    }
                    else
                        l->next = 0;
                    l->prev->next = l;		//c->b;
                    change = 1;
                }
                else
                    l = l->next;
            }
        }
        l = p.next;
        while(l != NULL){
            fprintf(f,"%f - %s - %f N %f E\n",l->index,l->name,l->k->lat,l->k->lon);
            l = l->next;
        }

        fclose(f);

        p.next->prev = NULL;				//Abkoppeln vom Pufferelement
        *(pt[6]) = p.next;						//Listenanfang setzen
    }
}
void OSM::save(char* filename, double minlat, double maxlat,
               double minlon, double maxlon){
    FILE* f;
    f = fopen(filename,"wb");
    if(	dmaxlat == -90
        && dminlat == 90
        && dmaxlon == -180
        && dminlon == 180){
        dmaxlat = maxlat;
        dminlat = minlat;
        dmaxlon = maxlon;
        dminlon = minlon;
    }

    fwrite(&minlat,sizeof(double),1,f);
    fwrite(&maxlat,sizeof(double),1,f);
    fwrite(&minlon,sizeof(double),1,f);
    fwrite(&maxlon,sizeof(double),1,f);
    for(int i = 0; i <= MaxNature; i++)
        save_lines(ln[i],f,minlat,maxlat,minlon,maxlon);

    // peak list is to be inverted to be loaded faster
    invert_points(pt[6]);
    for(int i = 0; i <= MaxPoints; i++)
        save_points(pt[i],f,minlat,maxlat,minlon,maxlon);
    invert_points(pt[6]);

    /*speichere Kreuzungsdaten ab, vorher Kreuzungen durchzählen, nachher
    diese Anzahl bei currentseek speichern*/
    int c = 0;
    qint64 currentseek = _ftelli64(f);
    fwrite(&c,sizeof(int),1,f);
    save_cross(*D.nodes, f, &c, minlat, maxlat, minlon, maxlon);
    _fseeki64(f,currentseek,SEEK_SET);
    fwrite(&c,sizeof(int),1,f);

    fclose(f);
    reset_KOO_status(*D.nodes);
}
void OSM::save_lines(lines** l, FILE* f, double minlat, double maxlat,
                    double minlon, double maxlon){
    int c = 0;
    int len;
    if(*l == NULL)
        fwrite(&c,sizeof(int),1,f);
    else{
        lines*  k = *l;
        while(k != NULL){
            if(isin(k->way,minlat,maxlat,minlon,maxlon))
                c++;
            k = k->next;
        }
        fwrite(&c,sizeof(int),1,f);
        k = *l;
        while(k != NULL){
            if(isin(k->way,minlat,maxlat,minlon,maxlon)){
                save_Way(k->way,f);
                len = 0;
                if(k->name == NULL && k->way->name != NULL)
                    k->name = k->way->name;
                if(k->name == NULL){
                    fwrite(&len,sizeof(int),1,f);
                }
                else{
                    len = strlen(k->name);
                    fwrite(&len,sizeof(int),1,f);
                    fwrite(k->name,sizeof(char),len,f);
                }
            }
            k = k->next;
        }
    }
}
void OSM::save_points(points** p, FILE* f, double minlat, double maxlat,
                    double minlon, double maxlon){
    int c = 0;
    int len;
    if(*p == NULL)
        fwrite(&c,sizeof(int),1,f);
    else{
        points*  k = *p;
        while(k != NULL){
            if(isin(k->k,minlat,maxlat,minlon,maxlon))
                c++;
            k = k->next;
        }
        fwrite(&c,sizeof(int),1,f);
        k = *p;
        while(k != NULL){
            if(isin(k->k,minlat,maxlat,minlon,maxlon)){
                save_node(k->k,f);
                fwrite(&(k->index),sizeof(float),1,f);
                len = 0;
                if(k->name == NULL){
                    fwrite(&len,sizeof(int),1,f);
                }
                else{
                    len = strlen(k->name);
                    fwrite(&len,sizeof(int),1,f);
                    fwrite(k->name,sizeof(char),len,f);
                }
            }
            k = k->next;
        }
    }
}
void OSM::save_node(KOO* k, FILE* f){
    fwrite(&(k->id),sizeof(qint64),1,f);
    fwrite(&(k->lon),sizeof(double),1,f);
    fwrite(&(k->lat),sizeof(double),1,f);
    if(k->cd != NULL)
        /*markiere den Node als noch zu speichernde Kreuzung*/
        k->status = 10;
}
void OSM::save_Way(Way* w, FILE* f){
    fwrite(&(w->IsBridge),sizeof(bool),1,f);
    fwrite(&(w->IsTunnel),sizeof(bool),1,f);
    fwrite(&(w->IsStair),sizeof(bool),1,f);
    fwrite(&(w->IsOneWay),sizeof(bool),1,f);
    fwrite(&(w->waycolors),sizeof(char),6,f);
    fwrite(&(w->id),sizeof(qint64),1,f);
    fwrite(&(w->type),sizeof(char),1,f);
    fwrite(&(w->nodec),sizeof(int),1,f);
    for(int i = 0; i < w->nodec; i++)
        save_node(w->node[i],f);
}
void OSM::save_cross(KOO* k, FILE* f, int* c, double minlat,
                     double maxlat, double minlon, double maxlon){
    if(k != NULL){
        if(k->status == 10
           && isin(k,dminlat,dmaxlat,dminlon,dmaxlon)){
            (*c)++;
            fwrite(&(k->id),sizeof(qint64),1,f);
            /*Anzahl i der Nachbarkreuzungen bestimmen*/
            short i = 0;
            Neighbour* n = k->cd->neighbours;
            while(n != NULL){
                /*gespeichert werden immer nur die VorwärtsNachbarrelationen,
                sonst ist das gespeichert redundant. Beim Laden werden dann
                beider Kreuzungen befuellt werden*/
                if(n->nb->status == 10 && n->rd == fw
                   &&isin(n->nb,dminlat,dmaxlat,dminlon,dmaxlon))
                    i++;
                n = n->next;
            }
            fwrite(&i,sizeof(short),1,f);

            n = k->cd->neighbours;
            while(n != NULL){
                if(n->nb->status == 10 && n->rd == fw
                   && isin(n->nb,dminlat,dmaxlat,dminlon,dmaxlon)){
                    fwrite(&(n->rd),		sizeof(readdirection),1,f);
                    fwrite(&(n->d),			sizeof(float),1,f);
                    fwrite(&(n->hm),		sizeof(float),1,f);
                    fwrite(&(n->nb->id),	sizeof(qint64),1,f);
                    fwrite(&(n->way->id),	sizeof(qint64),1,f);
                    fwrite(&(n->nodec),		sizeof(int),1,f);
                    fwrite(&(n->tr),		sizeof(char),1,f);
                    fwrite(&(n->brthr->tr),	sizeof(char),1,f);
                    int j = 0;
                    for(int l = n->way->nodec-1; l >=0; l--){
                        if(n->way->node[l] == k)
                            j = l;			//findet zugehoerigen Index in der
                    }						//KOO-Liste des Weges
                    fwrite(&j,sizeof(int),1,f);
                    fwrite(&(n->tpc),		sizeof(int),1,f);
                    fwrite(n->tp,			sizeof(trackpoint),n->tpc,f);
                }
                n = n->next;
            }
        }
        /*Rekursionsaufruf*/
        save_cross(k->l,f,c, minlat, maxlat, minlon, maxlon);
        save_cross(k->r,f,c, minlat, maxlat, minlon, maxlon);
    }
}
void OSM::reset_KOO_status(KOO* k){
    if(k != NULL){
        if(k->status == 10)
            k->status = 2;
        /*rekursiver Aufruf*/
        reset_KOO_status(k->l);
        reset_KOO_status(k->r);
    }
}
void OSM::load(char* filename){
    FILE* f;
    f = fopen(filename,"rb");
    if(f != NULL){
        double minlat, maxlat, minlon, maxlon;
        fread(&minlat,sizeof(double),1,f);
        fread(&maxlat,sizeof(double),1,f);
        fread(&minlon,sizeof(double),1,f);
        fread(&maxlon,sizeof(double),1,f);

        if(minlat < this->minlat) this->minlat = minlat;
        if(maxlat > this->maxlat) this->maxlat = maxlat;
        if(minlon < this->minlon) this->minlon = minlon;
        if(maxlon > this->maxlon) this->maxlon = maxlon;
        for (int i = 0; i <= MaxNature; i++)
            load_lines(ln[i],f);
        invert_points(pt[7]);           // to achieve comp. add_points function
        for (int i = 0; i <= MaxPoints; i++)
            load_points(pt[i],f);
        invert_points(pt[7]);           // reestablish increasing order
        int c;
        fread(&c,sizeof(int),1,f);
        for(int i = 0; i < c; i++)
            load_cross(f);
        fclose(f);
        //QMessageBox(QMessageBox::NoIcon,"OSMende","equal heights follow...").exec();

        equal_crossheight(get_first_KOO());
        //QMessageBox(QMessageBox::NoIcon,"OSMende","equal heights finished.").exec();
        status = 3;
    }
}
KOO* OSM::load_node(FILE* f){
    qint64 id;
    fread(&id,sizeof(qint64),1,f);
    KOO* k = insert_AVL(id,D.nodes);
    if(k->lat == 200){
        fread(&(k->lon),sizeof(double),1,f);
        fread(&(k->lat),sizeof(double),1,f);
        k->status = 0;
        return k;
    }
    else{
        fseek(f,2*sizeof(double),SEEK_CUR);
        return k;
    }
}
Way* OSM::load_Way(FILE* f){
    qint64 id;
    int length,i;
    char type,colors[6],bridge,tunnel,stairs,oneway;
    Way* w;
    fread(&bridge,sizeof(bool),1,f);
    fread(&tunnel,sizeof(bool),1,f);
    fread(&stairs,sizeof(bool),1,f);
    fread(&oneway,sizeof(bool),1,f);
    fread(&colors,sizeof(char),6,f);
    fread(&id,sizeof(qint64),1,f);
    fread(&type,sizeof(char),1,f);
    fread(&length,sizeof(int),1,f);
    w = insert_AVL(id,D.waysavl);
    if(w->type == -1 || w->nodec < length){
        /*lesen und fuellen, da noch nicht vorhanden oder kürzer*/
        KOO** newKOO = new KOO*[length];
        for(i = 0; i < length; i++)
            newKOO[i] = load_node(f);

        if(w->type != -1){
            replace_Way_KOO(w,newKOO,length);
            w = NULL;					//Der Weg war schon vorhanden, wurde
                                        //nur geaendert, deshalb return NULL
                                        //da OSM::load_lines damit arbeitet
        }
        else{
            w->IsTunnel = tunnel;
            w->IsBridge = bridge;
            w->IsStair = stairs;
            w->IsOneWay = oneway;
            w->waycolors[0] = colors[0];
            w->waycolors[1] = colors[1];
            w->waycolors[2] = colors[2];
            w->waycolors[3] = colors[3];
            w->waycolors[4] = colors[4];
            w->waycolors[5] = colors[5];
            w->type = type;
            w->nodec = length;
            w->node = newKOO;
            w->maxlat = -90;
            w->minlat =  90;
            w->maxlon =-180;
            w->minlon = 180;
            for(i = 0; i < length; i++){
                if(w->node[i]->lat > w->maxlat) w->maxlat = float(w->node[i]->lat);
                if(w->node[i]->lat < w->minlat) w->minlat = float(w->node[i]->lat);
                if(w->node[i]->lon > w->maxlon) w->maxlon = float(w->node[i]->lon);
                if(w->node[i]->lon < w->minlon) w->minlon = float(w->node[i]->lon);
            }
        }
    }
    else{
        fseek(f,length*(2*sizeof(double)+sizeof(qint64)),SEEK_CUR);
        w = NULL;						//auf keinen Fall neues Line
    }
    return w;
}
void OSM::load_cross(FILE* f){
    readdirection rd;
    float d,hm;
    qint64 id, nid, wid;
    int j,tpc,nodec;
    char tr_fw, tr_bw;
    short c;
    Neighbour *n1, *n2;
    trackpoint *tp;
    fread(&id, sizeof(qint64),1,f);
    fread(&c,sizeof(short),1,f);
    Way* w;
    KOO* k = search(id,*D.nodes), *kn;
    if(k != NULL){
        if(k->cd == NULL){
            /*Kreuzungsdaten anlegen, falls noch nicht vorhanden*/
            alloc_crossdata(k);
        }

        /*Alle Kreuzungsabgänge bearbeiten*/
        for(short i = 0; i < c; i++){
            /*einlesen*/

            fread(&(rd),	sizeof(readdirection),1,f);
            fread(&(d),		sizeof(float),1,f);
            fread(&(hm),	sizeof(float),1,f);
            fread(&(nid),	sizeof(qint64),1,f);
            fread(&(wid),	sizeof(qint64),1,f);
            fread(&(nodec),	sizeof(int),1,f);
            fread(&(tr_fw),	sizeof(char),1,f);
            fread(&(tr_bw),	sizeof(char),1,f);
            fread(&j,		sizeof(int),1,f);
            fread(&(tpc),	sizeof(int),1,f);

            kn = search(nid,*D.nodes);
            w = search(wid,*D.waysavl);

            if(kn != NULL && w != NULL && j+(nodec-1) < w->nodec){
                /*Achtung: bei verschieden alte OSM-DAtensätze gleichen sich
                die Wege nicht unbedingt aufgrund der Gleichheit ihrer ID.
                Im Falle der verschiedenheit muss wenigstens gewährleistet,
                dass Neighbour nicht über das nodearray hinausliest*/
                if(kn->cd == NULL){
                    /*Kreuzungsdaten anlegen, falls noch nicht vorhanden*/
                    alloc_crossdata(kn);
                }
                n1 = add_Neighbour(k ,&(w->node[j]),kn,w,d,nodec,fw);
                n2 = add_Neighbour(kn,&(w->node[j]),k ,w,d,nodec,bw);
                if(n1 != NULL && n2 != NULL){
                    /*Normalfalll: hin und rückweg existieren noch nicht*/
                    tp = new trackpoint[tpc];
                    fread(tp,sizeof(trackpoint),tpc,f);
                    n1->tp = tp;
                    n1->tpc = tpc;
                    n1->brthr = n2;
                    n1->hm = hm;
                    n1->tr = tr_fw;
                    n2->tp = tp;
                    n2->tpc = tpc;
                    n2->brthr = n1;
                    n2->hm = qMax(float(0),hm - (tp[tpc-1].h - tp[0].h));
                    n2->tr = tr_bw;
                    //n2->hm = (n1->hm+n2->hm)/2.;
                    //n1->hm = n2->hm;
                }
                else
                    if(n1 != NULL){
                        /*Ringstrasse bspw.*/
                        tp = new trackpoint[tpc];
                        fread(tp,sizeof(trackpoint),tpc,f);
                        n1->tp = tp;
                        n1->tpc = tpc;
                        n1->brthr = n1;
                        n1->hm = hm;
                        n1->tr = tr_fw;
                    }
                    else{
                        /*Straßen existieren bereits, seek setzen statt
                        weiter auslesen*/
                        del_Neighbour(k,n1);
                        del_Neighbour(kn,n2);
                        fseek(f,tpc*sizeof(trackpoint),SEEK_CUR);
                    }
            }
            else{
                fseek(f,tpc*sizeof(trackpoint),SEEK_CUR);
            }
        }
    }
    else
        /*sonst kann die Kreuzung nicht angelegt werden, sollte aber nicht
        passieren*/
        for(short i = 0; i < c; i++){
            fseek(f,  (  sizeof(readdirection)
                        +2*sizeof(float)
                        +2*sizeof(qint64)
                        +2*sizeof(char)
                        +2*sizeof(int)),SEEK_CUR);
            fread(&(tpc),	sizeof(int),1,f);
            fseek(f,tpc*sizeof(trackpoint),SEEK_CUR);
        }
}
void OSM::alloc_crossdata(KOO* k){
    k->cd = new crossdata;
    k->cd->neighbours = NULL;
    k->cd->s = new STRONG[2];
    k->cd->s->from = NULL;
    k->cd->s->via = NULL;
}

void OSM::load_lines(lines** l, FILE* f){
    int c;
    int len;
    Way* w;
    char name[10000];
    fread(&c,sizeof(int),1,f);
    for(int i = 0; i < c; i++){
        w = load_Way(f);
        fread(&len,sizeof(int),1,f);
        if(len > 0 && w != NULL){
            /*Way noch nicht vorhanden und besitzt Namen*/
            fread(name,sizeof(char),len,f);
            name[len]='\0';
            add_line(l,w,name);
        }
        else
            if(w != NULL)
                /*Way noch nicht vorhanden, aber kein Name*/
                add_line(l,w,NULL);
            else
                /*weg bereits vorhanden*/
                fseek(f,len*sizeof(char),SEEK_CUR);
    }
}
void OSM::load_points(points** p, FILE* f){
    int c;
    int len;
    float index;
    KOO* k;
    char name[1000];
    fread(&c,sizeof(int),1,f);
    points *pt = *p;
    for(int i = 0; i < c; i++){
        k = load_node(f);
        fread(&index,sizeof(float),1,f);
        fread(&len,sizeof(int),1,f);
        fread(name,sizeof(char),len,f);
        if(len > 0 && k->status == 0){
            name[len] = '\0';
            pt = add_points(&pt,new points(k,name,index));
        }
        else
            if(k->status == 0)
                pt = add_points(&pt,new points(k,NULL,index));
    }
    if(pt != NULL)
        *p = pt->root();
}
void OSM::check_lines(){
    for(int i = 0; i <= MaxNature; i++)
        check_lines_rek(ln[i]);
}
void OSM::check_lines_rek(lines** l){
    lines dummy, *t = &dummy, *b;
    dummy.next = *l;
    while(t->next != NULL){
        if(t->next->way->type >= 98){
            /*Element loeschen*/
            b = t->next;
            t->next = t->next->next;
            delete(b);
        }
        else
            t = t->next;
    }
    *l = dummy.next;
}
void OSM::quotationreplace(char* a){
    int i = 0;
    while(a[i] != '\0'){
        if(a[i] == '\"')
            a[i] = '\'';
        i++;
    }
}
void OSM::htmlreplace(char* c){
    if(c != NULL){
        char html[3][6] = {{'&','#','3','9',';','\0'},
                           {'&','#','3','8',';','\0'},
                           {'&','#','3','4',';','\0'}};
        char newchar[3] = {39,38,34};
        for(int i = 0; i < 3; i++){
            while(strstr(c,html[i]) != NULL){
                char* ct = strstr(c,html[i]);
                ct[0] = newchar[i];
                strcpy(&ct[1],&ct[strlen(html[i])]);
            }
        }
    }
}


void OSM::select_cross(double lat, double lon, char mintype, char maxtype,
                       KOO** kdest, float* d,KOO* ksrc){
    if(ksrc != NULL){
        if(ksrc->status == 2){
            bool reachable= false;
            Neighbour* n = ksrc->cd->neighbours;
            while(n != NULL && ! reachable){
                if(n->way->type <= maxtype && n->way->type >= mintype)
                    reachable = true;
                n = n->next;
            }
            if(reachable){
                float d2 = distance(lat,lon,ksrc->lat,ksrc->lon);
                if(d2 < *d){
                    *d = d2;
                    *kdest = ksrc;
                }
            }
        }
        select_cross(lat,lon,mintype,maxtype,kdest,d,ksrc->l);
        select_cross(lat,lon,mintype,maxtype,kdest,d,ksrc->r);
    }
}

void OSM::select_cross(double lat, double lon, KOO** kdest, float* d,KOO* ksrc){
    if(ksrc != NULL){
        if(ksrc->status == 2){
            float d2 = distance(lat,lon,ksrc->lat,ksrc->lon);
            if(d2 < *d){
                *d = d2;
                *kdest = ksrc;
            }
        }
        select_cross(lat,lon,kdest,d,ksrc->l);
        select_cross(lat,lon,kdest,d,ksrc->r);
    }
}
Way* OSM::select_Way(double lat, double lon){
    KOO *k1 = get_first_KOO(), *k2 = NULL;
    Way k;
    k.minlat = float(lat - 0.0001);
    k.maxlat = float(lat + 0.0001);
    k.minlon = float(lon - 0.00015);
    k.maxlon = float(lon + 0.00015);
    lines dummy, *l;
    dummy.next = NULL;
    extract_Way(&dummy,get_first_Way(),&k,0,MaxNature);

    Way* wmin = NULL;
    float dmin = 1e15;
    l = dummy.next;
    while(l != NULL){
        double dminw = min_distance(l->way,lat,lon,NULL,NULL);
        if(dminw < dmin){
            wmin = l->way;
            dmin = dminw;
        }
        l = l->next;
    }
    /*Liste löschen*/
    while(dummy.next != NULL){
        l = dummy.next;
        dummy.next = l->next;
        delete(l);
    }

    return wmin;
}
double OSM::min_distance(Way* wc, double lat, double lon,
                    int* mini, double *dti){
    double d0 = 0;
    double d1 = 1e15;
    double d2 = distance(wc->node[0]->lat, wc->node[0]->lon,lat,lon);
    double dmin = d2;
    int imin = 0;
    double dt = 0;
    for(int i = 0; i < wc->nodec-1; i++){
        d1 = d2;
        d2 = distance(wc->node[i+1]->lat, wc->node[i+1]->lon,lat,lon);
        d0 = distance(wc->node[i  ]->lat, wc->node[i  ]->lon,
                      wc->node[i+1]->lat, wc->node[i+1]->lon);
        if(d1*d1+d0*d0 > d2*d2 && d2*d2 + d0*d0 > d1*d1){
            // there is a less distant point on the current way segment
            double dtt = (d1*d1-d2*d2+d0*d0)/2./d0;
            double dmint = sqrt(d1*d1-dtt*dtt);
            if(dmint < dmin){
                dt = dtt;
                imin = i;
                dmin = dmint;
            }
         }
         else{
            // node[i+1] may be the new closest point
                if(d2 < dmin){
                    dt = 0;
                    imin = i+1;
                    dmin = d2;
                }
         }
    }
    if(mini != NULL) *mini = imin;
    if(dti  != NULL) *dti  = dt;
    return dmin;
}

void OSM::extract_Way(lines* l, Way* w, Way* k, char mintype, char maxtype){
    if(w != NULL){
        if(w->type <= maxtype && w->type >= mintype
            && isin(w,k->minlat,k->maxlat,k->minlon,k->maxlon)){
            lines* tl = new lines;
            tl->way = w;
            tl->next = l->next;
            l->next = tl;
        }
        extract_Way(l,w->l,k,mintype,maxtype);
        extract_Way(l,w->r,k,mintype,maxtype);
    }
}
void OSM::change_Waytype(Way* w, char newtype){
    // set up new new line
    lines** l = NULL;
    if(newtype < MaxNature)
        l = ln[newtype];
    if(l != NULL){
        add_line(l,w,w->name);
        // first remove old map element
        lines** lold = NULL;
        char* namebuf;
        if(w->type < MaxNature)
            lold = ln[w->type];
        if(lold != NULL)
            del_lines(w,lold,namebuf);
        w->type = newtype;
    }
}
void OSM::del_lines(Way* w, lines** l, char* keepname){
    if(*l != NULL){
        lines* tl = *l, *tlbuffer;
        while(tl ->next != NULL && tl->next->way != w)
            tl = tl->next;				//Weg raussuchen
        if(tl->next != NULL){
            /*falls gefunden, dann aus der Liste entfernen*/
            tlbuffer = tl->next;
            tl->next = tl->next->next;
            keepname = tlbuffer->name;
            delete(tlbuffer);
        }
        else{
            /*falls es das erste Listenelement ist*/
            if((*l)->way == w){
                tlbuffer = (*l);
                *l = tlbuffer->next;
                keepname = tlbuffer->name;
                delete(tlbuffer);
            }
        }
    }
}
void OSM::del_Way(Way* w){
    lines** l = NULL;
    if(w->type < MaxNature)
        l = ln[w->type];
    char *name = NULL;
    del_lines(w,l,name);
    if(name != NULL)
        delete(name);
    for(int i = 0; i < w->nodec; i++){
        if(w->node[i]->cd != NULL){
            del_Neighbour(w->node[i],w);
        }
    }
    w->type = 99;
}
void OSM::del_Neighbour(KOO* k, Way* w){
    if(k->cd != NULL){
        Neighbour dummy, *n = &dummy, *nbuffer;
        dummy.next = k->cd->neighbours;
        while(n->next != NULL){
            if(n->next->way == w){
                /*Die Nachbarrelation n->next muss geloescht werden*/
                nbuffer = n->next;
                n->next = nbuffer->next;
                delete(nbuffer);
            }
            else
                n = n->next;
        }
        k->cd->neighbours = dummy.next;
    }
}
void OSM::del_Neighbour(KOO* k, Neighbour* n){
    if(n != NULL && k != NULL && k->cd != NULL){
        Neighbour dummy, *tn = &dummy;
        char found = 0;
        dummy.next = k->cd->neighbours;
        while(tn->next != NULL){
            if(tn->next == n){
                tn->next = tn->next->next;
                found = 1;
            }
            else
                tn = tn->next;
        }
        if(found)
            delete(n);
        k->cd->neighbours = dummy.next;
    }
}
short OSM::check_AVL(Way* w){
    if(w != NULL){
        if(w->l != NULL && w->l->id > w->id)
            return 0;
        if(w->r != NULL && w->r->id < w->id)
            return 0;
        return check_AVL(w->l)*check_AVL(w->r);
    }
    else
        return 1;
}
void OSM::replace_Way_KOO(Way* w, KOO** newKOO, int newlength){
    KOO** KOObuffer = w->node;
    int lengthbuffer = w->nodec;
    w->node = newKOO;
    w->nodec = newlength;
    for(int i = 0; i < lengthbuffer; i++)
        /*alle Wegpunkte untersuchen, ob sie Kreuzungen sind*/
        if(KOObuffer[i]->cd != NULL){
            Neighbour* nb = KOObuffer[i]->cd->neighbours;
            short deletenow = 0;
            while(nb != NULL){
                /*alle Nachbarsschaftsbeziehung der Kreuzung untersuchen*/
                if(nb->way == w){
                    /*die Nachbarschaftsbeziehung verläuft auf dem zu
                    veraendernden Weg w*/
                    int KOOcount = 0;
                    short countnow = 0;
                    for(int j = 0; j < newlength; j++){
                        /*Pruefe, ob die alten KOO auf dem neuem Weg vorhanden
                        sind*/
                        if(countnow)
                            KOOcount++;
                        if(w->node[j] == KOObuffer[i]){
                            if(!countnow){
                                countnow = 1;
                                nb->node = &(w->node[j]);
                                nb->rd = fw;
                                nb->brthr->node = &(w->node[j]);
                                nb->brthr->rd = bw;
                            }
                            else{
                                countnow = 0;
                                nb->nodec = KOOcount + 1;
                                nb->brthr->nodec = KOOcount + 1;
                            }
                        }
                        if(w->node[j] == nb->nb){
                            if(!countnow){
                                countnow = 1;
                                nb->node = &(w->node[j]);
                                nb->rd = bw;
                                nb->brthr->node = &(w->node[j]);
                                nb->brthr->rd = fw;
                            }
                            else{
                                countnow = 0;
                                nb->nodec = KOOcount + 1;
                                nb->brthr->nodec = KOOcount + 1;
                            }
                        }
                    }
                    if(countnow || !KOOcount)
                        /*wenigstens einer der Kreuzungen ist im neuem Weg
                        nicht vorhanden. Diese Nachbarschaftsbeziehung muss
                        aufgegeben werden*/
                        deletenow = 1;

                }
                if(deletenow){
                    Neighbour* nbbuffer = nb;
                    nb = nb->next;
                    del_Neighbour(nbbuffer->nb,nbbuffer->brthr);
                    del_Neighbour(KOObuffer[i],nbbuffer);
                    deletenow = 0;
                }
                else
                    nb = nb->next;
            }
        }
    delete(KOObuffer);
}
void OSM::equal_crossheight(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            short equal = 1;
            double h1,h2;
            if(n->rd == fw) h1 = n->tp[0       ].h;
            else			h1 = n->tp[n->tpc-1].h;
            /*Gibt es überhaupt Diskrepanzen bezueglich der Kreuzungshoehe*/
            while(n->next != NULL && equal == 1){
                if(n->next->rd == fw) h2 = n->next->tp[0             ].h;
                else				  h2 = n->next->tp[n->next->tpc-1].h;
                if(h1 != h2) equal = 0;
                h1 = h2;
                n = n->next;
            }
            /*Falls ja, sollten die Diskrepanzen beseitigt werden*/
            if(!equal){
                double avh = 0;			//Average height
                int nbc = 0;			//Neighbour count
                n = k->cd->neighbours;
                /*Alle abgehenden Wege sollen die Duchschnittshoehe der
                Kreuzung an der Kreuzung aufzeigen*/
                while(n != NULL){
                    if(n->rd == fw) avh = avh + n->tp[0             ].h;
                    else			avh = avh + n->tp[n->tpc-1].h;
                    nbc++;
                    n = n->next;
                }
                avh = 1.*avh/nbc;
                /*Korriere alle abgehenden Wege der Kreuzung linear*/
                double dh;
                n = k->cd->neighbours;
                while(n != NULL){
                    if(n->d > 0){
                        if(n->rd == fw){
                            dh = n->tp[0].h - avh;
                            for(int i = 0; i < n->tpc; i++){
                                n->tp[i].h = float(n->tp[i].h
                                             - dh* (n->d - n->tp[i].d)/n->d);
                            }
                        }
                        else{
                            dh = n->tp[n->tpc-1].h - avh;
                            for(int i = 0; i < n->tpc; i++){
                                n->tp[i].h = float(n->tp[i].h - dh* (n->tp[i].d)/n->d);
                            }
                        }
                    }
                    else{
                        if(n->rd == fw) n->tp[0].h = float(avh);
                        else	 n->tp[n->tpc-1].h = float(avh);
                    }
                    n = n->next;
                }
            }
        }
        equal_crossheight(k->l);
        equal_crossheight(k->r);
    }
}
void OSM::normalize_Hm(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            while(n != NULL){
                if(n->rd == fw){
                    n->hm = (n->hm + n->brthr->hm)/2.;
                    n->brthr->hm = n->hm;
                }
                n->d = n->tp[n->tpc-1].d;
                n = n->next;
            }
        }
        normalize_Hm(k->l);
        normalize_Hm(k->r);
    }
}
void OSM::invert_Hm(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            float hmswap;
            while(n != NULL){
                if(n->rd == fw){
                    hmswap = n->hm;
                    n->hm = n->brthr->hm;
                    n->brthr->hm = hmswap;
                }
                n = n->next;
            }
        }
        invert_Hm(k->l);
        invert_Hm(k->r);
    }
}
void OSM::swap_Hm_d(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            float hmswap;
            while(n != NULL){
                hmswap = n->hm;
                n->hm = n->d*1000+0.00001;
                n->d = hmswap;
                n = n->next;
            }
        }
        swap_Hm_d(k->l);
        swap_Hm_d(k->r);
    }
}
void OSM::reset_Hm(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            while(n != NULL){
                n->brthr->hm = n->hm = 0;
                if(n->rd == fw){
                    for(int i = 1; i < n->tpc; i++)
                        if(n->tp[i].d > n->tp[i-1].d){
                            if(n->tp[i].h > n->tp[i-1].h)
                                n->hm += (n->tp[i].h - n->tp[i-1].h);
                            else
                                n->brthr->hm -= (n->tp[i].h - n->tp[i-1].h);
                        }
                }
                n = n->next;
            }
        }
        reset_Hm(k->l);
        reset_Hm(k->r);
    }
}

void OSM::points_Hm(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            while(n != NULL){
                n->brthr->hm = n->hm = 0;
                if(n->rd == fw){
                    for(int i = 1; i < n->tpc; i++)
                        if(n->tp[i].d > n->tp[i-1].d){
                            if(n->tp[i].h > n->tp[i-1].h)
                                n->hm += (n->tp[i].h - n->tp[i-1].h)
                                        *(n->tp[i].h - n->tp[i-1].h)
                                        /(n->tp[i].d - n->tp[i-1].d)/1000;
                            else
                                n->brthr->hm += (n->tp[i].h - n->tp[i-1].h)
                                               *(n->tp[i].h - n->tp[i-1].h)
                                               /(n->tp[i].d - n->tp[i-1].d)/1000;
                        }
                }
                n = n->next;
            }
        }
        points_Hm(k->l);
        points_Hm(k->r);
    }
}
void OSM::refresh_time(KOO* k, float (*t)(Neighbour* n, void* inst),void* buf){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            while(n != NULL){
                n->t = t(n,buf);
                n = n->next;
            }
        }
        refresh_time(k->l,t,buf);
        refresh_time(k->r,t,buf);
    }
}

void OSM::merge(KOO* knew, KOO* kold){
    if(	knew != NULL && kold != NULL &&
        knew->cd != NULL && kold->cd != NULL &&
        knew->cd->neighbours != NULL && kold->cd->neighbours != NULL){
        double newh;
        if(knew->cd->neighbours->rd == fw)
            newh = knew->cd->neighbours->tp[0].h;
        else
            newh = knew->cd->neighbours->tp[knew->cd->neighbours->tpc-1].h;
        Neighbour* tn = kold->cd->neighbours;
        while(tn != NULL){
            KOO* klast;
            if(tn->rd == fw)	klast = tn->node[0];
            else				klast = tn->node[tn->nodec-1];
            double dd =  distance(klast->lat,klast->lon,knew->lat,knew->lon)
                        -distance(klast->lat,klast->lon,kold->lat,kold->lon);
            tn->d = tn->brthr->d = tn->d + dd;
            if(tn->rd == fw){
                for(int i = 1; i < tn->tpc; i++)
                    tn->tp[i].d = tn->tp[i].d+dd;
                tn->node[0] = knew;
                tn->tp[0].h = newh;
            }
            else{
                tn->tp[tn->tpc-1].d = tn->tp[tn->tpc-1].d + dd;
                tn->node[tn->nodec-1] = knew;
                tn->tp[tn->tpc-1].h = newh;
            }
            tn->brthr->nb = knew;
            double hmfw= 0, hmbw = 0;
            for(int i = 1; i < tn->tpc; i++)
                if(tn->tp[i].h > tn->tp[i-1].h)
                    hmfw = hmfw + tn->tp[i].h - tn->tp[i-1].h;
                else
                    hmbw = hmbw - tn->tp[i].h + tn->tp[i-1].h;

            if(tn->rd == fw){
                tn->hm = hmfw;
                tn->brthr->hm = hmbw;
            }
            else{
                tn->hm = hmbw;
                tn->brthr->hm = hmfw;
            }

            if(tn->nb == knew && tn != tn->brthr){
                del_Neighbour(knew,tn->brthr);
                tn = tn->next;
            }
            else{
                Neighbour* tb = tn->next;
                tn->next = knew->cd->neighbours;
                knew->cd->neighbours = tn;
                tn = tb;
            }

        }
        kold->cd = NULL;
        kold->status = 0;
    }
}
void OSM::deleteRect(double minlat, double maxlat,
                     double minlon, double maxlon){
    /*alle Wege löschen*/
    del_Way(get_first_Way(),minlat,maxlat,minlon,maxlon);
    /*alle Punkte löschen*/
    for(int i = 0; i < MaxPoints; i++){
        points* p = *pt[i];
        points pbuf; pbuf.next = p;
        if(p != NULL) p->prev = &pbuf;
        while(p != NULL){
            if(isin(p,minlat,maxlat,minlon,maxlon)){
                p->prev->next = p->next;
                if(p->next != NULL)
                    p->next->prev = p->prev;
            }
            p = p->next;
        }
        if(pbuf.next != NULL)
            pbuf.next->prev = NULL;
        *pt[i] = pbuf.next;
    }
}
void OSM::del_Way(Way* w, double minlat, double maxlat,
                     double minlon, double maxlon){
    if(w != NULL){
        if(iscompletein(w,minlat,maxlat,minlon,maxlon))
            del_Way(w);
        del_Way(w->l,minlat,maxlat,minlon,maxlon);
        del_Way(w->r,minlat,maxlat,minlon,maxlon);
    }
}
void OSM::delete_AVL(Way* l){
    if(l != NULL){
        if(l->node != NULL) delete l->node;
        if(l->name != NULL) delete l->name;
        delete_AVL(l->l);
        delete_AVL(l->r);
        delete l;
    }
}

void OSM::delete_AVL(KOO* l){
    if(l != NULL){
        if(l->cd != NULL){
            Neighbour *n = l->cd->neighbours, *nt;
            while(n != NULL){
                nt = n;
                n = n->next;
                if(nt->rd == fw)
                    delete nt->tp;
                delete nt;
            }
            if(l->cd->s != NULL) delete l->cd->s;
            delete l->cd;
        }
        delete_AVL(l->l);
        delete_AVL(l->r);
        delete l;
    }
}

void OSM::delete_lines(lines* l){
    lines* lt;
    while(l != NULL){
        lt = l;
        l = l->next;
        delete lt;
    }
}

void OSM::delete_points(points* l){
    points* lt;
    while(l != NULL){
        lt = l;
        l = l->next;
        if(lt->name != NULL) delete lt->name;
        delete lt;
    }
}

void OSM::delete_relation(relation* l){
    relation* lt;
    while(l != NULL){
        lt = l;
        l = l->next;


        if(lt->members != NULL
                ) delete l->members;
        delete lt;

    }
}

bool OSM::insert_cross(Way* w, double latc, double lonc){
    int imin;
    double dti;
    min_distance(w,latc,lonc,&imin,&dti);
    if(w->node[imin]->cd != NULL && dti == 0)
        // crossroad data already exists at the nearest point
        return false;
    KOO* k = NULL;
    if(dti != 0){
        // if a new point is to be inserted
        double d0 = distance(w->node[imin  ]->lat,w->node[imin  ]->lon,
                             w->node[imin+1]->lat,w->node[imin+1]->lon);
        // create new coordinate
        k = new KOO;
        k->lat =     dti /d0 * w->node[imin+1]->lat
               + (d0-dti)/d0 * w->node[imin  ]->lat;
        k->lon =     dti /d0 * w->node[imin+1]->lon
               + (d0-dti)/d0 * w->node[imin  ]->lon;
        k->cd = NULL;
        // append to AVL tree
        KOO* pk = get_first_KOO(); while(pk->l != NULL) pk = pk->l;
        k->balance = 0; k->id = pk->id-1;
        k->r = k->l = NULL; k->prev = pk; pk->l = k;
        // change all affected data sets
        w->nodec++;
        KOO** node_new = new KOO*[w->nodec];
        imin++;
        for(int i = 0     ; i < imin; i++)     node_new[i] = w->node[i];
        for(int i = imin+1; i < w->nodec; i++) node_new[i] = w->node[i-1];
        node_new[imin] = k;
        w->node = node_new;

    }
    else{
        k = w->node[imin];
    }
    // split neighbour relation on way segment
    int iprev = imin-1;
    while(w->node[iprev]->cd == NULL) iprev--;
    int inext = imin+1;
    while(w->node[inext]->cd == NULL) inext++;
    double dtot = 0;
    for(int j = iprev; j < imin; j++)
        dtot += distance(w->node[j  ]->lat,w->node[j  ]->lon,
                         w->node[j+1]->lat,w->node[j+1]->lon);
    Neighbour* nprev = w->node[iprev]->cd->neighbours;
    Neighbour* nnext = w->node[inext]->cd->neighbours;
    while(nprev != NULL && (nprev->way != w || nprev->rd != fw)) nprev = nprev->next;
    while(nnext != NULL && (nnext->way != w || nnext->rd != bw)) nnext = nnext->next;
    Neighbour* new_prev = add_Neighbour(k,&(w->node[iprev]),w->node[iprev],w,dtot         ,imin-iprev+1,bw);
    Neighbour* new_next = add_Neighbour(k,&(w->node[ imin]),w->node[inext],w,nnext->d-dtot,inext-imin+1,fw);
    // generate new height data
    int last_tpi = 0; while(nprev->tp[last_tpi].d < dtot) last_tpi++;
    new_prev->tpc = last_tpi+1;            new_prev->tp = new trackpoint[new_prev->tpc];
    new_next->tpc = nprev->tpc-last_tpi+1; new_next->tp = new trackpoint[new_next->tpc];
    for(int j = 0; j < last_tpi; j++)           new_prev->tp[j           ] = nprev->tp[j];
    for(int j = last_tpi; j < nprev->tpc; j++)  new_next->tp[j-last_tpi+1] = nprev->tp[j];
    double new_height = (nprev->tp[last_tpi].d - dtot  )/(nprev->tp[last_tpi].d - nprev->tp[last_tpi-1].d) * nprev->tp[last_tpi-1].h
                      + (dtot - nprev->tp[last_tpi-1].d)/(nprev->tp[last_tpi].d - nprev->tp[last_tpi-1].d) * nprev->tp[last_tpi  ].h;
    new_prev->tp[new_prev->tpc-1].d = dtot;
    new_prev->tp[new_prev->tpc-1].h = new_height;
    new_next->tp[0].d = 0;
    new_next->tp[0].h = new_height;
    // create new neighbour brother relations and delete the old ones
    new_prev->brthr     = add_Neighbour(w->node[iprev],&(w->node[iprev]),k,w,dtot         ,imin-iprev+1,fw);
    new_prev->brthr->brthr = new_prev;
    new_prev->brthr->tp  = new_prev->tp ;
    new_prev->brthr->tpc = new_prev->tpc;
    new_next->brthr     = add_Neighbour(w->node[inext],&(w->node[ imin]),k,w,nnext->d-dtot,inext-imin+1,bw);
    new_next->brthr->brthr = new_next;
    new_next->brthr->tp  = new_next->tp ;
    new_next->brthr->tpc = new_next->tpc;
    del_Neighbour(w->node[iprev],nprev);
    del_Neighbour(w->node[inext],nnext);
    return true;
}
void OSM::add_user_track(KOOlist *kl, trackpoint *tpl, int tc){
    // prepare way
    Way *w = new Way;
    w->type = 4;
    // gett AVL tree branches
    KOO* pk = get_first_KOO();
    while(pk->l != NULL)
        pk = pk->l;
    Way* pw = get_first_Way();
    while(pw->l != NULL)
        pw = pw->l;
    // add KOO to AVL tree
    KOOlist* kt = kl;
    while(kt != NULL){
        if(kt->k->id == -1){
            kt->k->balance = 0;
            kt->k->id = qMin(qint64(-2),pk->id-1);
            kt->k->r = kt->k->l = NULL;
            kt->k->prev = pk;
            pk->l = kt->k;
            pk = kt->k;
        }
        kt = kt->next;
        w->nodec++;
    }
    // add way to AVL tree
    w->id = qMin(qint64(-2),pw->id-1);
    pw->l = w;
    w->prev = pw;
    w->node = new KOO*[w->nodec];
    kt = kl;
    int i = 0;
    while(kt != NULL){
        w->node[i] = kt->k;
        kt = kt->next;
        i++;
    }
    add_line(ln[MaxTrackType],w,"user defined");
    calc_brect(w);
    // create neighbour relation
    float hm = 0; for(int i = 1; i < tc; i++) hm += qMax(tpl[i].h-tpl[i-1].h,float(0));
    Neighbour *n1 = add_Neighbour(w->node[0]         ,&(w->node[0]),w->node[w->nodec-1],w,tpl[tc-1].d,w->nodec,fw);
    Neighbour *n2 = add_Neighbour(w->node[w->nodec-1],&(w->node[0]),w->node[0]         ,w,tpl[tc-1].d,w->nodec,bw);
    n1->tp = tpl;
    n1->tpc = tc;
    n1->brthr = n2;
    n1->hm = hm;
    n2->tp = tpl;
    n2->tpc = tc;
    n2->brthr = n1;
    n2->hm = qMax(float(0),hm - (tpl[tc-1].h - tpl[0].h));
}
void OSM::calc_brect(Way* l){
    l->maxlat = -90;
    l->maxlon = -180;
    l->minlat =  90;
    l->minlon =  180;
    for(int j = 0; j < l->nodec; j++){
    if(l->node[j]->lat > l->maxlat )
        l->maxlat = float(l->node[j]->lat);
    if(l->node[j]->lat < l->minlat )
        l->minlat = float(l->node[j]->lat);
    if(l->node[j]->lon > l->maxlon )
        l->maxlon = float(l->node[j]->lon);
    if(l->node[j]->lon < l->minlon )
        l->minlon = float(l->node[j]->lon);
    }
}
void OSM::trim_names(int dgrid, QList<points *> **elegrid, int cx, int cy){
    // for each cell on the grip
    for(int x = 0 ; x < cx; x++)
        for(int y = 0 ; y < cy; y++)
            //investigate each element of the grid
            for(int i = 0; i < elegrid[x][y].length(); i++){
                points* p = elegrid[x][y].at(i);
                if(p->index > 0 && p->name != NULL && p->name[0] != '\0'){
                    // compare it with all neighbour cells
                    for(int nx = qMax(0,x-dgrid) ; nx < qMin(cx,x+dgrid+1); nx++)
                        for(int ny = qMax(0,y-dgrid) ; ny < qMin(cy,y+dgrid+1); ny++)
                            // i.e. with each neighbour cell element
                            for(int j = 0; j < elegrid[nx][ny].length(); j++){
                                points* n = elegrid[nx][ny].at(j);
                                if(n != p &&
                                   n->name != NULL &&
                                   n->name[0] != '\0' &&
                                   (strstr(n->name,p->name) != NULL || strstr(p->name,n->name) != NULL)){
                                    // matching names found, kill the longer
                                    if(strlen(n->name) > strlen(p->name))
                                        n->name = false;
                                    else{
                                        p->name = n->name;
                                        n->name = false;
                                    }
                                }
                             }
                }
            }
}

void OSM::trim_nodes(float index, int dgrid, bool demand_names,
                     QList<points *> **elegrid, int cx, int cy){
    // for each cell on the grip
    for(int x = 0 ; x < cx; x++)
        for(int y = 0 ; y < cy; y++)
            //investigate each element of the grid
            for(int i = 0; i < elegrid[x][y].length(); i++){
                points* p = elegrid[x][y].at(i);
                if(p->index == index){
                    // element p has the required type
                    if(demand_names && (p->name == NULL || p->name[0] == '\0')){
                        // but does not fullfill demands one name
                        elegrid[x][y].removeAt(i);
                        i--;
                        p->remove();
                    }
                    else{
                        // compare it with all neighbour cells
                        for(int nx = qMax(0,x-dgrid) ; nx < qMin(cx,x+dgrid+1); nx++)
                            for(int ny = qMax(0,y-dgrid) ; ny < qMin(cy,y+dgrid+1); ny++){
                                int jmax = elegrid[nx][ny].length();
                                // i.e. with each neighbour cell element
                                for(int j = 0; j < jmax; j++){
                                    points* n = elegrid[nx][ny].at(j);
                                    if(n != p &&
                                       n->index == index &&
                                       (!demand_names || (n->name != NULL && n->name[0] != '\0'))){
                                       // a conquering element was found
                                       j = jmax;
                                       ny = cy;
                                       nx = cx;
                                       elegrid[x][y].removeAt(i);
                                       i--;
                                       p->remove();
                                    }

                                }
                            }
                    }
                }
            }
}
void OSM::trim_nodes(float index, double mind, bool demand_name){
    points* p = *(pt[Hikei]);
    double a = 1;
    while(p != NULL){
        if(p->index == index){
            // is tree, search fhor close neighboor trees
            points* n = p->next;
            if(p->name != NULL && strstr(p->name,"Forsth") != NULL)
                a++;
            while(n != NULL &&
                  !(n->index == index &&
                    distance(p->k->lat,p->k->lon,n->k->lat,n->k->lon) < mind &&
                    (!demand_name || (n->name != NULL && n->name[0] != '\0'))))
                n = n->next;

            if(n != NULL || (demand_name && (p->name == NULL || p->name[0] == '\0')))
                p->remove();
        }
        p = p->next;
    }
}
void OSM::trim_points(){
    invert_points(pt[Hikei]);
    // first try to extract nodes from Ways
    // mark nodes
    reset_KOO_status(get_first_KOO(),0);
    for(int i = 0; i <= MaxNature; i++){
        lines* t = *ln[i];
        while(t != NULL){
            for(int i = 0; i < t->way->nodec; i++)
                t->way->node[i]->status++;
            t = t->next;
        }
    }
    // try to condense hike map symbols
    way_to_node(get_first_Way());
    // reset KOO status
    restore_KOO_status();
    // create helper data structure
    maxlat = -90;
    minlat = 90;
    maxlon = -180;
    minlon = 180;
    calc_brect(pt[7]);
    int sy = (maxlat - minlat)*111                                  /0.1;
    int sx = (maxlon - minlon)*111*cos((maxlat+minlat)/2.*M_PI/180.)/0.1;
    if(sx <= 0 || sy <= 0 || sx*sy > 3000*3000)
        return;                         // too large area, give up trimming
    QList<points*> **elegrid = new QList<points*>*[sx];
    for(int i = 0; i < sx; i++)
        elegrid[i] = new QList<points*>[sy];
    points* p = *(pt[Hikei]);
    while(p != NULL){
        int ix = qMin(sx-1,int((p->k->lon - minlon)/(maxlon - minlon) * sx));
        int iy = qMin(sy-1,int((p->k->lat - minlat)/(maxlat - minlat) * sy));
        elegrid[ix][iy].append(p);
        p = p->next;
    }
    // remove too dense map elements
    trim_nodes(35.,2,false,elegrid,sx,sy);      // trees
    trim_nodes(21.,1,false,elegrid,sx,sy);      // parking
    trim_nodes(22.,1,false,elegrid,sx,sy);      // bus stations
    trim_nodes(37.,0,true ,elegrid,sx,sy);      // named building
    // remove same names
    trim_names(2,elegrid,sx,sy);
    invert_points(pt[Hikei]);
}
void OSM::way_to_node(Way *l){
    if(l != NULL){
        double a = 0;
        if(l->type >= 20 && l->type < 90){
            // try to transform way into node
            double lat = 0;
            double lon = 0;
            int besti = 0;
            for(int i = 0; i < l->nodec; i++){
                lat += l->node[i]->lat;
                lon += l->node[i]->lon;
                if(l->node[i]->status < l->node[besti]->status) besti = i;
            }
            if(l->node[besti]->status == 0){
                l->node[besti]->lat = lat / l->nodec;
                l->node[besti]->lon = lon / l->nodec;
                l->node[besti]->status = 100;
                add_points(pt[7],new points(l->node[besti],l->name,l->type));
            }
        }
        way_to_node(l->l);
        way_to_node(l->r);
    }
}
void OSM::assign_traffic_restrictions(){
    // overwrite Status
    reset_KOO_status(get_first_KOO(),0);
    points* p = *(pt[Hikei]);
    while(p != NULL){
        if(p->index == 10)  p->k->status = qMax(tr_give_way,p->k->status);
        if(p->index == 11)  p->k->status = qMax(tr_stop,p->k->status);
        if(p->index == 12)  p->k->status = qMax(tr_lights,p->k->status);
        p = p->next;
    }
    // investigate all neighbour relations
    assign_traffic_restrictions(get_first_KOO());

    // reset Status
    restore_KOO_status();
}
void OSM::assign_traffic_restrictions(KOO* k){
    if(k != NULL){
        if(k->cd != NULL && k->cd->neighbours != NULL){
            Neighbour* n = k->cd->neighbours;
            while(n != NULL){
                // identify end of road
                KOO* nend;
                if(n->rd == fw) nend = n->node[n->nodec-1];
                else            nend = n->node[0];
                // search for traffic restrictions close to end
                for(int i = 0; i < n->nodec; i++)
                    if(n->node[i]->status > 0 && distance(n->node[i],nend) < 0.05)
                        n->tr = qMax(n->tr, n->node[i]->status);
                // look if tere is one-way-regulation on the way
                if(n->rd == bw && n->way->IsOneWay)
                        n->tr = tr_one_way;
                n = n->next;
            }
        }
        assign_traffic_restrictions(k->l);
        assign_traffic_restrictions(k->r);
    }
}
void OSM::restore_KOO_status(){
    reset_KOO_status(get_first_KOO(),-1);
    for(int i = 0; i <= MaxNature; i++){
        lines* t = *ln[i];
        while(t != NULL){
            for(int i = 0; i < t->way->nodec; i++)
                t->way->node[i]->status = 1;
            t = t->next;
        }
    }
    points* p = *(pt[Hikei]);
    while(p != NULL){
        p->k->status = 3;
        p = p->next;
    }
}
float OSM::d(KOO* k1, KOO* k2){
    return distance(k1->lat,k1->lon,k2->lat,k2->lon);
}
