#ifndef TSPMATRIX_H
#define TSPMATRIX_H

#include <QList>

class edge{
public:
    edge();
    edge(edge* src);
    bool is_precessor_of(edge* l);      //!< try to link to another edge
    bool is_succecessor_of(edge* l);    //!< try to link to another edge
    bool link_pre(edge* l);             //!< try to link at the start of l
    bool link_suc(edge* l);             //!< try to link at the end of l
    int from0();                        //!< start of chain
    int to0();                          //!< end of chain
    int from;                           //!< edge start
    int to;                             //!< edge end
    int row;                            //!< bolinging in matrix
    int col;                            //!< bolinging in matrix
    void reset_link();
    edge* fw();                         //!< go to next
    edge* bw();                         //!< go to previous
    edge* start();                      //!< go to begin of chain
private:
    edge *prev;                         //!< if edge can be connected
    edge *next;                         //!< if edge can be connected
};



class TspMatrix
{
public:
    TspMatrix(double** data, int c);
    TspMatrix* clone();
    ~TspMatrix();
    double lower_bound();               //!< returns sum
    void deactivate_cell(int row, int column);
    void deactivate_cell(edge* e);
    void shrink(int row, int column);   //!< removes row and column
    void reduce();                      //!< eliminates row and column minima
    edge* candidate(double* penalty);   //!< best edge to contribute to solution
    double row_min(int irow);           //!< row minimum
    double col_min(int icol);           //!< col minimum
    double row_min(int irow, int exc);  //!< row smalles aside exc
    double col_min(int icol, int exc);  //!< col smalles asied exc
    double accept_candidate();          //!< search for and procceed with canditate
    double refuse_candidate();          //!< search for and refuse canditate
    void fill_lists(QList<edge*> *solution0, QList<int>* row0,
                    QList<int>* col0, double sum);
    void primitive_solution();
    bool finished(int maxi);            //!< check if Tsp Solution is ready
    edge* get_solution();               //!< returns first element of solution
    QList<edge*>* solutionlist();
        double** d;                         //!< data
            int c;                              //!< dimension
private:
    QList<edge*> solution;               //!< List of all removed edges
    QList<int> row;                     //!< indeces of rows;
    QList<int> col;                  //!< indeces of columns;
    double sum;                         //!< accumulated penalty
};
#endif // TSPMATRIX_H
