#ifndef WEBDATA_H
#define WEBDATA_H
#include "osm.h"
#include "HeightData.h"
#include "MAP.h"
#include "strongcalc.h"
#include "Routing.h"
#include <QThread>
#include <QFileInfo>
#include <QProcess>
#include <QDir>

enum RunModeWebData{CreateTilesBTP,CreateTiles,SplitEurope,ReadOSM,CreateHTML,
                    CreateSTRONGLayers,SummitCalc};

//! multithreading working command, defines the task to be executed
typedef struct webdatacmd{
    RunModeWebData mode;                //!< type of task
    //!@{
    //! bounding rect to fulfil the tast
    double minlat, maxlat, minlon, maxlon;
    //!@}
    int width;                          //!< rendering image size
    int height;                         //!< rendering image size
    config* c;                          //!< BTP configuration to be used
    QString dest;                       //!< destination file name
    OSM* osm;                           //!< data to be painted
    //!@{
    //! Google tile sheme
    int z,x,y,dx,dy;
    //!@}
    QString src;                        //!< source file name
}webdatacmd;
Q_DECLARE_METATYPE(webdatacmd)

//! creates data for web site
/*!
  this class is a threat which allows to parallely draw tile, compute climb
  data, read osm and split raw pbf-$osm data. Aim is to fasten the calculation
  by using multiple cores of the processor. That means this is the basic class
  of web content calculation with multithreating.
*/
class WebData:public QThread{
Q_OBJECT
public:
    //! contructor to create png tiles of bounding rect
    WebData(webdatacmd w);
    void run();
    void exit();
    webdatacmd w;
public slots:
    void split_europe();                 /*!< splite osm europe file*/
signals:
    void add(webdatacmd* w);
    void split(WebData* wd);
private:
    QString dest;                       /*!< destination to store result*/
    void generate_map();                /*!< generate tiles without BTP data*/
    void generate_map_BTP();            /*!< generate tiles with BTP data*/
    void generate_STRONG();             /*!< generates STRONG tiles*/
    //! creats *.osm file from pbf and reads in with external BTP process
    void read_osm();
    //! analyses climbs on btp files and saves html data
    void analyse_climbs();

    double xztolon(int x, int z);       //!< longitude of tile x in zoom z
    double yztolat(int y, int z);       //!< lattitude of tile y in zoom z

};

#endif // WEBDATA_H
